﻿// AngleOps.cs
// .NET adapters for the IAU SOFA AngleOps functions.
// Copyright Paul S. Hirose 2017

// revision history
#if false
2017-09-17 Created for the SOFA AngleOps functions, which were originally part of another file.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {

    /* VectorMatrix/AngleOps */


    // DECLSPEC void iauA2af(int ndp, double angle, char *sign, int idmsf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauA2af(int ndp, double angle, out char sign, int[] idmsf);

    /// <summary>
    /// Decompose radians into degrees, arcminutes, arcseconds, fraction.
    /// </summary>
    /// <param name="ndp">resolution</param>
    /// <param name="angle">angle in radians</param>
    /// <param name="sign">'+' or '-'</param>
    /// <param name="idmsf">degrees, arcminutes, arcseconds, fraction</param>
    /// <exception cref="ArgumentException"> if <paramref name="idmsf"/> does not have dimension [4]
    /// </exception>
    public static void A2af(int ndp, double angle, out char sign, int[] idmsf) {
        SofaValidate.VerifyInt1x4(idmsf);
        iauA2af(ndp, angle, out sign, idmsf);
    }


    // DECLSPEC void iauA2tf(int ndp, double angle, char *sign, int ihmsf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauA2tf(int ndp, double angle, out char sign, int[] ihmsf);

    /// <summary>
    /// Decompose radians into hours, minutes, seconds, fraction.
    /// </summary>
    /// <param name="ndp">resolution</param>
    /// <param name="angle">angle in radians</param>
    /// <param name="sign">'+' or '-'</param>
    /// <param name="ihmsf">hours, minutes, seconds, fraction</param>
    /// <exception cref="ArgumentException"> if <paramref name="ihmsf"/> does not have dimension [4]
    /// </exception>
    public static void A2tf(int ndp, double angle, out char sign, int[] ihmsf) {
        SofaValidate.VerifyInt1x4(ihmsf);
        iauA2tf(ndp, angle, out sign, ihmsf);
    }


    // DECLSPEC int iauAf2a(char s, int ideg, int iamin, double asec, double *rad);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauAf2a(char s, int ideg, int iamin, double asec, out double rad);

    /// <summary>
    /// Convert degrees, arcminutes, arcseconds to radians.
    /// </summary>
    /// <param name="s">'-' = negative, otherwise positive</param>
    /// <param name="ideg">degrees</param>
    /// <param name="iamin">arcminutes</param>
    /// <param name="asec">arcseconds</param>
    /// <param name="rad">angle in radians</param>
    /// <remarks>The result is computed even if any of the range checks fail. Negative ideg, iamin
    /// and/or asec produce a warning status, but the absolute value is used in the conversion.
    /// (SOFA revision of 2013 June 18)
    /// </remarks>
    /// <returns>0 = OK</returns>
    public static int Af2a(char s, int ideg, int iamin, double asec, out double rad) {
        return iauAf2a(s, ideg, iamin, asec, out rad);
    }


    // DECLSPEC double iauAnp(double a);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauAnp(double a);

    /// <summary>
    /// Normalize angle into the range [0, 2pi).
    /// </summary>
    /// <param name="a">angle (radians)</param>
    /// <returns>angle in range 0-2pi</returns>
    public static double Anp(double a) {
        return iauAnp(a);
    }


    // DECLSPEC double iauAnpm(double a);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauAnpm(double a);

    /// <summary>
    /// Normalize angle into the interval [-pi, +pi).
    /// </summary>
    /// <param name="a">angle (radians)</param>
    /// <returns>angle in range -pi to +pi</returns>
    /// <remarks>SOFA revision of 2013 June 18</remarks>
    public static double Anpm(double a) {
        return iauAnpm(a);
    }


    // D2dtf() depends on SOFA UTC code and therefore is excluded from production code.
#if DEBUG
    // DECLSPEC int iauD2dtf(const char *scale, int ndp, double d1, double d2,
    //         int *iy, int *im, int *id, int ihmsf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauD2dtf(char scale, int ndp, double d1, double d2,
             out int iy, out int im, out int id, int[] ihmsf);

    /// <summary>
    /// Format for output a 2-part Julian Date (or in the case of UTC a quasi-JD form that includes
    /// special provision for leap seconds).
    /// </summary>
    /// <param name="scale">only "UTC" (case sensitive) is legal</param>
    /// <param name="ndp">number of decimal places in seconds</param>
    /// <param name="d1">Julian date part 1</param>
    /// <param name="d2">Julian date part 2</param>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day (Gregorian calendar)</param>
    /// <param name="ihmsf">hours, minutes, seconds, fraction</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException"> if <paramref name="ihmsf"/> does not have
    /// dimension [4].</exception>
    /// <remarks>This revision:  2014 February 1</remarks>
    public static int D2dtf(char scale, int ndp, double d1, double d2,
             out int iy, out int im, out int id, int[] ihmsf) {
        SofaValidate.VerifyInt1x4(ihmsf);
        return iauD2dtf(scale, ndp, d1, d2, out iy, out im, out id, ihmsf);
    }
#endif


    // DECLSPEC void iauD2tf(int ndp, double days, char *sign, int ihmsf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauD2tf(int ndp, double days, out char sign, int[] ihmsf);

    /// <summary>
    /// Decompose days to hours, minutes, seconds, fraction.
    /// </summary>
    /// <param name="ndp">resolution</param>
    /// <param name="days">interval in days</param>
    /// <param name="sign">'+' or '-'</param>
    /// <param name="ihmsf">hours, minutes, seconds, fraction</param>
    /// <exception cref="ArgumentException"> if <paramref name="ihmsf"/> does not have
    /// dimension [4]</exception>
    public static void D2tf(int ndp, double days, out char sign, int[] ihmsf) {
        SofaValidate.VerifyInt1x4(ihmsf);
        iauD2tf(ndp, days, out sign, ihmsf);
    }


    // DECLSPEC int iauTf2a(char s, int ihour, int imin, double sec, double *rad);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTf2a(char s, int ihour, int imin, double sec, out double rad);

    /// <summary>
    /// Convert hours, minutes, seconds to radians.
    /// </summary>
    /// <param name="s">'-' = negative, otherwise positive</param>
    /// <param name="ihour">hours</param>
    /// <param name="imin">minutes</param>
    /// <param name="sec">seconds</param>
    /// <param name="rad">angle in radians</param>
    /// <returns>0 = OK</returns>
    public static int Tf2a(char s, int ihour, int imin, double sec, out double rad) {
        return iauTf2a(s, ihour, imin, sec, out rad);
    }


    // DECLSPEC int iauTf2d(char s, int ihour, int imin, double sec, double *days);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTf2d(char s, int ihour, int imin, double sec, out double days);

    /// <summary>
    /// Convert hours, minutes, seconds to days.
    /// </summary>
    /// <param name="s">'-' = negative, otherwise positive</param>
    /// <param name="ihour">hours</param>
    /// <param name="imin">minutes</param>
    /// <param name="sec">seconds</param>
    /// <param name="days">days</param>
    /// <returns>0 = OK</returns>
    public static int Tf2d(char s, int ihour, int imin, double sec, out double days) {
        return iauTf2d(s, ihour, imin, sec, out days);
    }


    // Self test this part of the class.
#if DEBUG

    /// <summary>
    /// Self test AngleOps methods.
    /// </summary>
    public static void TestAngleOps() {
        double a, d;
        int j, iy, im, id;
        int[] idmsf;
        int[] ihmsf;
        char s;

        idmsf = new int[4];
        ihmsf = new int[4];

        /* VectorMatrix/AngleOps */

        // DECLSPEC void iauA2af(int ndp, double angle, char *sign, int idmsf[4]);
        A2af(4, 2.345, out s, idmsf);
        if (s != '+')
            throw new Exception("Wrong sign.");
        SofaValidate.EqualsApprox(idmsf[0], 134, 0);
        SofaValidate.EqualsApprox(idmsf[1], 21, 0);
        SofaValidate.EqualsApprox(idmsf[2], 30, 0);
        SofaValidate.EqualsApprox(idmsf[3], 9706, 0);

        // DECLSPEC void iauA2tf(int ndp, double angle, char *sign, int ihmsf[4]);
        A2tf(4, -3.01234, out s, ihmsf);
        if (s != '-')
            throw new Exception("Wrong sign.");
        SofaValidate.EqualsApprox(ihmsf[0], 11, 0);
        SofaValidate.EqualsApprox(ihmsf[1], 30, 0);
        SofaValidate.EqualsApprox(ihmsf[2], 22, 0);
        SofaValidate.EqualsApprox(ihmsf[3], 6484, 0);

        // DECLSPEC int iauAf2a(char s, int ideg, int iamin, double asec, double *rad);
        j = Af2a('-', 45, 13, 27.2, out a);
        SofaValidate.EqualsApprox(a, -0.7893115794313644842, 1e-12);
        SofaValidate.EqualsApprox(j, 0, 0);

        // DECLSPEC double iauAnp(double a);
        SofaValidate.EqualsApprox(Anp(-0.1), 6.183185307179586477, 1e-12);

        // DECLSPEC double iauAnpm(double a);
        SofaValidate.EqualsApprox(Anpm(-4.0), 2.283185307179586477, 1e-12);

        // DECLSPEC int iauD2dtf(const char *scale, int ndp, double d1, double d2,
        //         int *iy, int *im, int *id, int ihmsf[4]);
        j = D2dtf("UTC", 5, 2400000.5, 49533.99999, out iy, out im, out id, ihmsf);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(iy, 1994, 0);
        SofaValidate.EqualsApprox(im, 6, 0);
        SofaValidate.EqualsApprox(id, 30, 0);
        SofaValidate.EqualsApprox(ihmsf[0], 23, 0);
        SofaValidate.EqualsApprox(ihmsf[1], 59, 0);
        SofaValidate.EqualsApprox(ihmsf[2], 60, 0);
        SofaValidate.EqualsApprox(ihmsf[3], 13599, 0);

        // DECLSPEC void iauD2tf(int ndp, double days, char *sign, int ihmsf[4]);
        D2tf(4, -0.987654321, out s, ihmsf);
        if (s != '-')
            throw new Exception("Wrong sign.");
        SofaValidate.EqualsApprox(ihmsf[0], 23, 0);
        SofaValidate.EqualsApprox(ihmsf[1], 42, 0);
        SofaValidate.EqualsApprox(ihmsf[2], 13, 0);
        SofaValidate.EqualsApprox(ihmsf[3], 3333, 0);

        // DECLSPEC int iauTf2a(char s, int ihour, int imin, double sec, double *rad);
        j = Tf2a('+', 4, 58, 20.2, out a);
        SofaValidate.EqualsApprox(a, 1.301739278189537429, 1e-12);
        SofaValidate.EqualsApprox(j, 0, 0);

        // DECLSPEC int iauTf2d(char s, int ihour, int imin, double sec, double *days);
        SofaValidate.EqualsApprox(Tf2d(' ', 23, 55, 10.9, out d), 0, 0);
        SofaValidate.EqualsApprox(d, 0.9966539351851851852, 1e-12);
    }
#endif  // end self test
    }   // end class Sofa
}   // end namespace HirosePS
