﻿// Astrometry.cs
// .NET adapters for the SOFA Astrometry and SpaceMotion routines.
// Copyright Paul S. Hirose, 2017

// revision notes
#if false
2014-07-25  Began work.
2014-08-16  About a week delay to find a way to pass an array of structs. Resumed work and got
as far as Apco13(). Self tests all OK.
2014-08-22  Done, and self test OK! One loose end is iauPmsafe(), which seems to appear twice.
Need to verify that.
2014-08-29  iauPmsafe() is indeed declared twice, so I have completed all methods.
2014-08-31  Added missing self test for Atic13(). Several tests called the SOFA routine directly
instead of my method.
2014-10-03  Ensured all array arguments are checked for proper dimensions.
2014-10-04  Verified the self test calls my methods, not SOFA. Next check is to verify no blank
XML comments.
2014-10-05  No blank XML comments. Excluded methods with UTC parameters from the release build.

2017-09-10  SofaJpl 2.1. Modified tests of Apc family to match minor modification to SOFA code.
Modified PMSafe(), Pvstar(), Starpv() tests too.
#endif

// Anything with UTC is excluded from the production build so I don't have to rebuild the product
// every time there's a leap second. My classes provide their own UTC facilities.



using System;
using System.Runtime.InteropServices;


namespace HirosePS {
    /// <summary>
    /// Static class of .NET adapters for the IAU SOFA routines.
    /// </summary>
public static partial class Sofa {

    /* Astronomy/Astrometry */

    // DECLSPEC void iauAb(double pnat[3], double v[3], double s, double bm1,
    //       double ppr[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAb(double[] pnat, double[] v, double s, double bm1, double[] ppr);

    /// <summary>
    /// Apply aberration to transform natural direction into proper direction.
    /// </summary>
    /// <param name="pnat">natural direction to the source (unit vector)</param>
    /// <param name="v">observer barycentric velocity in units of c</param>
    /// <param name="s">distance between the Sun and the observer (au)</param>
    /// <param name="bm1">sqrt(1-|v|^2): reciprocal of Lorenz factor</param>
    /// <param name="ppr">proper direction to source (unit vector)</param>
    /// <exception cref="ArgumentException">if any array has incorrect dimensions.</exception>
    public static void Ab(double[] pnat, double[] v, double s, double bm1, double[] ppr) {
        SofaValidate.VerifyDouble1x3(pnat);
        SofaValidate.VerifyDouble1x3(v);
        SofaValidate.VerifyDouble1x3(ppr);
        iauAb(pnat, v, s, bm1, ppr);
    }


    // DECLSPEC void iauApcg(double date1, double date2,
    //         double ebpv[2][3], double ehp[3],
    //         iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApcg(double date1, double date2, double[,] ebpv, double[] ehp,
    [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a geocentric observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and GCRS coordinates. The Earth ephemeris is supplied by
    /// the caller.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="ebpv">Earth barycentric pos/vel (au, au/day)</param>
    /// <param name="ehp">Earth heliocentric position (au)</param>
    /// <param name="astrom">returned star-independent astrometry parameters</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    /// <exception cref="ArgumentException">if any array has incorrect dimensions.</exception>
    public static void Apcg(double date1, double date2, double[,] ebpv, double[] ehp,
    SofaAstrom astrom) {
        SofaValidate.VerifyDouble2x3(ebpv);
        SofaValidate.VerifyDouble1x3(ehp);
        iauApcg(date1, date2, ebpv, ehp, astrom);
    }


    // DECLSPEC void iauApcg13(double date1, double date2, iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApcg13(double date1, double date2, [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a geocentric observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and GCRS coordinates. The caller supplies the date, and SOFA
    /// models are used to predict the Earth ephemeris.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Apcg13(double date1, double date2, SofaAstrom astrom) {
        iauApcg13(date1, date2, astrom);
    }


    // DECLSPEC void iauApci(double date1, double date2,
    //         double ebpv[2][3], double ehp[3],
    //         double x, double y, double s,
    //         iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApci(double date1, double date2, double[,] ebpv, double[] ehp,
    double x, double y, double s, [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and geocentric CIRS coordinates. The Earth ephemeris and
    /// CIP/CIO are supplied by the caller.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="ebpv">Earth barycentric position/velocity (au, au/day)</param>
    /// <param name="ehp">Earth heliocentric position (au)</param>
    /// <param name="x">CIP X (component of unit vector)</param>
    /// <param name="y">CIP Y (component of unit vector)</param>
    /// <param name="s">the CIO locator s (radians)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Apci(double date1, double date2, double[,] ebpv, double[] ehp,
    double x, double y, double s, SofaAstrom astrom) {
        SofaValidate.VerifyDouble2x3(ebpv);
        SofaValidate.VerifyDouble1x3(ehp);
        iauApci(date1, date2, ebpv, ehp, x, y, s, astrom);
    }


    // DECLSPEC void iauApci13(double date1, double date2,
    //           iauASTROM *astrom, double *eo);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApci13(double date1, double date2, [In, Out] SofaAstrom astrom,
        out double eo);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and geocentric CIRS coordinates. The caller supplies the date,
    /// and SOFA models are used to predict the Earth ephemeris and CIP/CIO.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <param name="eo">equation of the origins (ERA-GST)</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Apci13(double date1, double date2, SofaAstrom astrom, out double eo) {
        iauApci13(date1, date2, astrom, out eo);
    }


    // DECLSPEC void iauApco(double date1, double date2,
    //         double ebpv[2][3], double ehp[3],
    //         double x, double y, double s, double theta,
    //         double elong, double phi, double hm,
    //         double xp, double yp, double sp,
    //         double refa, double refb,
    //         iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApco(double date1, double date2,
    double[,] ebpv, double[] ehp,
    double x, double y, double s, double theta,
    double elong, double phi, double hm,
    double xp, double yp, double sp,
    double refa, double refb,
    [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and observed coordinates.  The caller supplies the Earth
    /// ephemeris, the Earth rotation information and the refraction constants as well as the
    /// site coordinates.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="ebpv">Earth barycentric PV (au, au/day)</param>
    /// <param name="ehp">Earth heliocentric P (au)</param>
    /// <param name="x">CIP X (component of unit vector)</param>
    /// <param name="y">CIP Y (component of unit vector)</param>
    /// <param name="s">CIO locator s (radians)</param>
    /// <param name="theta">Earth rotation angle (radians)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic north latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="sp">TIO locator s' (radians)</param>
    /// <param name="refa">refraction constant A (radians)</param>
    /// <param name="refb">refraction constant B (radians)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Apco(double date1, double date2,
    double[,] ebpv, double[] ehp,
    double x, double y, double s, double theta,
    double elong, double phi, double hm,
    double xp, double yp, double sp,
    double refa, double refb,
    SofaAstrom astrom) {
        SofaValidate.VerifyDouble2x3(ebpv);
        SofaValidate.VerifyDouble1x3(ehp);
        iauApco(date1, date2, ebpv, ehp, x, y, s, theta, elong, phi, hm, xp, yp, sp, refa, refb,
        astrom);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauApco13(double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          iauASTROM *astrom, double *eo);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauApco13(double utc1, double utc2, double dut1,
    double elong, double phi, double hm, double xp, double yp,
    double phpa, double tk, double rh, double wl,
    [In, Out] SofaAstrom astrom, out double eo);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between ICRS and observed coordinates.  The caller supplies UTC, site
    /// coordinates, ambient air conditions and observing wavelength, and SOFA models are used to
    /// obtain the Earth ephemeris, CIP/CIO and refraction constants.
    /// </summary>
    /// <param name="utc1">UTC Julian date part 1</param>
    /// <param name="utc2">UTC Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at the observer (millibars)</param>
    /// <param name="tk">ambient temperature at the observer (C)</param>
    /// <param name="rh">relative humidity at the observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <param name="eo">equation of the origins (ERA-GST)</param>
    /// <returns>0 = UTC OK</returns>
    public static int Apco13(double utc1, double utc2, double dut1,
    double elong, double phi, double hm, double xp, double yp,
    double phpa, double tk, double rh, double wl,
    [In, Out] SofaAstrom astrom, out double eo) {
        return iauApco13(utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tk, rh, wl,
            astrom, out eo);
    }
#endif


    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    // DECLSPEC void iauApcs(double date1, double date2, double pv[2][3],
    //         double ebpv[2][3], double ehp[3],
    //         iauASTROM *astrom);
    extern static void iauApcs(double date1, double date2, double[,] pv,
    double[,] ebpv, double[] ehp, [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For an observer whose geocentric position and velocity are known, prepare star-independent
    /// astrometry parameters for transformations between ICRS and GCRS.  The Earth ephemeris is
    /// supplied by the caller.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="pv">observer's geocentric pos/vel (m, m/s)</param>
    /// <param name="ebpv">Earth barycentric PV (au, au/day)</param>
    /// <param name="ehp">Earth heliocentric P (au)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    /// <exception cref="ArgumentException">if any array is null or wrong dimensions</exception>
    public static void Apcs(double date1, double date2, double[,] pv,
    double[,] ebpv, double[] ehp, SofaAstrom astrom) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(ebpv);
        SofaValidate.VerifyDouble1x3(ehp);
        iauApcs(date1, date2, pv, ebpv, ehp, astrom);
    }


    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    // DECLSPEC void iauApcs13(double date1, double date2, double pv[2][3],
    //           iauASTROM *astrom);
    extern static void iauApcs13(double date1, double date2, double[,] pv,
        [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For an observer whose geocentric position and velocity are known, prepare star-independent
    /// astrometry parameters for transformations between ICRS and GCRS.  The Earth ephemeris is
    /// from SOFA models.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="pv">observer's geocentric pos/vel (m and m/s)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <exception cref="ArgumentException">if <paramref name="pv"/> is null or wrong dimensions
    /// </exception>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Apcs13(double date1, double date2, double[,] pv, SofaAstrom astrom) {
        SofaValidate.VerifyDouble2x3(pv);
         iauApcs13(date1, date2, pv, astrom);
    }


    // DECLSPEC void iauAper(double theta, iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAper(double theta, [In, Out] SofaAstrom astrom);

    /// <summary>
    /// In the star-independent astrometry parameters, update only the Earth rotation angle,
    /// supplied by the caller explicitly.
    /// </summary>
    /// <param name="theta">Earth rotation angle (radians)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <remarks>For targets expressed as equinox based positions, such as classical geocentric
    /// apparent (RA,Dec), the supplied theta can be Greenwich apparent sidereal time rather than
    /// Earth rotation angle.</remarks>
    public static void Aper(double theta, SofaAstrom astrom) {
         iauAper(theta, astrom);
    }


    // DECLSPEC void iauAper13(double ut11, double ut12, iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAper13(double ut11, double ut12, [In, Out] SofaAstrom astrom);

    /// <summary>
    /// In the star-independent astrometry parameters, update only the Earth rotation angle.
    /// The caller provides UT1.
    /// </summary>
    /// <param name="ut11">UT1 Julian date part 1 (at midnight for max precision)</param>
    /// <param name="ut12">UT1 Julian date part 2 (&gt;=0 and &lt;1 for max precision)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    public static void Aper13(double ut11, double ut12, SofaAstrom astrom) {
         iauAper13(ut11, ut12, astrom);
    }


    // DECLSPEC void iauApio(double sp, double theta,
    //         double elong, double phi, double hm, double xp, double yp,
    //         double refa, double refb,
    //         iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauApio(double sp, double theta,
        double elong, double phi, double hm, double xp, double yp,
        double refa, double refb,
        [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between CIRS and observed coordinates. The caller supplies the Earth
    /// orientation information and the refraction constants as well as the site coordinates.
    /// </summary>
    /// <param name="sp">the TIO locator s' (radians)</param>
    /// <param name="theta">Earth rotation angle (radians)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic north latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="refa">refraction constant A (radians)</param>
    /// <param name="refb">refraction constant B (radians)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    public static void Apio(double sp, double theta,
        double elong, double phi, double hm, double xp, double yp, double refa, double refb,
        SofaAstrom astrom) {
        iauApio(sp, theta, elong, phi, hm, xp, yp, refa, refb, astrom);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauApio13(double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          iauASTROM *astrom);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauApio13(double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        [In, Out] SofaAstrom astrom);

    /// <summary>
    /// For a terrestrial observer, prepare star-independent astrometry parameters for
    /// transformations between CIRS and observed coordinates.  The caller supplies UTC, site
    /// coordinates, ambient air conditions and observing wavelength.
    /// </summary>
    /// <param name="utc1">UTC quasi Julian date part 1</param>
    /// <param name="utc2">UTC quasi Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic north latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">temperature at observer (C)</param>
    /// <param name="rh">relative humidity at observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="astrom">output: star-independent astrometry parameters</param>
    /// <returns>0 = UTC date OK</returns>
    public static int Apio13(double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        SofaAstrom astrom) {
        return iauApio13(utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tk, rh, wl, astrom);
    }
#endif


    // DECLSPEC void iauAtci13(double rc, double dc,
    //           double pr, double pd, double px, double rv,
    //           double date1, double date2,
    //           double *ri, double *di, double *eo);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtci13(double rc, double dc,
        double pr, double pd, double px, double rv,
        double date1, double date2,
        out double ri, out double di, out double eo);

    /// <summary>
    /// Transform ICRS star data, epoch J2000.0, to CIRS.
    /// </summary>
    /// <param name="rc">ICRS right ascension at J2000.0 (radians)</param>
    /// <param name="dc">ICRS declination at J2000.0 (radians)</param>
    /// <param name="pr">RA proper motion (radians/year, dRA/dt not cos(Dec)*dRA/dt)
    /// </param>
    /// <param name="pd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="ri">CIRS geocentric RA (radians)</param>
    /// <param name="di">CIRS geocentric Dec (radians)</param>
    /// <param name="eo">equation of the origins (ERA-GST)</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Atci13(double rc, double dc,
        double pr, double pd, double px, double rv,
        double date1, double date2,
        out double ri, out double di, out double eo) {
        iauAtci13(rc, dc, pr, pd, px, rv, date1, date2, out ri, out di, out eo);
    }


    // DECLSPEC void iauAtciq(double rc, double dc, double pr, double pd,
    //          double px, double rv, iauASTROM *astrom,
    //          double *ri, double *di);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtciq(double rc, double dc, double pr, double pd,
        double px, double rv, SofaAstrom astrom, out double ri, out double di);

    /// <summary>
    /// Quick ICRS, epoch J2000.0, to CIRS transformation, given precomputed star-independent
    /// astrometry parameters.
    /// </summary>
    /// <param name="rc">ICRS RA at J2000.0 (radians)</param>
    /// <param name="dc">ICRS Dec at J2000.0 (radians)</param>
    /// <param name="pr">RA proper motion (radians/year, dRA/dt not cos(Dec)*dRA/dt)</param>
    /// <param name="pd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="ri">CIRS RA (radians)</param>
    /// <param name="di">CIRS Dec (radians)</param>
    public static void Atciq(double rc, double dc, double pr, double pd,
        double px, double rv, SofaAstrom astrom, out double ri, out double di) {
        iauAtciq(rc, dc, pr, pd, px, rv, astrom, out ri, out di);
    }


    // DECLSPEC void iauAtciqn(double rc, double dc, double pr, double pd,
    //           double px, double rv, iauASTROM *astrom,
    //           int n, iauLDBODY b[], double *ri, double *di);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtciqn(double rc, double dc, double pr, double pd,
        double px, double rv, SofaAstrom astrom,
        int n, SofaLDBody[] b, out double ri, out double di);

    /// <summary>
    /// Quick ICRS, epoch J2000.0, to CIRS transformation, given precomputed star-independent
    /// astrometry parameters plus a list of light-deflecting bodies.
    /// </summary>
    /// <param name="rc">ICRS RA at J2000.0 (radians)</param>
    /// <param name="dc">ICRS Dec at J2000.0 (radians)</param>
    /// <param name="pr">RA proper motion (radians/year, dRA/dt not cos(Dec)*dRA/dt)</param>
    /// <param name="pd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="n">number of light deflecting bodies</param>
    /// <param name="b">data for each of the light deflecting bodies</param>
    /// <param name="ri">CIRS RA (radians)</param>
    /// <param name="di">CIRS Dec (radians)</param>
    public static void Atciqn(double rc, double dc, double pr, double pd,
        double px, double rv, SofaAstrom astrom,
        int n, SofaLDBody[] b, out double ri, out double di) {
        iauAtciqn(rc, dc, pr, pd, px, rv, astrom, n, b, out ri, out di);
    }


    // DECLSPEC void iauAtciqz(double rc, double dc, iauASTROM *astrom,
    //           double *ri, double *di);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtciqz(double rc, double dc, SofaAstrom astrom,
         out double ri, out double di);

    /// <summary>
    /// Quick ICRS to CIRS transformation, given precomputed star-independent astrometry parameters,
    /// and assuming zero parallax and proper motion.
    /// </summary>
    /// <param name="rc">ICRS astrometric RA (radians)</param>
    /// <param name="dc">ICRS astrometric Dec (radians)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="ri">CIRS RA (radians)</param>
    /// <param name="di">CIRS Dec (radians)</param>
    public static void Atciqz(double rc, double dc, SofaAstrom astrom,
         out double ri, out double di) {
         iauAtciqz(rc, dc, astrom, out ri, out di);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauAtco13(double rc, double dc,
    //          double pr, double pd, double px, double rv,
    //          double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          double *aob, double *zob, double *hob,
    //          double *dob, double *rob, double *eo);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauAtco13(double rc, double dc,
        double pr, double pd, double px, double rv,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double aob, out double zob, out double hob,
        out double dob, out double rob, out double eo);

    /// <summary>
    /// ICRS RA,Dec to observed place.  The caller supplies UTC, site coordinates, ambient air
    /// conditions and observing wavelength. SOFA models are used for the Earth ephemeris,
    /// bias precession nutation, Earth orientation and refraction.
    /// </summary>
    /// <param name="rc">ICRS right ascension at J2000.0 (radians)</param>
    /// <param name="dc">ICRS declination at J2000.0 (radians)</param>
    /// <param name="pr">RA proper motion (radians/year, dRA/dt not cos(Dec)*dRA/dt)</param>
    /// <param name="pd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="utc1">UTC quasi Julian date part 1</param>
    /// <param name="utc2">UTC quasi Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">temperature at observer (C)</param>
    /// <param name="rh">relative humidity at observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="aob">observed azimuth (radians: N=0,E=90)</param>
    /// <param name="zob">observed zenith distance (radians)</param>
    /// <param name="hob">observed hour angle (radians)</param>
    /// <param name="dob">observed declination (radians)</param>
    /// <param name="rob">observed right ascension (CIO-based, radians)</param>
    /// <param name="eo">equation of the origins (ERA-GST)</param>
    /// <returns>0 = UTC date OK</returns>
    public static int Atco13(double rc, double dc,
        double pr, double pd, double px, double rv,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double aob, out double zob, out double hob,
        out double dob, out double rob, out double eo) {

        return iauAtco13(rc, dc, pr, pd, px, rv, utc1, utc2, dut1, elong, phi, hm, xp, yp,
        phpa, tk, rh, wl, out aob, out zob, out hob, out dob, out rob, out eo);
    }
#endif


    // DECLSPEC void iauAtic13(double ri, double di,
    //           double date1, double date2,
    //           double *rc, double *dc, double *eo);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtic13(double ri, double di, double date1, double date2,
        out double rc, out double dc, out double eo);

    /// <summary>
    /// Transform star RA,Dec from geocentric CIRS to ICRS astrometric.
    /// </summary>
    /// <param name="ri">CIRS geocentric RA (radians)</param>
    /// <param name="di">CIRS geocentric Dec (radians)</param>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="rc">ICRS astrometric RA (radians)</param>
    /// <param name="dc">ICRS astrometric Dec (radians)</param>
    /// <param name="eo">equation of the origins (ERA-GST)</param>
    /// <remarks>TT can be used instead of TDB without any significant impact on accuracy.</remarks>
    public static void Atic13(double ri, double di, double date1, double date2,
        out double rc, out double dc, out double eo) {
        iauAtic13(ri, di, date1, date2, out rc, out dc, out eo);
    }


    // DECLSPEC void iauAticq(double ri, double di, iauASTROM *astrom,
    //          double *rc, double *dc);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAticq(double ri, double di, SofaAstrom astrom,
        out double rc, out double dc);

    /// <summary>
    /// Quick CIRS RA,Dec to ICRS astrometric place, given the star-independent astrometry
    /// parameters.
    /// </summary>
    /// <param name="ri">CIRS RA (radians)</param>
    /// <param name="di">CIRS Dec (radians)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="rc">ICRS astrometric RA (radians)</param>
    /// <param name="dc">ICRS astrometric Dec (radians)</param>
    public static void Aticq(double ri, double di, SofaAstrom astrom,
        out double rc, out double dc) {
        iauAticq(ri, di, astrom, out rc, out dc);
    }


    // DECLSPEC void iauAticqn(double ri, double di, iauASTROM *astrom,
    //           int n, iauLDBODY b[], double *rc, double *dc);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAticqn(double ri, double di, SofaAstrom astrom, int n, SofaLDBody[] b,
        out double rc, out double dc);

    /// <summary>
    /// Quick CIRS to ICRS astrometric place transformation, given the star-independent astrometry
    /// parameters plus a list of light-deflecting bodies.
    /// </summary>
    /// <param name="ri">CIRS RA (radians)</param>
    /// <param name="di">CIRS declination (radians)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="n">number of light-deflecting bodies</param>
    /// <param name="b">data for each of the n bodies</param>
    /// <param name="rc">ICRS astrometric RA (radians)</param>
    /// <param name="dc">ICRS astrometric declination (radians)</param>
    public static void Aticqn(double ri, double di, SofaAstrom astrom, int n, SofaLDBody[] b,
        out double rc, out double dc) {
            iauAticqn(ri, di, astrom, n, b, out rc, out dc);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauAtio13(double ri, double di,
    //          double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          double *aob, double *zob, double *hob,
    //          double *dob, double *rob);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauAtio13(double ri, double di,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double aob, out double zob, out double hob,
        out double dob, out double rob);

    /// <summary>
    /// CIRS RA,Dec to observed place.  The caller supplies UTC, site coordinates, ambient air
    /// conditions and observing wavelength.
    /// </summary>
    /// <param name="ri">CIRS right ascension (CIO-based, radians)</param>
    /// <param name="di">CIRS declination (radians)</param>
    /// <param name="utc1">UTC quasi Julian date part 1</param>
    /// <param name="utc2">UTC quasi Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">temperature at observer (C)</param>
    /// <param name="rh">relative humidity at the observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="aob">observed azimuth (radians: N=0,E=90)</param>
    /// <param name="zob">observed zenith distance (radians)</param>
    /// <param name="hob">observed hour angle (radians)</param>
    /// <param name="dob">observed declination (radians)</param>
    /// <param name="rob">observed right ascension (CIO-based, radians)</param>
    /// <returns>0 = UTC date OK</returns>
    public static int Atio13(double ri, double di,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double aob, out double zob, out double hob,
        out double dob, out double rob) {

        return iauAtio13(ri, di, utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tk, rh, wl,
            out aob, out zob, out hob, out dob, out rob);
    }
#endif


    // DECLSPEC void iauAtioq(double ri, double di, iauASTROM *astrom,
    //          double *aob, double *zob,
    //          double *hob, double *dob, double *rob);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauAtioq(double ri, double di, SofaAstrom astrom,
        out double aob, out double zob, out double hob, out double dob, out double rob);

    /// <summary>
    /// Quick CIRS to observed place transformation.
    /// </summary>
    /// <param name="ri">CIRS right ascension (radians?)</param>
    /// <param name="di">CIRS declination (radians?)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="aob">observed azimuth (radians: N=0,E=90)</param>
    /// <param name="zob">observed zenith distance (radians)</param>
    /// <param name="hob">observed hour angle (radians)</param>
    /// <param name="dob">observed declination (radians)</param>
    /// <param name="rob">observed right ascension (CIO-based, radians)</param>
    public static void Atioq(double ri, double di, SofaAstrom astrom,
        out double aob, out double zob, out double hob, out double dob, out double rob) {
        iauAtioq(ri, di, astrom, out aob, out zob, out hob, out dob, out rob);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauAtoc13(const char *type, double ob1, double ob2,
    //          double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          double *rc, double *dc);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl,
        CharSet = CharSet.Auto)]
    extern static int iauAtoc13(byte[] type, double ob1, double ob2,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double rc, out double dc);

    /// <summary>
    /// Observed place at a groundbased site to to ICRS astrometric RA,Dec. The caller supplies UTC,
    /// site coordinates, ambient air conditions and observing wavelength.
    /// </summary>
    /// <param name="type">type of coordinates ("R", "H" or "A")</param>
    /// <param name="ob1">observed Az, HA or RA (radians; Az is N=0,E=90)</param>
    /// <param name="ob2">observed ZD or Dec (radians)</param>
    /// <param name="utc1">UTC quasi Julian date part 1</param>
    /// <param name="utc2">UTC quasi Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">temperature at observer (C)</param>
    /// <param name="rh">relative humidity at the observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="rc">ICRS astrometric RA (radians)</param>
    /// <param name="dc">ICRS astrometric Dec (radians)</param>
    /// <returns>0 = UTC date OK</returns>
    public static int Atoc13(string type, double ob1, double ob2,
        double utc1, double utc2, double dut1,
        double elong, double phi, double hm, double xp, double yp,
        double phpa, double tk, double rh, double wl,
        out double rc, out double dc)
    {
        System.Text.ASCIIEncoding ascii = new System.Text.ASCIIEncoding();
        byte[] bytes = ascii.GetBytes(type);
        return iauAtoc13(bytes, ob1, ob2, utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa,
                tk, rh, wl, out rc, out dc);
    }
#endif


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauAtoi13(const char *type, double ob1, double ob2,
    //          double utc1, double utc2, double dut1,
    //          double elong, double phi, double hm, double xp, double yp,
    //          double phpa, double tk, double rh, double wl,
    //          double *ri, double *di);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl,
        CharSet = CharSet.Auto)]
    extern static int iauAtoi13(byte[] type, double ob1, double ob2, double utc1, double utc2,
        double dut1, double elong, double phi, double hm, double xp, double yp, double phpa,
        double tk, double rh, double wl, out double ri, out double di);

    /// <summary>
    /// Observed place to CIRS.  The caller supplies UTC, site coordinates, ambient air conditions
    /// and observing wavelength.
    /// </summary>
    /// <param name="type">type of coordinates ("R", "H" or "A")</param>
    /// <param name="ob1">observed Az, HA or RA (radians; Az is N=0,E=90)</param>
    /// <param name="ob2">observed ZD or Dec (radians)</param>
    /// <param name="utc1">UTC quasi Julian date part 1</param>
    /// <param name="utc2">UTC quasi Julian date part 2</param>
    /// <param name="dut1">UT1-UTC (seconds)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">temperature at observer (C)</param>
    /// <param name="rh">relative humidity at observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="ri">CIRS right ascension (CIO-based, radians)</param>
    /// <param name="di">CIRS declination (radians)</param>
    /// <returns>0 = UTC date OK</returns>
    public static int Atoi13(string type, double ob1, double ob2, double utc1, double utc2,
        double dut1, double elong, double phi, double hm, double xp, double yp, double phpa,
        double tk, double rh, double wl, out double ri, out double di)
    {
        System.Text.ASCIIEncoding ascii = new System.Text.ASCIIEncoding();
        byte[] bytes = ascii.GetBytes(type);
        return iauAtoi13(bytes, ob1, ob2, utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa,
        tk, rh, wl, out ri, out di);
    }
#endif


    // DECLSPEC void iauAtoiq(const char *type,
    //          double ob1, double ob2, iauASTROM *astrom,
    //          double *ri, double *di);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl,
        CharSet = CharSet.Auto)]
    extern static void iauAtoiq(byte[] type, double ob1, double ob2, SofaAstrom astrom,
        out double ri, out double di);

    /// <summary>
    /// Quick observed place to CIRS, given the star-independent astrometry parameters.
    /// </summary>
    /// <param name="type">type of coordinates: ("R", "H" or "A")</param>
    /// <param name="ob1">observed Az, HA or RA (radians; Az is N=0,E=90)</param>
    /// <param name="ob2">observed ZD or Dec (radians)</param>
    /// <param name="astrom">star-independent astrometry parameters</param>
    /// <param name="ri">CIRS right ascension (CIO-based, radians)</param>
    /// <param name="di">CIRS declination (radians)</param>
    public static void Atoiq(string type, double ob1, double ob2, SofaAstrom astrom,
        out double ri, out double di)
    {
        System.Text.ASCIIEncoding ascii = new System.Text.ASCIIEncoding();
        byte[] bytes = ascii.GetBytes(type);
        iauAtoiq(bytes, ob1, ob2, astrom, out ri, out di);
    }


    // DECLSPEC void iauLd(double bm, double p[3], double q[3], double e[3],
    //       double em, double dlim, double p1[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLd(double bm, double[] p, double[] q, double[] e,
        double em, double dlim, double[] p1);

    /// <summary>
    /// Apply light deflection by a solar-system body, as part of transforming coordinate direction
    /// into natural direction.
    /// </summary>
    /// <param name="bm">mass of the gravitating body (solar masses)</param>
    /// <param name="p">direction from observer to source (unit vector)</param>
    /// <param name="q">direction from body to source (unit vector)</param>
    /// <param name="e">direction from body to observer (unit vector)</param>
    /// <param name="em">distance from body to observer (au)</param>
    /// <param name="dlim">deflection limiter</param>
    /// <param name="p1">observer to deflected source (unit vector)</param>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    /// <remarks>The arguments <paramref name="p"/> and <paramref name="p1"/> can be the same array.
    /// </remarks>
    public static void Ld(double bm, double[] p, double[] q, double[] e,
       double em, double dlim, double[] p1)
    {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(q);
        SofaValidate.VerifyDouble1x3(e);
        SofaValidate.VerifyDouble1x3(p1);
       iauLd(bm, p, q, e, em, dlim, p1);
    }


    // DECLSPEC void iauLdn(int n, iauLDBODY b[], double ob[3], double sc[3],
    //        double sn[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLdn(int n, SofaLDBody[] b, double[] ob, double[] sc, double[] sn);

    /// <summary>
    /// For a star, apply light deflection by multiple solar-system bodies, as part of transforming
    /// coordinate direction into natural direction.
    /// </summary>
    /// <param name="b">data for each deflecting body</param>
    /// <param name="ob">barycentric position of the observer (au)</param>
    /// <param name="sc">observer to star coord direction (unit vector)</param>
    /// <param name="sn">observer to deflected star (unit vector)</param>
    /// <exception cref="ArgumentException">if any vector has wrong dimensions.</exception>
    /// <remarks>The arguments <paramref name="sc"/> and <paramref name="sn"/> can be the same
    /// array.</remarks>
    public static void Ldn(SofaLDBody[] b, double[] ob, double[] sc, double[] sn)
    {
        SofaValidate.VerifyDouble1x3(ob);
        SofaValidate.VerifyDouble1x3(sc);
        SofaValidate.VerifyDouble1x3(sn);
        iauLdn(b.Length, b, ob, sc, sn);
    }


    // DECLSPEC void iauLdsun(double p[3], double e[3], double em, double p1[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLdsun(double[] p, double[] e, double em, double[] p1);

    /// <summary>
    /// Light deflection by the Sun.
    /// </summary>
    /// <param name="p">direction from observer to source (unit vector)</param>
    /// <param name="e">direction from Sun to observer (unit vector)</param>
    /// <param name="em">distance from Sun to observer (au)</param>
    /// <param name="p1">observer to deflected source (unit vector)</param>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    /// <remarks>The arguments <paramref name="p"/> and <paramref name="p1"/> can be the same array.
    /// </remarks>
    public static void Ldsun(double[] p, double[] e, double em, double[] p1) {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(e);
        SofaValidate.VerifyDouble1x3(p1);
         iauLdsun(p, e, em, p1);
    }


    // DECLSPEC void iauPmpx(double rc, double dc, double pr, double pd,
    //         double px, double rv, double pmt, double vob[3],
    //         double pco[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPmpx(double rc, double dc, double pr, double pd,
        double px, double rv, double pmt, double[] vob, double[] pco);

    /// <summary>
    /// Apply proper motion and parallax.
    /// </summary>
    /// <param name="rc">ICRS RA at catalog epoch (radians)</param>
    /// <param name="dc">ICRS Dec at catalog epoch (radians)</param>
    /// <param name="pr">RA proper motion (radians/year, dRA/dt rather than cos(Dec)*dRA/dt)</param>
    /// <param name="pd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="pmt">proper motion time interval (SSB, Julian years)</param>
    /// <param name="vob">SSB to observer vector (au)</param>
    /// <param name="pco">coordinate direction (BCRS unit vector)</param>
    /// <exception cref="ArgumentException">if either array does not have dimension [3].
    /// </exception>
    public static void Pmpx(double rc, double dc, double pr, double pd,
        double px, double rv, double pmt, double[] vob, double[] pco)
    {
        SofaValidate.VerifyDouble1x3(vob);
        SofaValidate.VerifyDouble1x3(pco);
        iauPmpx(rc, dc, pr, pd, px, rv, pmt, vob, pco);
    }


    // DECLSPEC int iauPmsafe(double ra1, double dec1, double pmr1, double pmd1,
    //          double px1, double rv1,
    //          double ep1a, double ep1b, double ep2a, double ep2b,
    //          double *ra2, double *dec2, double *pmr2, double *pmd2,
    //          double *px2, double *rv2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauPmsafe(double ra1, double dec1, double pmr1, double pmd1,
        double px1, double rv1, double ep1a, double ep1b, double ep2a, double ep2b,
        out double ra2, out double dec2, out double pmr2, out double pmd2,
        out double px2, out double rv2);

    /// <summary>
    /// Star proper motion:  update star catalog data for space motion, with special handling to
    /// handle the zero parallax case.
    /// </summary>
    /// <param name="ra1">right ascension (radians), before</param>
    /// <param name="dec1">declination (radians), before</param>
    /// <param name="pmr1">RA proper motion (radians/year, coordinate rate not great circle rate),
    /// before</param>
    /// <param name="pmd1">Dec proper motion (radians/year), before</param>
    /// <param name="px1">parallax (arcseconds), before</param>
    /// <param name="rv1">radial velocity (km/s receding), before</param>
    /// <param name="ep1a">"before" epoch, part A (2451545.0 for max resolution)</param>
    /// <param name="ep1b">"before" epoch, part B</param>
    /// <param name="ep2a">"after" epoch, part A (2451545.0 for max resolution)</param>
    /// <param name="ep2b">"after" epoch, part B</param>
    /// <param name="ra2">right ascension (radians), after</param>
    /// <param name="dec2">declination (radians), after</param>
    /// <param name="pmr2">RA proper motion (radians/year, coordinate rate not great circle rate),
    /// after</param>
    /// <param name="pmd2">Dec proper motion (radians/year), after</param>
    /// <param name="px2">parallax (arcseconds), after</param>
    /// <param name="rv2">radial velocity (km/s receding), after</param>
    /// <returns>0 = no warnings or errors</returns>
    public static int Pmsafe(double ra1, double dec1, double pmr1, double pmd1,
        double px1, double rv1, double ep1a, double ep1b, double ep2a, double ep2b,
        out double ra2, out double dec2, out double pmr2, out double pmd2,
        out double px2, out double rv2) 
    {
        return iauPmsafe(ra1, dec1, pmr1, pmd1, px1, rv1, ep1a, ep1b, ep2a, ep2b,
            out ra2, out dec2, out pmr2, out pmd2, out px2, out rv2);
    }


    // DECLSPEC void iauPvtob(double elong, double phi, double hm,
    //          double xp, double yp, double sp, double theta,
    //          double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvtob(double elong, double phi, double hm,
        double xp, double yp, double sp, double theta, double[,] pv);

    /// <summary>
    /// Position and velocity of a terrestrial observing station with respect to the CIRS.
    /// </summary>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="hm">height above ellipsoid (meters)</param>
    /// <param name="xp">polar motion X coordinate (radians)</param>
    /// <param name="yp">polar motion Y coordinate (radians)</param>
    /// <param name="sp">TIO locator s' (radians)</param>
    /// <param name="theta">Earth rotation angle (radians)</param>
    /// <param name="pv">CIRS position/velocity vector (m, m/s)</param>
    /// <exception cref="ArgumentException">if array has wrong dimensions.</exception>
    public static void Pvtob(double elong, double phi, double hm,
        double xp, double yp, double sp, double theta, double[,] pv)
    {
        SofaValidate.VerifyDouble2x3(pv);
        iauPvtob(elong, phi, hm, xp, yp, sp, theta, pv);
    }


    // DECLSPEC void iauRefco(double phpa, double tk, double rh, double wl,
    //          double *refa, double *refb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRefco(double phpa, double tk, double rh, double wl,
        out double refa, out double refb);

    /// <summary>
    /// Determine the constants A and B in the atmospheric refraction model
    /// dZ = A tan Z + B tan^3 Z.
    /// </summary>
    /// <param name="phpa">pressure at observer (millibars)</param>
    /// <param name="tk">ambient temperature at observer (C)</param>
    /// <param name="rh">relative humidity at observer (range 0-1)</param>
    /// <param name="wl">wavelength (micrometers)</param>
    /// <param name="refa">tan Z coefficient (radians)</param>
    /// <param name="refb">tan^3 Z coefficient (radians)</param>
    /// <remarks>Z is the "observed" zenith distance (i.e. affected by refraction) and dZ is what to
    /// add to Z to give the "topocentric" (i.e. in vacuo) zenith distance.</remarks>
    public static void Refco(double phpa, double tk, double rh, double wl,
        out double refa, out double refb)
    {
        iauRefco(phpa, tk, rh, wl, out refa, out refb);
    }


    // self test code
#if DEBUG

    /// <summary>
    /// Self test astrometry methods.
    /// </summary>
    public static void TestAstrometry() {

        double utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tc, rh, wl, ob1, ob2, bm, em,
            dlim;
        double eo, rc, dc, pr, pd, px, rv, pmt, ri, di, aob, zob, hob, dob, rob;
        double ra1, dec1, pmr1, pmd1, px1, rv1, ep1a, ep1b, ep2a, ep2b;
        double ra2, dec2, pmr2, pmd2, px2, rv2;
        double refa, refb;
        double[] vec1, vec2, vec3, p, p1, q, e, sn, pob, pco;
        double[,] bpn, identityMatrix, pv;
        identityMatrix = new double[,] { { 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } };
        SofaAstrom astrom;
        SofaLDBody dbody1, dbody2, dbody3;      // light deflection bodies
        SofaLDBody[] b;
        int status;

        /* Astronomy/Astrometry */

        // DECLSPEC void iauAb(double pnat[3], double v[3], double s, double bm1,
        //       double ppr[3]);
        vec1 = new double[]
            { -0.76321968546737951, -0.60869453983060384, -0.21676408580639883 };
        vec2 = new double[]
            { 2.1044018893653786e-5, -8.9108923304429319e-5, -3.8633714797716569e-5 };
        vec3 = new double[3];
        Ab(vec1, vec2, 0.99980921395708788, 0.99999999506209258, vec3);
        SofaValidate.EqualsApprox(vec3,
            new double[] { -0.7631631094219556269, -0.6087553082505590832, -0.2167926269368471279 },
            1e-12);

        // DECLSPEC void iauApcg(double date1, double date2,
        //         double ebpv[2][3], double ehp[3],
        //         iauASTROM *astrom);
        // 2017-09-10 Edited V[] and Bml to match minor change to SOFA routine.
        bpn = new double[3, 3];
        astrom = new SofaAstrom();
        Apcg(2456165.5, 0.401182685, new double[,]{
            {0.901310875, -0.417402664, -0.180982288},
            {0.00742727954,  0.0140507459,  0.00609045792}},
            new double[] { 0.903358544, -0.415395237, -0.180084014 },
            astrom);
        SofaValidate.EqualsApprox(astrom.Pmt, 12.65133794027378508, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, 
            new double[] { 0.901310875, -0.417402664, -0.180982288 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[] 
        { 0.8940025429324143045, -0.4110930268679817955, -0.1782189004872870264 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 1.010465295811013146, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.4289638913597693554e-4, 0.8115034051581320575e-4, 0.3517555136380563427e-4 }, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999951686012981, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, identityMatrix, 0.0);


        // DECLSPEC void iauApcg13(double date1, double date2, iauASTROM *astrom);
        // 2017-09-10 Edited V[] and Bml to match minor change to SOFA routine.
        astrom = new SofaAstrom();
        Apcg13(2456165.5, 0.401182685, astrom);
        SofaValidate.EqualsApprox(astrom.Pmt, 12.65133794027378508, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, new double[]
        { 0.9013108747340644755, -0.4174026640406119957, -0.1809822877867817771 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[]
        {0.8940025429255499549, -0.4110930268331896318, -0.1782189006019749850}, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 1.010465295964664178, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.4289638912941341125e-4, 0.8115034032405042132e-4, 0.3517555135536470279e-4 }, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999951686013142, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, identityMatrix, 0.0);

        // DECLSPEC void iauApci(double date1, double date2,
        //         double ebpv[2][3], double ehp[3],
        //         double x, double y, double s,
        //         iauASTROM *astrom);
        astrom = new SofaAstrom();
        Apci(2456165.5, 0.401182685, new double[,]
        {{0.901310875, -0.417402664, -0.180982288}, {0.00742727954, 0.0140507459, 0.00609045792}},
        new double[] { 0.903358544, -0.415395237, -0.180084014 }, 0.0013122272, -2.92808623e-5,
        3.05749468e-8, astrom);
        SofaValidate.EqualsApprox(astrom.Pmt, 12.65133794027378508, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, new double[]
        { 0.901310875, -0.417402664, -0.180982288 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[] {
            0.8940025429324143045, -0.4110930268679817955, -0.1782189004872870264}, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 1.010465295811013146, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.4289638913597693554e-4, 0.8115034051581320575e-4, 0.3517555136380563427e-4 }, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999951686012981, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, new double[,]
            {{0.9999991390295159156, -0.1136336653771609630e-7, -0.1312227200895260194e-2},
            { 0.4978650072505016932e-7, 0.9999999995713154868, 0.2928082217872315680e-4},
            { 0.1312227200000000000e-2, -0.2928086230000000000e-4, 0.9999991386008323373}}, 1e-12);

        // DECLSPEC void iauApci13(double date1, double date2,
        //           iauASTROM *astrom, double *eo);
        // 2017-09-10 Edited V[] and Bml to match minor change to SOFA routine.
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        SofaValidate.EqualsApprox(astrom.Pmt, 12.65133794027378508, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, new double[]
            { 0.9013108747340644755, -0.4174026640406119957, -0.1809822877867817771 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[]
            { 0.8940025429255499549, -0.4110930268331896318, -0.1782189006019749850 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 1.010465295964664178, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.4289638912941341125e-4, 0.8115034032405042132e-4, 0.3517555135536470279e-4 }, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999951686013142, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, new double[,] {
            { 0.9999992060376761710, -0.1282291987222130690e-7,-0.1260128571661374559e-2},
            { 0.4124244860106037157e-7, 0.9999999997456835325,0.2255285422953395494e-4},
            { 0.1260128571051709670e-2, -0.2255288829420524935e-4,0.9999992057833604343}}, 1e-12);
        SofaValidate.EqualsApprox(eo, -0.2900618712657375647e-2, 1e-12);

        // DECLSPEC void iauApco(double date1, double date2,
        //         double ebpv[2][3], double ehp[3],
        //         double x, double y, double s, double theta,
        //         double elong, double phi, double hm,
        //         double xp, double yp, double sp,
        //         double refa, double refb,
        //         iauASTROM *astrom);
        astrom = new SofaAstrom();
        Apco(2456384.5, 0.970031644,
        new double[,]{  {-0.974170438, -0.211520082, -0.0917583024},
                        { 0.00364365824, -0.0154287319, -0.00668922024}},
        new double[] { -0.973458265, -0.209215307, -0.0906996477 },
        0.0013122272, -2.92808623e-5, 3.05749468e-8, 3.14540971,
        -0.527800806, -1.2345856, 2738.0,
        2.47230737e-7, 1.82640464e-6, -3.01974337e-11,
        0.000201418779, -2.36140831e-7, astrom);

        SofaValidate.EqualsApprox(astrom.Pmt, 13.25248468622587269, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb[0], -0.9741827110630322720, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eb[1], -0.2115130190135344832, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eb[2], -0.09179840186949532298, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[0], -0.9736425571689739035, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[1], -0.2092452125849330936, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[2], -0.09075578152243272599, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 0.9998233241709957653, 1e-12);
        SofaValidate.EqualsApprox(astrom.V[0], 0.2078704992916728762e-4, 1e-16);
        SofaValidate.EqualsApprox(astrom.V[1], -0.8955360107151952319e-4, 1e-16);
        SofaValidate.EqualsApprox(astrom.V[2], -0.3863338994288951082e-4, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999950277561236, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, new double[,]{
            {0.9999991390295159156, -0.1136336653771609630e-7, -0.1312227200895260194e-2},
            {0.4978650072505016932e-7, 0.9999999995713154868, 0.2928082217872315680e-4},
            {0.1312227200000000000e-2, -0.2928086230000000000e-4, 0.9999991386008323373}}, 1e-12);
        SofaValidate.EqualsApprox(astrom.Along, -0.5278008060301974337, 1e-12);
        SofaValidate.EqualsApprox(astrom.Xpl, 0.1133427418174939329e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Ypl, 0.1453347595745898629e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Sphi, -0.9440115679003211329, 1e-12);
        SofaValidate.EqualsApprox(astrom.Cphi, 0.3299123514971474711, 1e-12);
        SofaValidate.EqualsApprox(astrom.Diurab, 0, 0);
        SofaValidate.EqualsApprox(astrom.Eral, 2.617608903969802566, 1e-12);
        SofaValidate.EqualsApprox(astrom.Refa, 0.2014187790000000000e-3, 1e-15);
        SofaValidate.EqualsApprox(astrom.Refb, -0.2361408310000000000e-6, 1e-18);

        // DECLSPEC int iauApco13(double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          iauASTROM *astrom, double *eo);
        // 2017-09-10 edited Eb[], Eh[], Em, V[], Bm1 to match change to SOFA routine.
        astrom = new SofaAstrom();
        SofaValidate.EqualsApprox(Apco13(2456384.5, 0.969254051, 0.1550675, -0.527800806,
             -1.2345856, 2738.0, 2.47230737e-7, 1.82640464e-6, 731.0, 12.8, 0.59, 0.55, astrom,
             out eo), 0, 0);
        SofaValidate.EqualsApprox(astrom.Pmt, 13.25248468622475727, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb[0], -0.9741827107320875162, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eb[1], -0.2115130190489716682, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eb[2], -0.09179840189496755339, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[0], -0.9736425572586935247, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[1], -0.2092452121603336166, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh[2], -0.09075578153885665295, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 0.9998233240913898141, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.2078704994520489246e-4, -0.8955360133238868938e-4, -0.3863338993055887398e-4}, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999950277561004, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, new double[,] {
            {0.9999991390295147999, -0.1136336652812486604e-7, -0.001312227201745553566},
            { 0.4978650075315529277e-7, 0.9999999995713154865, 0.2928082218847679162e-4},
            {0.001312227200850293372, -0.2928086230975367296e-4, 0.9999991386008312212}}, 1e-12);
        SofaValidate.EqualsApprox(astrom.Along, -0.5278008060301974337, 1e-12);
        SofaValidate.EqualsApprox(astrom.Xpl, 0.1133427418174939329e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Ypl, 0.1453347595745898629e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Sphi, -0.9440115679003211329, 1e-12);
        SofaValidate.EqualsApprox(astrom.Cphi, 0.3299123514971474711, 1e-12);
        SofaValidate.EqualsApprox(astrom.Diurab, 0, 0);
        SofaValidate.EqualsApprox(astrom.Eral, 2.617608909189066140, 1e-12);
        SofaValidate.EqualsApprox(astrom.Refa, 0.2014187785940396921e-3, 1e-15);
        SofaValidate.EqualsApprox(astrom.Refb, -0.2361408314943696227e-6, 1e-18);
        SofaValidate.EqualsApprox(eo, -0.003020548354802412839, 1e-14);

        // DECLSPEC void iauApcs(double date1, double date2, double pv[2][3],
        //         double ebpv[2][3], double ehp[3],
        //         iauASTROM *astrom);
        astrom = new SofaAstrom();
        Apcs(2456384.5, 0.970031644,
            new double[,] {
            {-1836024.09, 1056607.72, -5998795.26},
            { -77.0361767, -133.310856, 0.0971855934}},
            new double[,] {
            {-0.974170438, -0.211520082, -0.0917583024},
            {0.00364365824, -0.0154287319, -0.00668922024}
            }, new double[] { -0.973458265, -0.209215307, -0.0906996477 }, astrom);
        SofaValidate.EqualsApprox(astrom.Pmt, 13.25248468622587269, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, new double[] {
            -0.9741827110629881886, -0.2115130190136415986, -0.09179840186954412099 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[] {
            -0.9736425571689454706, -0.2092452125850435930, -0.09075578152248299218 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 0.9998233241709796859, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.2078704993282685510e-4, -0.8955360106989405683e-4, -0.3863338994289409097e-4 },
            1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999950277561237, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, identityMatrix, 0.0);

        // DECLSPEC void iauApcs13(double date1, double date2, double pv[2][3],
        //           iauASTROM *astrom);
        astrom = new SofaAstrom();
        Apcs13(2456165.5, 0.401182685, new double[,] {
            {-6241497.16, 401346.896, -1251136.04}, {-29.264597, -455.021831, 0.0266151194}},
            astrom);
        SofaValidate.EqualsApprox(astrom.Pmt, 12.65133794027378508, 1e-11);
        SofaValidate.EqualsApprox(astrom.Eb, new double[]
            { 0.9012691529023298391, -0.4173999812023068781, -0.1809906511146821008 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eh, new double[]
            { 0.8939939101759726824, -0.4111053891734599955, -0.1782336880637689334 }, 1e-12);
        SofaValidate.EqualsApprox(astrom.Em, 1.010428384373318379, 1e-12);
        SofaValidate.EqualsApprox(astrom.V, new double[] {
            0.4279877294121697570e-4, 0.7963255087052120678e-4, 0.3517564013384691531e-4 }, 1e-16);
        SofaValidate.EqualsApprox(astrom.Bm1, 0.9999999952947980978, 1e-12);
        astrom.GetBpn(bpn);
        SofaValidate.EqualsApprox(bpn, identityMatrix, 0.0);

        // DECLSPEC void iauAper(double theta, iauASTROM *astrom);
        astrom = new SofaAstrom();
        astrom.Along = 1.234;
        Aper(5.678, astrom);
        SofaValidate.EqualsApprox(astrom.Eral, 6.912000000000000000, 1e-12);

        // DECLSPEC void iauAper13(double ut11, double ut12, iauASTROM *astrom);
        astrom = new SofaAstrom();
        astrom.Along = 1.234;
        Aper13(2456165.5, 0.401182685, astrom);
        SofaValidate.EqualsApprox(astrom.Eral, 3.316236661789694933, 1e-12);
        
        // DECLSPEC void iauApio(double sp, double theta,
        //         double elong, double phi, double hm, double xp, double yp,
        //         double refa, double refb,
        //         iauASTROM *astrom);
        astrom = new SofaAstrom();
        Apio(-3.01974337e-11, 3.14540971, -0.527800806, -1.2345856, 2738.0, 2.47230737e-7,
            1.82640464e-6, 0.000201418779, -2.36140831e-7, astrom);
        SofaValidate.EqualsApprox(astrom.Along, -0.5278008060301974337, 1e-12);
        SofaValidate.EqualsApprox(astrom.Xpl, 0.1133427418174939329e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Ypl, 0.1453347595745898629e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Sphi, -0.9440115679003211329, 1e-12);
        SofaValidate.EqualsApprox(astrom.Cphi, 0.3299123514971474711, 1e-12);
        SofaValidate.EqualsApprox(astrom.Diurab, 0.5135843661699913529e-6, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eral, 2.617608903969802566, 1e-12);
        SofaValidate.EqualsApprox(astrom.Refa, 0.2014187790000000000e-3, 1e-15);
        SofaValidate.EqualsApprox(astrom.Refb, -0.2361408310000000000e-6, 1e-18);

        // DECLSPEC int iauApio13(double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          iauASTROM *astrom);
        astrom = new SofaAstrom();
        SofaValidate.EqualsApprox(Apio13(2456384.5, 0.969254051, 0.1550675, -0.527800806,
            -1.2345856, 2738.0, 2.47230737e-7, 1.82640464e-6, 731.0, 12.8, 0.59, 0.55, astrom),
            0, 0);
        SofaValidate.EqualsApprox(astrom.Along, -0.5278008060301974337, 1e-12);
        SofaValidate.EqualsApprox(astrom.Xpl, 0.1133427418174939329e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Ypl, 0.1453347595745898629e-5, 1e-17);
        SofaValidate.EqualsApprox(astrom.Sphi, -0.9440115679003211329, 1e-12);
        SofaValidate.EqualsApprox(astrom.Cphi, 0.3299123514971474711, 1e-12);
        SofaValidate.EqualsApprox(astrom.Diurab, 0.5135843661699913529e-6, 1e-12);
        SofaValidate.EqualsApprox(astrom.Eral, 2.617608909189066140, 1e-12);
        SofaValidate.EqualsApprox(astrom.Refa, 0.2014187785940396921e-3, 1e-15);
        SofaValidate.EqualsApprox(astrom.Refb, -0.2361408314943696227e-6, 1e-18);

        // DECLSPEC void iauAtci13(double rc, double dc,
        //           double pr, double pd, double px, double rv,
        //           double date1, double date2,
        //           double *ri, double *di, double *eo);
        Atci13(2.71, 0.174, 1e-5, 5e-6, 0.1, 55.0, 2456165.5, 0.401182685, out ri, out di, out eo);
        SofaValidate.EqualsApprox(ri, 2.710121572969038991, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1729371367218230438, 1e-12);
        SofaValidate.EqualsApprox(eo, -0.002900618712657375647, 1e-14);


        // DECLSPEC void iauAtciq(double rc, double dc, double pr, double pd,
        //          double px, double rv, iauASTROM *astrom,
        //          double *ri, double *di);
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        Atciq(2.71, 0.174, 1e-5, 5e-6, 0.1, 55.0, astrom, out ri, out di);
        SofaValidate.EqualsApprox(ri, 2.710121572969038991, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1729371367218230438, 1e-12);

        // DECLSPEC void iauAtciqn(double rc, double dc, double pr, double pd,
        //           double px, double rv, iauASTROM *astrom,
        //           int n, iauLDBODY b[], double *ri, double *di);
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        dbody1 = new SofaLDBody(0.00028574, 3e-10, new double[,] {
            {-7.81014427, -5.60956681, -1.98079819},
            {0.0030723249, -0.00406995477, -0.00181335842}});
        dbody2=new SofaLDBody(0.00095435, 3e-9, new double[,] {
            {0.738098796, 4.63658692, 1.9693136},
            {-0.00755816922, 0.00126913722, 0.000727999001}
        });
        dbody3 = new SofaLDBody(1.0, 6e-6, new double[,] {
            {-0.000712174377, -0.00230478303, -0.00105865966},
            {6.29235213e-6, -3.30888387e-7, -2.96486623e-7}});
        b = new SofaLDBody[] {dbody1, dbody2, dbody3};
        Atciqn(2.71, 0.174, 1e-5, 5e-6, 0.1, 55.0, astrom, 3, b, out ri, out di);
        SofaValidate.EqualsApprox(ri, 2.710122008105325582, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1729371916491459122, 1e-12);

        
        // DECLSPEC void iauAtciqz(double rc, double dc, iauASTROM *astrom,
        //           double *ri, double *di);
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        Atciqz(2.71, 0.174, astrom, out ri, out di);
        SofaValidate.EqualsApprox(ri, 2.709994899247599271, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1728740720983623469, 1e-12);


        // DECLSPEC int iauAtco13(double rc, double dc,
        //          double pr, double pd, double px, double rv,
        //          double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          double *aob, double *zob, double *hob,
        //          double *dob, double *rob, double *eo);
        SofaValidate.EqualsApprox(Atco13(2.71, 0.174, 1e-5, 5e-6, 0.1, 55.0, 2456384.5, 0.969254051,
            0.1550675, -0.527800806, -1.2345856, 2738.0, 2.47230737e-7, 1.82640464e-6, 731.0, 12.8,
            0.59, 0.55, out aob, out zob, out hob, out dob, out rob, out eo), 0, 0);
        SofaValidate.EqualsApprox(aob, 0.09251774485358230653, 1e-12);
        SofaValidate.EqualsApprox(zob, 1.407661405256767021, 1e-12);
        SofaValidate.EqualsApprox(hob, -0.09265154431403157925, 1e-12);
        SofaValidate.EqualsApprox(dob, 0.1716626560075591655, 1e-12);
        SofaValidate.EqualsApprox(rob, 2.710260453503097719, 1e-12);
        SofaValidate.EqualsApprox(eo, -0.003020548354802412839, 1e-14);


        // DECLSPEC void iauAtic13(double ri, double di,
        //           double date1, double date2,
        //           double *rc, double *dc, double *eo);
        Atic13(2.710121572969038991, 0.1729371367218230438, 2456165.5, 0.401182685,
            out rc, out dc, out eo);
        SofaValidate.EqualsApprox(rc, 2.710126504531374930, 1e-12);
        SofaValidate.EqualsApprox(dc, 0.1740632537628342320, 1e-12);
        SofaValidate.EqualsApprox(eo, -0.002900618712657375647, 1e-14);

        // DECLSPEC void iauAticq(double ri, double di, iauASTROM *astrom,
        //          double *rc, double *dc);
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        Aticq(2.710121572969038991, 0.1729371367218230438, astrom, out rc, out dc);
        SofaValidate.EqualsApprox(rc, 2.710126504531374930, 1e-12);
        SofaValidate.EqualsApprox(dc, 0.1740632537628342320, 1e-12);


        // DECLSPEC void iauAticqn(double ri, double di, iauASTROM *astrom,
        //           int n, iauLDBODY b[], double *rc, double *dc);
        dbody1 = new SofaLDBody(0.00028574, 3e-10, new double[,] {
            {-7.81014427, -5.60956681, -1.98079819},
            { 0.0030723249, -0.00406995477, -0.00181335842}});
        dbody2 = new SofaLDBody(0.00095435, 3e-9, new double[,] {
            { 0.738098796, 4.63658692, 1.9693136},
            { -0.00755816922,  0.00126913722,  0.000727999001}});
        dbody3 = new SofaLDBody(1.0, 6e-6, new double[,]{
            {-0.000712174377, -0.00230478303, -0.00105865966},
            {6.29235213e-6, -3.30888387e-7, -2.96486623e-7}});
        b = new SofaLDBody[] { dbody1, dbody2, dbody3 };
        astrom = new SofaAstrom();
        Apci13(2456165.5, 0.401182685, astrom, out eo);
        Aticqn(2.709994899247599271, 0.1728740720983623469, astrom, 3, b, out rc, out dc);
        SofaValidate.EqualsApprox(rc, 2.709999575032685412, 1e-12);
        SofaValidate.EqualsApprox(dc, 0.1739999656317778034, 1e-12);

        // DECLSPEC int iauAtio13(double ri, double di,
        //          double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          double *aob, double *zob, double *hob,
        //          double *dob, double *rob);
        status = Atio13(2.710121572969038991, 0.1729371367218230438, 2456384.5, 0.969254051,
            0.1550675, -0.527800806, -1.2345856, 2738.0, 2.47230737e-7, 1.82640464e-6, 731.0, 12.8,
            0.59, 0.55, out aob, out zob, out hob, out dob, out rob);
        SofaValidate.EqualsApprox(status, 0, 0);
        SofaValidate.EqualsApprox(aob, 0.09233952224794989993, 1e-12);
        SofaValidate.EqualsApprox(zob, 1.407758704513722461, 1e-12);
        SofaValidate.EqualsApprox(hob, -0.09247619879782006106, 1e-12);
        SofaValidate.EqualsApprox(dob, 0.1717653435758265198, 1e-12);
        SofaValidate.EqualsApprox(rob, 2.710085107986886201, 1e-12);

        // DECLSPEC void iauAtioq(double ri, double di, iauASTROM *astrom,
        //          double *aob, double *zob,
        //          double *hob, double *dob, double *rob);
        astrom = new SofaAstrom();
        utc1 = 2456384.5;
        utc2 = 0.969254051;
        dut1 = 0.1550675;
        elong = -0.527800806;
        phi = -1.2345856;
        hm = 2738.0;
        xp = 2.47230737e-7;
        yp = 1.82640464e-6;
        phpa = 731.0;
        tc = 12.8;
        rh = 0.59;
        wl = 0.55;
        Apio13(utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tc, rh, wl, astrom);
        ri = 2.710121572969038991;
        di = 0.1729371367218230438;
        Atioq(ri, di, astrom, out aob, out zob, out hob, out dob, out rob);
        SofaValidate.EqualsApprox(aob, 0.09233952224794989993, 1e-12);
        SofaValidate.EqualsApprox(zob, 1.407758704513722461, 1e-12);
        SofaValidate.EqualsApprox(hob, -0.09247619879782006106, 1e-12);
        SofaValidate.EqualsApprox(rob, 2.710085107986886201, 1e-12);

        // DECLSPEC int iauAtoc13(const char *type, double ob1, double ob2,
        //          double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          double *rc, double *dc);
        utc1 = 2456384.5;
        utc2 = 0.969254051;
        dut1 = 0.1550675;
        elong = -0.527800806;
        phi = -1.2345856;
        hm = 2738.0;
        xp = 2.47230737e-7;
        yp = 1.82640464e-6;
        phpa = 731.0;
        tc = 12.8;
        rh = 0.59;
        wl = 0.55;

        ob1 = 2.710085107986886201;
        ob2 = 0.1717653435758265198;

        status = Atoc13("R", ob1, ob2, utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tc, rh, wl,
                   out rc, out dc);
        SofaValidate.EqualsApprox(rc, 2.709956744661000609, 1e-12);
        SofaValidate.EqualsApprox(dc, 0.1741696500895398562, 1e-12);
        SofaValidate.EqualsApprox(status, 0, 0);

        // DECLSPEC int iauAtoi13(const char *type, double ob1, double ob2,
        //          double utc1, double utc2, double dut1,
        //          double elong, double phi, double hm, double xp, double yp,
        //          double phpa, double tk, double rh, double wl,
        //          double *ri, double *di);
        utc1 = 2456384.5;
        utc2 = 0.969254051;
        dut1 = 0.1550675;
        elong = -0.527800806;
        phi = -1.2345856;
        hm = 2738.0;
        xp = 2.47230737e-7;
        yp = 1.82640464e-6;
        phpa = 731.0;
        tc = 12.8;
        rh = 0.59;
        wl = 0.55;
        ob1 = 2.710085107986886201;
        ob2 = 0.1717653435758265198;
        status = Atoi13("R", ob1, ob2, utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tc, rh, wl,
            out ri, out di);
        SofaValidate.EqualsApprox(ri, 2.710121574449135955, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1729371839114567725, 1e-12);
        SofaValidate.EqualsApprox(status, 0, 0);

        // DECLSPEC void iauAtoiq(const char *type,
        //          double ob1, double ob2, iauASTROM *astrom,
        //          double *ri, double *di);
        astrom = new SofaAstrom();
        utc1 = 2456384.5;
        utc2 = 0.969254051;
        dut1 = 0.1550675;
        elong = -0.527800806;
        phi = -1.2345856;
        hm = 2738.0;
        xp = 2.47230737e-7;
        yp = 1.82640464e-6;
        phpa = 731.0;
        tc = 12.8;
        rh = 0.59;
        wl = 0.55;
        Apio13(utc1, utc2, dut1, elong, phi, hm, xp, yp, phpa, tc, rh, wl, astrom);
        ob1 = 2.710085107986886201;
        ob2 = 0.1717653435758265198;
        Atoiq("R", ob1, ob2, astrom, out ri, out di);
        SofaValidate.EqualsApprox(ri, 2.710121574449135955, 1e-12);
        SofaValidate.EqualsApprox(di, 0.1729371839114567725, 1e-12);

        // DECLSPEC void iauLd(double bm, double p[3], double q[3], double e[3],
        //       double em, double dlim, double p1[3]);
        p = new double[3];
        p1 = new double[3];
        q = new double[3];
        e = new double[3];
        bm = 0.00028574;
        p[0] = -0.763276255;
        p[1] = -0.608633767;
        p[2] = -0.216735543;
        q[0] = -0.763276255;
        q[1] = -0.608633767;
        q[2] = -0.216735543;
        e[0] = 0.76700421;
        e[1] = 0.605629598;
        e[2] = 0.211937094;
        em = 8.91276983;
        dlim = 3e-10;
        Ld(bm, p, q, e, em, dlim, p1);
        SofaValidate.EqualsApprox(p1, new double[]
            {-0.7632762548968159627, -0.6086337670823762701, -0.2167355431320546947}, 1e-12);

        // DECLSPEC void iauLdn(int n, iauLDBODY b[], double ob[3], double sc[3],
        //        double sn[3]);
        dbody1 = new SofaLDBody(0.00028574, 3e-10, new double[,] {
            {-7.81014427, -5.60956681, -1.98079819},
            {0.0030723249, -0.00406995477, -0.00181335842}});
        dbody2 = new SofaLDBody(0.00095435, 3e-9, new double[,] {
            { 0.738098796, 4.63658692, 1.9693136},
            {-0.00755816922, 0.00126913722, 0.000727999001}});
        dbody3 = new SofaLDBody(1.0, 6e-6, new double[,] {
            {-0.000712174377, -0.00230478303, -0.00105865966},
            {6.29235213e-6, -3.30888387e-7, -2.96486623e-7}});
        b = new SofaLDBody[] { dbody1, dbody2, dbody3 };
        sn = new double[3];
        Ldn(b, new double[] { -0.974170437, -0.2115201, -0.0917583114 },
            new double[] { -0.763276255, -0.608633767, -0.216735543 }, sn);
        SofaValidate.EqualsApprox(sn, new double[]
            { -0.7632762579693333866, -0.6086337636093002660, -0.2167355420646328159 }, 1e-12);

        // DECLSPEC void iauLdsun(double p[3], double e[3], double em, double p1[3]);
        p1 = new double[3];
        Ldsun(new double[] { -0.763276255, -0.608633767, -0.216735543 },
            new double[] { -0.973644023, -0.20925523, -0.0907169552 }, 0.999809214, p1);
        SofaValidate.EqualsApprox(p1, new double[]
            { -0.7632762580731413169, -0.6086337635262647900, -0.2167355419322321302 }, 1e-12);

        // DECLSPEC void iauPmpx(double rc, double dc, double pr, double pd,
        //         double px, double rv, double pmt, double vob[3],
        //         double pco[3]);
        rc = 1.234;
        dc = 0.789;
        pr = 1e-5;
        pd = -2e-5;
        px = 1e-2;
        rv = 10.0;
        pmt = 8.75;
        pob = new double[]{ 0.9, 0.4, 0.1 };
        pco = new double[3];
        Pmpx(rc, dc, pr, pd, px, rv, pmt, pob, pco);
        SofaValidate.EqualsApprox(pco, new double[]
            { 0.2328137623960308440, 0.6651097085397855317, 0.7095257765896359847 }, 1e-12);

        // DECLSPEC int iauPmsafe(double ra1, double dec1, double pmr1, double pmd1,
        //          double px1, double rv1,
        //          double ep1a, double ep1b, double ep2a, double ep2b,
        //          double *ra2, double *dec2, double *pmr2, double *pmd2,
        //          double *px2, double *rv2);
        ra1 = 1.234;
        dec1 = 0.789;
        pmr1 = 1e-5;
        pmd1 = -2e-5;
        px1 = 1e-2;
        rv1 = 10.0;
        ep1a = 2400000.5;
        ep1b = 48348.5625;
        ep2a = 2400000.5;
        ep2b = 51544.5;
        status = Pmsafe(ra1, dec1, pmr1, pmd1, px1, rv1, ep1a, ep1b, ep2a, ep2b,
            out ra2, out dec2, out pmr2, out pmd2, out px2, out rv2);
        SofaValidate.EqualsApprox(ra2, 1.234087484501017061, 1e-12);
        SofaValidate.EqualsApprox(dec2, 0.7888249982450468574, 1e-12);
        SofaValidate.EqualsApprox(pmr2, 0.9996457663586073988e-5, 1e-16);
        SofaValidate.EqualsApprox(pmd2, -0.2000040085106737816e-4, 1e-16);
        SofaValidate.EqualsApprox(px2, 0.9999997295356765185e-2, 1e-12);
        SofaValidate.EqualsApprox(rv2, 10.38468380293920069, 1e-10);
        SofaValidate.EqualsApprox(status, 0, 0);

        // DECLSPEC void iauPvtob(double elong, double phi, double hm,
        //          double xp, double yp, double sp, double theta,
        //          double pv[2][3]);
        pv = new double[2, 3];
        Pvtob(2.0, 0.5, 3000.0, 1e-6, -0.5e-6, 1e-8, 5.0, pv);
        SofaValidate.EqualsApprox(pv[0, 0], 4225081.367071159207, 1e-5);
        SofaValidate.EqualsApprox(pv[0, 1], 3681943.215856198144, 1e-5);
        SofaValidate.EqualsApprox(pv[0, 2], 3041149.399241260785, 1e-5);
        SofaValidate.EqualsApprox(pv[1, 0], -268.4915389365998787, 1e-9);
        SofaValidate.EqualsApprox(pv[1, 1], 308.0977983288903123, 1e-9);
        SofaValidate.EqualsApprox(pv[1, 2], 0, 0);

        // DECLSPEC void iauRefco(double phpa, double tk, double rh, double wl,
        //          double *refa, double *refb);
        Refco(800.0, 10.0, 0.9, 0.4, out refa, out refb);
        SofaValidate.EqualsApprox(refa, 0.2264949956241415009e-3, 1e-15);
        SofaValidate.EqualsApprox(refb, -0.2598658261729343970e-6, 1e-18);
    }

#endif      // end self test code

}       // end class Sofa
}       // end namespace HirosePS
