﻿// Calendars.cs
// .NET adapters for the IAU SOFA Astronomy/Calendars methods.
// Copyright Paul S. Hirose 2017.

// revision history
#if false
2017-09-17 Originally part of another file, but spun off into its own file to make the methods
easier to locate during maintenance.
#endif

using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {
    /* Astronomy/Calendars */

    // DECLSPEC int iauCal2jd(int iy, int im, int id, double *djm0, double *djm);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauCal2jd(int iy, int im, int id, out double djm0, out double djm);

    /// <summary>
    /// Convert Gregorian Calendar date to Julian Date.
    /// </summary>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day</param>
    /// <param name="djm0">always 2400000.5</param>
    /// <param name="djm">Modified Julian Date at 0 h</param>
    /// <returns>zero if OK</returns>
    /// <remarks>rejects dates before -4799 January 1</remarks>
    public static int Cal2jd(int iy, int im, int id, out double djm0, out double djm) {
        return iauCal2jd(iy, im, id, out djm0, out djm);
    }


    // DECLSPEC double iauEpb(double dj1, double dj2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEpb(double dj1, double dj2);

    /// <summary>
    /// Convert Julian Date to Besselian Epoch.
    /// </summary>
    /// <param name="dj1">JD part 1 (2451545.0 for max resolution)</param>
    /// <param name="dj2">JD part 2</param>
    /// <returns>Besselian Epoch</returns>
    public static double Epb(double dj1, double dj2) {
        return iauEpb(dj1, dj2);
    }


    // DECLSPEC void iauEpb2jd(double epb, double *djm0, double *djm);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern void iauEpb2jd(double epb, out double djm0, out double djm);

    /// <summary>
    /// Convert Besselian Epoch to Julian Date.
    /// </summary>
    /// <param name="epb">Besselian Epoch></param>
    /// <param name="djm0">always 2400000.5</param>
    /// <param name="djm">Modified Julian Date</param>
    public static void Epb2jd(double epb, out double djm0, out double djm) {
        iauEpb2jd(epb, out djm0, out djm);
    }


    // DECLSPEC double iauEpj(double dj1, double dj2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEpj(double dj1, double dj2);

    /// <summary>
    /// Convert Julian Date to Julian Epoch.
    /// </summary>
    /// <param name="dj1">JD part 1 (2451545.0 for max resolution)</param>
    /// <param name="dj2">JD part 2</param>
    /// <returns>Julian Epoch (e.g. 1996.8)</returns>
    public static double Epj(double dj1, double dj2) {
        return iauEpj(dj1, dj2);
    }


    // DECLSPEC void iauEpj2jd(double epj, double *djm0, double *djm);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern void iauEpj2jd(double epj, out double djm0, out double djm);

    /// <summary>
    /// Convert Julian Epoch to Julian Date.
    /// </summary>
    /// <param name="epj">Julian Epoch (e.g. 1996.8)</param>
    /// <param name="djm0">always 2400000.5</param>
    /// <param name="djm">Modified Julian Date</param>
    public static void Epj2jd(double epj, out double djm0, out double djm) {
        iauEpj2jd(epj, out djm0, out djm);
    }


    // DECLSPEC int iauJd2cal(double dj1, double dj2, int *iy, int *im, int *id, double *fd);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauJd2cal(double dj1, double dj2, out int iy, out int im, out int id,
        out double fd);

    /// <summary>
    /// Convert Julian Date to Gregorian year, month, day, and fraction of a day.
    /// </summary>
    /// <param name="dj1">JD part 1</param>
    /// <param name="dj2">JD part 2</param>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day</param>
    /// <param name="fd">fractional day</param>
    /// <returns>zero if OK</returns>
    /// <remarks>The earliest valid date is -68569.5 (-4900 March 1). The largest value accepted
    /// is 1e9.</remarks>
    public static int Jd2cal(double dj1, double dj2,
    out int iy, out int im, out int id, out double fd) {
        return iauJd2cal(dj1, dj2, out iy, out im, out id, out fd);
    }


    // DECLSPEC int iauJdcalf(int ndp, double dj1, double dj2, int iymdf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauJdcalf(int ndp, double dj1, double dj2, int[] iymdf);

    /// <summary>
    /// Julian Date to Gregorian Calendar, rounded to a specified precision.
    /// </summary>
    /// <param name="ndp">requested number of decimal places of days in fraction</param>
    /// <param name="dj1">JD part 1</param>
    /// <param name="dj2">JD part 2</param>
    /// <param name="iymdf">year, month, day, fractional day</param>
    /// <returns>zero if OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="iymdf"/> length is not 4</exception>
    public static int Jdcalf(int ndp, double dj1, double dj2, int[] iymdf) {
        SofaValidate.VerifyInt1x4(iymdf);
        return iauJdcalf(ndp, dj1, dj2, iymdf);
    }



    // Test the methods in this file.
#if DEBUG
    /// <summary>
    /// Self test the SOFA calendar methods.
    /// </summary>
    public static void TestCalendars() {
        double jd1, jd2, tolerance, fracday;
        int status, y, m, d;
        int[] ymdf;

        // DECLSPEC int iauCal2jd(int iy, int im, int id, double *djm0, double *djm);
        status = Cal2jd(2003, 06, 01, out jd1, out jd2);
        if (status != 0)
            throw new Exception();
        tolerance = 1e-6;
        SofaValidate.EqualsApprox(jd1, 2400000.5, tolerance);
        SofaValidate.EqualsApprox(jd2, 52791.0, tolerance);

        // DECLSPEC double iauEpb(double dj1, double dj2);
        SofaValidate.EqualsApprox(Epb(2415019.8135, 30103.18648), 1982.418424159278580, 1e-12);

        // DECLSPEC void iauEpb2jd(double epb, double *djm0, double *djm);
        Epb2jd(1957.3, out jd1, out jd2);
        SofaValidate.EqualsApprox(jd1, 2400000.5, 1e-9);
        SofaValidate.EqualsApprox(jd2, 35948.1915101513, 1e-9);

        // DECLSPEC double iauEpj(double dj1, double dj2);
        SofaValidate.EqualsApprox(Epj(2451545, -7392.5), 1979.760438056125941, 1e-12);

        // DECLSPEC void iauEpj2jd(double epj, double *djm0, double *djm);
        Epj2jd(1996.8, out jd1, out jd2);
        SofaValidate.EqualsApprox(jd1, 2400000.5, 1e-9);
        SofaValidate.EqualsApprox(jd2, 50375.7, 1e-9);

        // DECLSPEC int iauJd2cal(double dj1, double dj2,
        //                     int *iy, int *im, int *id, double *fd);
        if (Jd2cal(2400000.5, 50123.9999, out y, out m, out d, out fracday) != 0)
            throw new Exception();
        SofaValidate.EqualsApprox(y, 1996, 0.0);
        SofaValidate.EqualsApprox(m, 2, 0.0);
        SofaValidate.EqualsApprox(d, 10, 0.0);
        SofaValidate.EqualsApprox(fracday, 0.9999, 1e-7);

        // DECLSPEC int iauJdcalf(int ndp, double dj1, double dj2, int iymdf[4]);
        ymdf = new int[4];
        SofaValidate.EqualsApprox(Jdcalf(4, 2400000.5, 50123.9999, ymdf), 0, 0);
        SofaValidate.EqualsApprox(ymdf[0], 1996, 0);
        SofaValidate.EqualsApprox(ymdf[1], 2, 0);
        SofaValidate.EqualsApprox(ymdf[2], 10, 0);
        SofaValidate.EqualsApprox(ymdf[3], 9999, 0);

    }   // end self test of class
#endif
}   // end partial class Sofa
}   // end namespace HirosePS
