﻿// CatEphem.cs
// .NET adapters for the IAU SOFA StarCatalogs and Ephemerides routines.
// Copyright 2017 by Paul S. Hirose.


// revision notes
#if false
2014-08-30  Began work.
2014-10-03  Ensured array parameters are checked for correct dimensions.
2014-10-06  XML comments have no parameter field blank and optimum JD format noted. Self tests
call my methods, not SOFA.

2017-09-10  Building SofaJpl 2.1. Starpm() test modified to match change in SOFA code.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {


    /* Astronomy/Ephemerides */


    //DECLSPEC int iauEpv00(double date1, double date2,
    //             double pvh[2][3], double pvb[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauEpv00(double date1, double date2, double[,] pvh, double[,] pvb);

    /// <summary>
    /// Earth position and velocity, heliocentric and barycentric, with respect to the Barycentric
    /// Celestial Reference System.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="pvh">heliocentric Earth position/velocity (au, au/day)</param>
    /// <param name="pvb">barycentric Earth position/velocity (au, au/day)</param>
    /// <returns>1 = warning: date outside the range 1900-2100 AD</returns>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    public static int Epv00(double date1, double date2, double[,] pvh, double[,] pvb) {
        SofaValidate.VerifyDouble2x3(pvh);
        SofaValidate.VerifyDouble2x3(pvb);
        return iauEpv00(date1, date2, pvh, pvb);
    }


    //DECLSPEC int iauPlan94(double date1, double date2, int np, double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauPlan94(double date1, double date2, int np, double[,] pv);

    /// <summary>
    /// Approximate heliocentric position and velocity of a nominated major planet:  Mercury, Venus,
    /// EMB, Mars, Jupiter, Saturn, Uranus or Neptune (but not the Earth itself).
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="np">planet (1=Mercury, 2=Venus, 3=EMB, 4=Mars, 5=Jupiter, 6=Saturn, 7=Uranus,
    /// 8=Neptune)</param>
    /// <param name="pv">planet p,v (heliocentric, J2000.0, AU,AU/d)</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="pv"/> does not have dimensions [2,3].
    /// </exception>
    public static int Plan94(double date1, double date2, int np, double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        return iauPlan94(date1, date2, np, pv);
    }


    /* Astronomy/SpaceMotion */


    // identical to iauPmsafe() already declared
#if false
    //DECLSPEC int iauPmsafe(double ra1, double dec1, double pmr1, double pmd1,
    //              double px1, double rv1, double ep1a, double ep1b,
    //              double ep2a, double ep2b, double *ra2, double *dec2,
    //              double *pmr2, double *pmd2, double *px2, double *rv2);
#endif


    // DECLSPEC int iauPvstar(double pv[2][3], double *ra, double *dec,
    //          double *pmr, double *pmd, double *px, double *rv);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauPvstar(double[,] pv, out double ra, out double dec,
        out double pmr, out double pmd, out double px, out double rv);

    /// <summary>
    /// Convert star position+velocity vector to catalog coordinates.
    /// </summary>
    /// <param name="pv">pv-vector (AU, AU/day)</param>
    /// <param name="ra">right ascension (radians)</param>
    /// <param name="dec">declination (radians)</param>
    /// <param name="pmr">RA proper motion (radians/year, coordinate rate not great circle rate)
    /// </param>
    /// <param name="pmd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcsec)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if array does not have dimensions [2,3].</exception>
    public static int Pvstar(double[,] pv, out double ra, out double dec,
        out double pmr, out double pmd, out double px, out double rv) {
        SofaValidate.VerifyDouble2x3(pv);
        return iauPvstar(pv, out ra, out dec, out pmr, out pmd, out px, out rv);
    }


    // DECLSPEC int iauStarpv(double ra, double dec,
    //          double pmr, double pmd, double px, double rv,
    //          double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauStarpv(double ra, double dec, double pmr, double pmd, double px, double rv,
        double[,] pv);

    /// <summary>
    /// Convert star catalog coordinates to position+velocity vector.
    /// </summary>
    /// <param name="ra">right ascension (radians)</param>
    /// <param name="dec">declination (radians)</param>
    /// <param name="pmr">RA proper motion (radians/year, coordinate rate not great circle rate)
    /// </param>
    /// <param name="pmd">Dec proper motion (radians/year)</param>
    /// <param name="px">parallax (arcseconds)</param>
    /// <param name="rv">radial velocity (km/s receding)</param>
    /// <param name="pv">pv-vector (AU, AU/day)</param>
    /// <returns>0 = no warnings</returns>
    /// <exception cref="ArgumentException">if array does not have dimensions [2,3].</exception>
    public static int Starpv(double ra, double dec, double pmr, double pmd, double px, double rv,
        double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        return iauStarpv(ra, dec, pmr, pmd, px, rv, pv);
    }

    
/* Astronomy/StarCatalogs */


    // DECLSPEC void iauFk52h(double r5, double d5,
    //              double dr5, double dd5, double px5, double rv5,
    //              double *rh, double *dh,
    //              double *drh, double *ddh, double *pxh, double *rvh);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauFk52h(double r5, double d5, double dr5, double dd5, double px5,
        double rv5, out double rh, out double dh, out double drh, out double ddh, out double pxh,
        out double rvh);

    /// <summary>
    /// Transform FK5 (J2000.0) star data into the Hipparcos system.
    /// </summary>
    /// <param name="r5">FK5 RA (radians)</param>
    /// <param name="d5">FK5 Dec (radians)</param>
    /// <param name="dr5">FK5 proper motion in RA (dRA/dt, rad/Jyear)</param>
    /// <param name="dd5">FK5 proper motion in Dec (dDec/dt, rad/Jyear)</param>
    /// <param name="px5">FK5 parallax (arcsec)</param>
    /// <param name="rv5">FK5 radial velocity (km/s receding)</param>
    /// <param name="rh">Hipparcos RA (radians)</param>
    /// <param name="dh">Hipparcos Dec (radians)</param>
    /// <param name="drh">Hipparcos proper motion in RA (dRA/dt, rad/Jyear)</param>
    /// <param name="ddh">Hipparcos proper motion in Dec (dDec/dt, rad/Jyear)</param>
    /// <param name="pxh">Hipparcos parallax (arcsec)</param>
    /// <param name="rvh">Hipparcos radial velocity (km/s, positive = receding)</param>
    /// <remarks>The proper motions in RA are dRA/dt rather than cos(Dec)*dRA/dt, and are per year
    /// rather than per century.</remarks>
    public static void Fk52h(double r5, double d5, double dr5, double dd5, double px5,
        double rv5, out double rh, out double dh, out double drh, out double ddh, out double pxh,
        out double rvh)
    {
         iauFk52h(r5, d5, dr5, dd5, px5, rv5, out rh, out dh, out drh, out ddh, out pxh, out rvh);
    }


    // DECLSPEC void iauFk5hip(double r5h[3][3], double s5h[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauFk5hip(double[,] r5h, double[] s5h);

    /// <summary>
    /// FK5 to Hipparcos rotation and spin.
    /// </summary>
    /// <param name="r5h">r-matrix: FK5 rotation wrt Hipparcos</param>
    /// <param name="s5h">r-vector: FK5 spin wrt Hipparcos</param>
    /// <exception cref="ArgumentException">if any array has wrong dimensions.</exception>
    public static void Fk5hip(double[,] r5h, double[] s5h) {
        SofaValidate.VerifyDouble3x3(r5h);
        SofaValidate.VerifyDouble1x3(s5h);
         iauFk5hip(r5h, s5h);
    }


    // DECLSPEC void iauFk5hz(double r5, double d5, double date1, double date2,
    //              double *rh, double *dh);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauFk5hz(double r5, double d5, double date1, double date2,
        out double rh, out double dh);

    /// <summary>
    /// Transform an FK5 (J2000.0) star position into the system of the Hipparcos catalogue,
    /// assuming zero Hipparcos proper motion.
    /// </summary>
    /// <param name="r5">FK5 RA (radians), equinox J2000.0, at date</param>
    /// <param name="d5">FK5 Dec (radians), equinox J2000.0, at date</param>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="rh">Hipparcos RA (radians)</param>
    /// <param name="dh">Hipparcos Dec (radians)</param>
    public static void Fk5hz(double r5, double d5, double date1, double date2,
        out double rh, out double dh)
    {
        iauFk5hz(r5, d5, date1, date2, out rh, out dh);
    }


    // DECLSPEC void iauH2fk5(double rh, double dh,
    //              double drh, double ddh, double pxh, double rvh,
    //              double *r5, double *d5,
    //              double *dr5, double *dd5, double *px5, double *rv5);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauH2fk5(double rh, double dh, double drh, double ddh, double pxh,
        double rvh, out double r5, out double d5, out double dr5, out double dd5, out double px5,
        out double rv5);

    /// <summary>
    /// Transform Hipparcos star data into the FK5 (J2000.0) system.
    /// </summary>
    /// <param name="rh">Hipparcos RA (radians)</param>
    /// <param name="dh">Hipparcos Dec (radians)</param>
    /// <param name="drh">Hipparcos proper motion in RA (dRA/dt, rad/Jyear)</param>
    /// <param name="ddh">Hipparcos proper motion in Dec (dDec/dt, rad/Jyear)</param>
    /// <param name="pxh">Hipparcos parallax (arcsec)</param>
    /// <param name="rvh">Hipparcos radial velocity (km/s receding)</param>
    /// <param name="r5">FK5 RA (radians)</param>
    /// <param name="d5">FK5 Dec (radians)</param>
    /// <param name="dr5">FK5 proper motion in RA (dRA/dt, rad/Jyear)</param>
    /// <param name="dd5">FK5 proper motion in Dec (dDec/dt, rad/Jyear)</param>
    /// <param name="px5">FK5 parallax (arcsec)</param>
    /// <param name="rv5">FK5 radial velocity (km/s receding)</param>
    public static void H2fk5(double rh, double dh, double drh, double ddh, double pxh,
        double rvh, out double r5, out double d5, out double dr5, out double dd5, out double px5,
        out double rv5)
    {
        iauH2fk5(rh, dh, drh, ddh, pxh, rvh, out r5, out d5, out dr5, out dd5, out px5, out rv5);
    }


    //DECLSPEC void iauHfk5z(double rh, double dh, double date1, double date2,
    //              double *r5, double *d5, double *dr5, double *dd5);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauHfk5z(double rh, double dh, double date1, double date2,
        out double r5, out double d5, out double dr5, out double dd5);

    /// <summary>
    /// Transform a Hipparcos star position into FK5 J2000.0, assuming zero Hipparcos proper motion.
    /// </summary>
    /// <param name="rh">Hipparcos RA (radians)</param>
    /// <param name="dh">Hipparcos Dec (radians)</param>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="r5">RA (radians)</param>
    /// <param name="d5">Dec (radians)</param>
    /// <param name="dr5">FK5 RA proper motion (rad/year)</param>
    /// <param name="dd5">Dec proper motion (rad/year)</param>
    /// <remarks>The proper motion in RA is dRA/dt, not cos(Dec)*dRA/dt.</remarks>
    public static void Hfk5z(double rh, double dh, double date1, double date2,
        out double r5, out double d5, out double dr5, out double dd5)
    {
        iauHfk5z(rh, dh, date1, date2, out r5, out d5, out dr5, out dd5);
    }


    //DECLSPEC int iauStarpm(double ra1, double dec1,
    //              double pmr1, double pmd1, double px1, double rv1,
    //              double ep1a, double ep1b, double ep2a, double ep2b,
    //              double *ra2, double *dec2,
    //              double *pmr2, double *pmd2, double *px2, double *rv2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauStarpm(double ra1, double dec1, double pmr1, double pmd1, double px1,
        double rv1, double ep1a, double ep1b, double ep2a, double ep2b,
        out double ra2, out double dec2, out double pmr2, out double pmd2, out double px2,
        out double rv2);

    /// <summary>
    /// Star proper motion: update star catalog data for space motion.
    /// </summary>
    /// <param name="ra1">right ascension (radians), before</param>
    /// <param name="dec1">declination (radians), before</param>
    /// <param name="pmr1">RA proper motion (radians/year), before</param>
    /// <param name="pmd1">Dec proper motion (radians/year), before</param>
    /// <param name="px1">parallax (arcseconds), before</param>
    /// <param name="rv1">radial velocity (km/s receding), before</param>
    /// <param name="ep1a">"before" TDB Julian date, part A (2451545.0 for optimum resolution)
    /// </param>
    /// <param name="ep1b">"before" TDB Julian date, part B</param>
    /// <param name="ep2a">"after" TDB Julian date, part A (2451545.0 for optimum resolution)
    /// </param>
    /// <param name="ep2b">"after" TDB Julian date, part B</param>
    /// <param name="ra2">right ascension (radians), after</param>
    /// <param name="dec2">declination (radians), after</param>
    /// <param name="pmr2">RA proper motion (radians/year), after</param>
    /// <param name="pmd2">Dec proper motion (radians/year), after</param>
    /// <param name="px2">parallax (arcseconds), after</param>
    /// <param name="rv2">radial velocity (km/s receding), after</param>
    /// <returns>0 = no warnings or errors</returns>
    /// <remarks>The RA proper motion is in terms of coordinate angle, not great circle angle.
    /// </remarks>
    public static int Starpm(double ra1, double dec1, double pmr1, double pmd1, double px1,
        double rv1, double ep1a, double ep1b, double ep2a, double ep2b,
        out double ra2, out double dec2, out double pmr2, out double pmd2, out double px2,
        out double rv2)
    {
        return iauStarpm(ra1, dec1, pmr1, pmd1, px1, rv1, ep1a, ep1b, ep2a, ep2b,
            out ra2, out dec2, out pmr2, out pmd2, out px2, out rv2);
    }



    // self test code
#if DEBUG

    /// <summary>
    /// Self test the star catalog and ephemeris methods.
    /// </summary>
    public static void TestCatEphem() {
        double r5, d5, dr5, dd5, px5, rv5, rh, dh, drh, ddh, pxh, rvh;
        double ra1, dec1, pmr1, pmd1, px1, rv1;
        double ra2, dec2, pmr2, pmd2, px2, rv2;
        double[] s5h;
        double[,] r5h, pvh, pvb, pv;
        int j, status;


        /* Astronomy/Ephemerides */


        //DECLSPEC int iauEpv00(double date1, double date2,
        //             double pvh[2][3], double pvb[2][3]);
        pvh = new double[2, 3];
        pvb = new double[2, 3];
        j = Epv00(2400000.5, 53411.52501161, pvh, pvb);
        SofaValidate.EqualsApprox(pvh[0, 0], -0.7757238809297706813, 1e-14);
        SofaValidate.EqualsApprox(pvh[0, 1], 0.5598052241363340596, 1e-14);
        SofaValidate.EqualsApprox(pvh[0, 2], 0.2426998466481686993, 1e-14);
        SofaValidate.EqualsApprox(pvh[1, 0], -0.1091891824147313846e-1, 1e-15);
        SofaValidate.EqualsApprox(pvh[1, 1], -0.1247187268440845008e-1, 1e-15);
        SofaValidate.EqualsApprox(pvh[1, 2], -0.5407569418065039061e-2, 1e-15);
        SofaValidate.EqualsApprox(pvb[0, 0], -0.7714104440491111971, 1e-14);
        SofaValidate.EqualsApprox(pvb[0, 1], 0.5598412061824171323, 1e-14);
        SofaValidate.EqualsApprox(pvb[0, 2], 0.2425996277722452400, 1e-14);
        SofaValidate.EqualsApprox(pvb[1, 0], -0.1091874268116823295e-1, 1e-15);
        SofaValidate.EqualsApprox(pvb[1, 1], -0.1246525461732861538e-1, 1e-15);
        SofaValidate.EqualsApprox(pvb[1, 2], -0.5404773180966231279e-2, 1e-15);

        //DECLSPEC int iauPlan94(double date1, double date2, int np, double pv[2][3]);
        pv = new double[2, 3];
        j = Plan94(2400000.5, 1e6, 0, pv);
        SofaValidate.EqualsApprox(pv, new double[,] { { 0, 0, 0 }, { 0, 0, 0 } }, 0.0);
        SofaValidate.EqualsApprox(j, -1, 0);
        j = Plan94(2400000.5, 1e6, 10, pv);
        SofaValidate.EqualsApprox(j, -1, 0);
        j = Plan94(2400000.5, -320000, 3, pv);
        SofaValidate.EqualsApprox(pv, new double[,]{
            {0.9308038666832975759, 0.3258319040261346000, 0.1422794544481140560},
            {-0.6429458958255170006e-2, 0.1468570657704237764e-1, 0.6406996426270981189e-2}},
            1e-11);
        SofaValidate.EqualsApprox(j, 1, 0);
        j = Plan94(2400000.5, 43999.9, 1, pv);
        SofaValidate.EqualsApprox(pv, new double[,] {
            {0.2945293959257430832, -0.2452204176601049596, -0.1615427700571978153},
            {0.1413867871404614441e-1, 0.1946548301104706582e-1, 0.8929809783898904786e-2}},
            1e-11);
        SofaValidate.EqualsApprox(j, 0, 0);


        /* Astronomy/SpaceMotion */


        // duplicate
        //DECLSPEC int iauPmsafe(double ra1, double dec1, double pmr1, double pmd1,
        //              double px1, double rv1, double ep1a, double ep1b,
        //              double ep2a, double ep2b, double *ra2, double *dec2,
        //              double *pmr2, double *pmd2, double *px2, double *rv2);


        // DECLSPEC int iauPvstar(double pv[2][3], double *ra, double *dec,
        //          double *pmr, double *pmd, double *px, double *rv);
        pv = new double[,] {
            {126668.5912743160601, 2136.792716839935195, -245251.2339876830091},
            {-0.4051854035740712739e-2, -0.6253919754866173866e-2, 0.1189353719774107189e-1}};
        status = Pvstar(pv, out ra1, out dec1, out pmr1, out pmd1, out px1, out rv1);
        SofaValidate.EqualsApprox(ra1, 0.1686756e-1, 1e-12);
        SofaValidate.EqualsApprox(dec1, -1.093989828, 1e-12);
        SofaValidate.EqualsApprox(pmr1, -0.178323516e-4, 1e-16);
        SofaValidate.EqualsApprox(pmd1, 0.2336024047e-5, 1e-16);
        SofaValidate.EqualsApprox(px1, 0.74723, 1e-12);
        SofaValidate.EqualsApprox(rv1, -21.60000010107306010, 1e-11);
        SofaValidate.EqualsApprox(status, 0, 0);

        // DECLSPEC int iauStarpv(double ra, double dec,
        //          double pmr, double pmd, double px, double rv,
        //          double pv[2][3]);
        ra1 = 0.01686756;
        dec1 = -1.093989828;
        pmr1 = -1.78323516e-5;
        pmd1 = 2.336024047e-6;
        px1 = 0.74723;
        rv1 = -21.6;
        status = Starpv(ra1, dec1, pmr1, pmd1, px1, rv1, pv);
        SofaValidate.EqualsApprox(pv[0, 0], 126668.5912743160601, 1e-10);
        SofaValidate.EqualsApprox(pv[0, 1], 2136.792716839935195, 1e-12);
        SofaValidate.EqualsApprox(pv[0, 2], -245251.2339876830091, 1e-10);
        SofaValidate.EqualsApprox(pv[1, 0], -0.4051854008955659551e-2, 1e-13);
        SofaValidate.EqualsApprox(pv[1, 1], -0.6253919754414777970e-2, 1e-15);
        SofaValidate.EqualsApprox(pv[1, 2], 0.1189353714588109341e-1, 1e-13);


        /* Astronomy/StarCatalogs */


        // DECLSPEC void iauFk52h(double r5, double d5,
        //              double dr5, double dd5, double px5, double rv5,
        //              double *rh, double *dh,
        //              double *drh, double *ddh, double *pxh, double *rvh);
        r5 = 1.76779433;
        d5 = -0.2917517103;
        dr5 = -1.91851572e-7;
        dd5 = -5.8468475e-6;
        px5 = 0.379210;
        rv5 = -7.6;
        Fk52h(r5, d5, dr5, dd5, px5, rv5, out rh, out dh, out drh, out ddh, out pxh, out rvh);
        SofaValidate.EqualsApprox(rh, 1.767794226299947632, 1e-14);
        SofaValidate.EqualsApprox(dh, -0.2917516070530391757, 1e-14);
        SofaValidate.EqualsApprox(drh, -0.19618741256057224e-6, 1e-19);
        SofaValidate.EqualsApprox(ddh, -0.58459905176693911e-5, 1e-19);
        SofaValidate.EqualsApprox(pxh, 0.37921, 1e-14);
        SofaValidate.EqualsApprox(rvh, -7.6000000940000254, 1e-11);

        // DECLSPEC void iauFk5hip(double r5h[3][3], double s5h[3]);
        s5h = new double[3];
        r5h = new double[3, 3];
        Fk5hip(r5h, s5h);
        SofaValidate.EqualsApprox(r5h[0,0], 0.9999999999999928638, 1e-14);
        SofaValidate.EqualsApprox(r5h[0,1], 0.1110223351022919694e-6, 1e-17);
        SofaValidate.EqualsApprox(r5h[0,2], 0.4411803962536558154e-7, 1e-17);
        SofaValidate.EqualsApprox(r5h[1,0], -0.1110223308458746430e-6, 1e-17);
        SofaValidate.EqualsApprox(r5h[1,1], 0.9999999999999891830, 1e-14);
        SofaValidate.EqualsApprox(r5h[1,2], -0.9647792498984142358e-7, 1e-17);
        SofaValidate.EqualsApprox(r5h[2,0], -0.4411805033656962252e-7, 1e-17);
        SofaValidate.EqualsApprox(r5h[2,1], 0.9647792009175314354e-7, 1e-17);
        SofaValidate.EqualsApprox(r5h[2,2], 0.9999999999999943728, 1e-14);
        SofaValidate.EqualsApprox(s5h, new double[] 
        { -0.1454441043328607981e-8, 0.2908882086657215962e-8, 0.3393695767766751955e-8 }, 1e-17);
        
        // DECLSPEC void iauFk5hz(double r5, double d5, double date1, double date2,
        //              double *rh, double *dh);
        r5 = 1.76779433;
        d5 = -0.2917517103;
        Fk5hz(r5, d5, 2400000.5, 54479.0, out rh, out dh);
        SofaValidate.EqualsApprox(rh, 1.767794191464423978, 1e-12);
        SofaValidate.EqualsApprox(dh, -0.2917516001679884419, 1e-12);

        // DECLSPEC void iauH2fk5(double rh, double dh,
        //              double drh, double ddh, double pxh, double rvh,
        //              double *r5, double *d5,
        //              double *dr5, double *dd5, double *px5, double *rv5);
        rh = 1.767794352;
        dh = -0.2917512594;
        drh = -2.76413026e-6;
        ddh = -5.92994449e-6;
        pxh = 0.379210;
        rvh = -7.6;
        H2fk5(rh, dh, drh, ddh, pxh, rvh, out r5, out d5, out dr5, out dd5, out px5, out rv5);
        SofaValidate.EqualsApprox(r5, 1.767794455700065506, 1e-13);
        SofaValidate.EqualsApprox(d5, -0.2917513626469638890, 1e-13);
        SofaValidate.EqualsApprox(dr5, -0.27597945024511204e-5, 1e-18);
        SofaValidate.EqualsApprox(dd5, -0.59308014093262838e-5, 1e-18);
        SofaValidate.EqualsApprox(px5, 0.37921, 1e-13);
        SofaValidate.EqualsApprox(rv5, -7.6000001309071126, 1e-10);

        //DECLSPEC void iauHfk5z(double rh, double dh, double date1, double date2,
        //              double *r5, double *d5, double *dr5, double *dd5);
        rh = 1.767794352;
        dh = -0.2917512594;
        Hfk5z(rh, dh, 2400000.5, 54479.0, out r5, out d5, out dr5, out dd5);
        SofaValidate.EqualsApprox(r5, 1.767794490535581026, 1e-13);
        SofaValidate.EqualsApprox(d5, -0.2917513695320114258, 1e-14);
        SofaValidate.EqualsApprox(dr5, 0.4335890983539243029e-8, 1e-22);
        SofaValidate.EqualsApprox(dd5, -0.8569648841237745902e-9, 1e-23);

        //DECLSPEC int iauStarpm(double ra1, double dec1,
        //              double pmr1, double pmd1, double px1, double rv1,
        //              double ep1a, double ep1b, double ep2a, double ep2b,
        //              double *ra2, double *dec2,
        //              double *pmr2, double *pmd2, double *px2, double *rv2);
        ra1 = 0.01686756;
        dec1 = -1.093989828;
        pmr1 = -1.78323516e-5;
        pmd1 = 2.336024047e-6;
        px1 = 0.74723;
        rv1 = -21.6;
        j = Starpm(ra1, dec1, pmr1, pmd1, px1, rv1, 2400000.5, 50083.0, 2400000.5, 53736.0,
            out ra2, out dec2, out pmr2, out pmd2, out px2, out rv2);
        SofaValidate.EqualsApprox(ra2, 0.01668919069414256149, 1e-13);
        SofaValidate.EqualsApprox(dec2, -1.093966454217127897, 1e-13);
        SofaValidate.EqualsApprox(pmr2, -0.1783662682153176524e-4, 1e-17);
        SofaValidate.EqualsApprox(pmd2, 0.2338092915983989595e-5, 1e-17);
        SofaValidate.EqualsApprox(px2, 0.7473533835317719243, 1e-13);
        SofaValidate.EqualsApprox(rv2, -21.59905170476417175, 1e-11);
        SofaValidate.EqualsApprox(j, 0, 0);
    }
#endif

}   // end partial class class Sofa
}   // end namespace HirosePS