﻿// Ecliptic.cs
// .NET adapters for the IAU SOFA Astronomy/EclipticCoordinates functions.
// Copyright 2017 Paul S. Hirose

// revision history
#if false
2017-09-14 Created as part of SofaJpl 2.1.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {
    // DECLSPEC void iauEceq06(double date1, double date2, double dl, double db,
    //           double *dr, double *dd);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauEceq06(double date1, double date2, double dl, double db,
        out double dr, out double dd);

    /// <summary>
    /// Transform ecliptic coordinates (mean equinox and ecliptic of date) to ICRS RA / Dec,
    /// using the IAU 2006 precession model.
    /// </summary>
    /// <param name="date1">Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">Julian date part 2</param>
    /// <param name="dl">ecliptic longitude (radians)</param>
    /// <param name="db">ecliptic latitude (radians)</param>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    public static void Eceq06(double date1, double date2, double dl, double db,
        out double dr, out double dd)
    {
        iauEceq06(date1, date2, dl, db, out dr, out dd);
    }


    // DECLSPEC void iauEcm06(double date1, double date2, double rm[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauEcm06(double date1, double date2, double[,] rm);

    /// <summary>
    /// ICRS equatorial to ecliptic and mean equinox rotation matrix, IAU 2006.
    /// </summary>
    /// <param name="date1">Terrestrial Time part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">Terrestrial Time part 2</param>
    /// <param name="rm">ICRS to ecliptic (mean equinox) rotation matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rm"/> does not have
    /// dimensions [3,3].</exception>
    public static void Ecm06(double date1, double date2, double[,] rm) {
        SofaValidate.VerifyDouble3x3(rm);
        iauEcm06(date1, date2, rm);
    }


    // DECLSPEC void iauEqec06(double date1, double date2, double dr, double dd,
    //           double *dl, double *db);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauEqec06(double date1, double date2, double dr, double dd,
        out double dl, out double db);

    /// <summary>
    /// Transform ICRS equatorial coordinates to ecliptic coordinates (mean equinox and ecliptic of
    /// date) using IAU 2006 precession model.
    /// </summary>
    /// <param name="date1">Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">Julian date part 2</param>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    /// <param name="dl">ecliptic longitude (radians)</param>
    /// <param name="db">ecliptic latitude (radians)</param>
    public static void Eqec06(double date1, double date2, double dr, double dd,
        out double dl, out double db)
    {
        iauEqec06(date1, date2, dr, dd, out dl, out db);
    }


    // DECLSPEC void iauLteceq(double epj, double dl, double db, double *dr, double *dd);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLteceq(double epj, double dl, double db, out double dr, out double dd);

    /// <summary>
    /// Transform ecliptic coordinates (mean equinox and ecliptic of date)
    /// to ICRS RA,Dec, using a long-term precession model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="dl">ecliptic longitude (radians)</param>
    /// <param name="db">ecliptic latitude (radians)</param>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    public static void Lteceq(double epj, double dl, double db, out double dr, out double dd) {
        iauLteceq(epj, dl, db, out dr, out dd);
    }


    // DECLSPEC void iauLtecm(double epj, double rm[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLtecm(double epj, double[,] rm);

    /// <summary>
    /// Rotation matrix, ICRS to ecliptic and mean equinox (long term model).
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="rm">ICRS to ecliptic rotation matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rm"/> does not have
    /// dimensions [3,3].</exception>
    public static void Ltecm(double epj, double[,] rm) {
        SofaValidate.VerifyDouble3x3(rm);
        iauLtecm(epj, rm);
    }


    // DECLSPEC void iauLteqec(double epj, double dr, double dd, double *dl, double *db);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLteqec(double epj, double dr, double dd, out double dl, out double db);

    /// <summary>
    /// Transform from ICRS to ecliptic coordinates (mean equinox and ecliptic of date) using a
    /// long-term precession model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    /// <param name="dl">ecliptic longitude (radians)</param>
    /// <param name="db">ecliptic latitude (radians)</param>
    public static void Lteqec(double epj, double dr, double dd, out double dl, out double db) {
        iauLteqec(epj, dr, dd, out dl, out db);
    }


    // self test code
#if DEBUG
    public static void TestEcliptic() {
        double date1, date2, db, dd, dl, dr, epj;
        double[,] rm;

        rm = new double[3, 3];


        // DECLSPEC void iauEceq06(double date1, double date2, double dl, double db,
        //           double *dr, double *dd);
        date1 = 2456165.5;
        date2 = 0.401182685;
        dl = 5.1;
        db = -0.9;

        Eceq06(date1, date2, dl, db, out dr, out dd);
        SofaValidate.EqualsApprox(dr, 5.533459733613627767, 1e-14);
        SofaValidate.EqualsApprox(dd, -1.246542932554480576, 1e-14);


        // DECLSPEC void iauEcm06(double date1, double date2, double rm[3][3]);
        date1 = 2456165.5;
        date2 = 0.401182685;

        Ecm06(date1, date2, rm);
        SofaValidate.EqualsApprox(rm, new double[,] {
            {0.9999952427708701137, -0.2829062057663042347e-2, -0.1229163741100017629e-2},
            {0.3084546876908653562e-2, 0.9174891871550392514, 0.3977487611849338124},
            {0.2488512951527405928e-5, -0.3977506604161195467, 0.9174935488232863071}},
            1e-14);


        // DECLSPEC void iauEqec06(double date1, double date2, double dr, double dd,
        //           double *dl, double *db);
        date1 = 1234.5;
        date2 = 2440000.5;
        dr = 1.234;
        dd = 0.987;

        Eqec06(date1, date2, dr, dd, out dl, out db);
        SofaValidate.EqualsApprox(dl, 1.342509918994654619, 1e-14);
        SofaValidate.EqualsApprox(db, 0.5926215259704608132, 1e-14);


        // DECLSPEC void iauLteceq(double epj, double dl, double db, double *dr, double *dd);
        epj = 2500.0;
        dl = 1.5;
        db = 0.6;
        Lteceq(epj, dl, db, out dr, out dd);
        SofaValidate.EqualsApprox(dr, 1.275156021861921167, 1e-14);
        SofaValidate.EqualsApprox(dd, 0.9966573543519204791, 1e-14);


        // DECLSPEC void iauLtecm(double epj, double rm[3][3]);
        epj = -3000.0;
        Ltecm(epj, rm);
        SofaValidate.EqualsApprox(rm, new double[,] {
            {0.3564105644859788825, 0.8530575738617682284, 0.3811355207795060435},
            {-0.9343283469640709942, 0.3247830597681745976, 0.1467872751535940865},
            {0.1431636191201167793e-2, -0.4084222566960599342, 0.9127919865189030899}
        }, 1e-14);


        // DECLSPEC void iauLteqec(double epj, double dr, double dd, double *dl, double *db);
        epj = -1500.0;
        dr = 1.234;
        dd = 0.987;
        Lteqec(epj, dr, dd, out dl, out db);
        SofaValidate.EqualsApprox(dl, 0.5039483649047114859, 1e-14);
        SofaValidate.EqualsApprox(db, 0.5848534459726224882, 1e-14);
    }
#endif  // end self test code
}   // end partial class Sofa
}   // end namespace HirosePS
