﻿// FundamentalArgs.cs
// .NET adapters for IAU SOFA routines in the FundamentalArgs category.
// Copyright Paul S. Hirose 2017

// revision history
#if false
2017-09-17 Created. Originally this was part of another file, but putting these methods in a
file of their own will make them easier to locate for maintenance.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {

    /* Astronomy/FundamentalArgs */

    // DECLSPEC double iauFad03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFad03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean elongation of the Moon from the Sun.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean elongation (radians)</returns>
    public static double Fad03(double t) {
        return iauFad03(t);
    }


    // DECLSPEC double iauFae03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFae03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Earth.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Earth (radians)</returns>
    public static double Fae03(double t) {
        return iauFae03(t);
    }


    // DECLSPEC double iauFaf03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFaf03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of the Moon minus mean
    /// longitude of the ascending node.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>F, radians</returns>
    public static double Faf03(double t) {
        return iauFaf03(t);
    }


    // DECLSPEC double iauFaju03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFaju03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Jupiter.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Jupiter, radians</returns>
    public static double Faju03(double t) {
        return iauFaju03(t);
    }


    // DECLSPEC double iauFal03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFal03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean anomaly of the Moon.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>l, radians</returns>
    public static double Fal03(double t) {
        return iauFal03(t);
    }


    // DECLSPEC double iauFalp03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFalp03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean anomaly of the Sun.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>l', radians</returns>
    public static double Falp03(double t) {
        return iauFalp03(t);
    }


    // DECLSPEC double iauFama03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFama03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Mars.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Mars, radians </returns>
    public static double Fama03(double t) {
        return iauFama03(t);
    }


    // DECLSPEC double iauFame03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFame03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Mercury.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Mercury, radians</returns>
    public static double Fame03(double t) {
        return iauFame03(t);
    }


    // DECLSPEC double iauFane03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFane03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Neptune.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Neptune, radians</returns>
    public static double Fane03(double t) {
        return iauFane03(t);
    }


    // DECLSPEC double iauFaom03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFaom03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of the Moon's ascending node.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>Omega, radians</returns>
    public static double Faom03(double t) {
        return iauFaom03(t);
    }


    // DECLSPEC double iauFapa03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFapa03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): general accumulated precession in longitude.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>general precession in longitude, radians</returns>
    public static double Fapa03(double t) {
        return iauFapa03(t);
    }


    // DECLSPEC double iauFasa03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFasa03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Saturn.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Saturn, radians</returns>
    public static double Fasa03(double t) {
        return iauFasa03(t);
    }


    // DECLSPEC double iauFaur03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFaur03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Uranus.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Uranus, radians</returns>
    public static double Faur03(double t) {
        return iauFaur03(t);
    }


    // DECLSPEC double iauFave03(double t);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauFave03(double t);

    /// <summary>
    /// Fundamental argument, IERS Conventions (2003): mean longitude of Venus.
    /// </summary>
    /// <param name="t">TDB, Julian centuries since J2000.0</param>
    /// <returns>mean longitude of Venus, radians</returns>
    public static double Fave03(double t) {
        return iauFave03(t);
    }


    // self test code
#if DEBUG
    public static void TestFundamentalArgs() {

        /* Astronomy/FundamentalArgs */

        // DECLSPEC double iauFad03(double t);
        SofaValidate.EqualsApprox(Fad03(0.80), 1.946709205396925672, 1e-12);

        // DECLSPEC double iauFae03(double t);
        SofaValidate.EqualsApprox(Fae03(0.80), 1.744713738913081846, 1e-12);

        // DECLSPEC double iauFaf03(double t);
        SofaValidate.EqualsApprox(Faf03(0.80), 0.2597711366745499518, 1e-12);

        // DECLSPEC double iauFaju03(double t);
        SofaValidate.EqualsApprox(Faju03(0.80), 5.275711665202481138, 1e-12);

        // DECLSPEC double iauFal03(double t);
        SofaValidate.EqualsApprox(Fal03(0.80), 5.132369751108684150, 1e-12);

        // DECLSPEC double iauFalp03(double t);
        SofaValidate.EqualsApprox(Falp03(0.80), 6.226797973505507345, 1e-12);

        // DECLSPEC double iauFama03(double t);
        SofaValidate.EqualsApprox(Fama03(0.80), 3.275506840277781492, 1e-12);

        // DECLSPEC double iauFame03(double t);
        SofaValidate.EqualsApprox(Fame03(0.80), 5.417338184297289661, 1e-12);

        // DECLSPEC double iauFane03(double t);
        SofaValidate.EqualsApprox(Fane03(0.80), 2.079343830860413523, 1e-12);

        // DECLSPEC double iauFaom03(double t);
        SofaValidate.EqualsApprox(Faom03(0.80), -5.973618440951302183, 1e-12);

        // DECLSPEC double iauFapa03(double t);
        SofaValidate.EqualsApprox(Fapa03(0.80), 0.1950884762240000000e-1, 1e-12);

        // DECLSPEC double iauFasa03(double t);
        SofaValidate.EqualsApprox(Fasa03(0.80), 5.371574539440827046, 1e-12);

        // DECLSPEC double iauFaur03(double t);
        SofaValidate.EqualsApprox(Faur03(0.80), 5.180636450180413523, 1e-12);

        // DECLSPEC double iauFave03(double t);
        SofaValidate.EqualsApprox(Fave03(0.80), 3.424900460533758000, 1e-12);

    }
#endif  // end self test code
}   // end partial class Sofa
}   // end namespace HirosePS
