﻿// Galactic.cs
// .NET adapters for the SOFA GalacticCoordinates routines.
// Copyright Paul S. Hirose, 2017

// revision notes
#if false

2015-08-20  Created.

2017-09-10  Building SofaJpl 2.1.
#endif



using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {

    // DECLSPEC void iauG2icrs ( double dl, double db, double *dr, double *dd );
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauG2icrs(double dl, double db, out double dr, out double dd);

    /// <summary>
    /// Transform galactic coordinates to the ICRS.
    /// </summary>
    /// <param name="dl">galactic longitude (radians)</param>
    /// <param name="db">galactic latitude (radians)</param>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    public static void G2icrs(double dl, double db, out double dr, out double dd) {
        iauG2icrs(dl, db, out dr, out dd);
    }


    // DECLSPEC void iauIcrs2g ( double dr, double dd, double *dl, double *db );
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauIcrs2g(double dr, double dd, out double dl, out double db);

    /// <summary>
    /// Transform ICRS coordinates to galactic.
    /// </summary>
    /// <param name="dr">ICRS right ascension (radians)</param>
    /// <param name="dd">ICRS declination (radians)</param>
    /// <param name="dl">galactic longitude (radians)</param>
    /// <param name="db">galactic latitude (radians)</param>
    public static void Icrs2g(double dr, double dd, out double dl, out double db) {
         iauIcrs2g(dr, dd, out dl, out db);
    }



    // Self-test
#if DEBUG

    /// <summary>
    /// Self test galactic methods.
    /// </summary>
    public static void TestGalactic() {
        double dl, db, dr, dd;

        dl = 5.5850536063818546461558105;
        db = -0.7853981633974483096156608;
        G2icrs(dl, db, out dr, out dd);
        SofaValidate.EqualsApprox(dr, 5.9338074302227188048671, 1e-14);
        SofaValidate.EqualsApprox(dd, -1.1784870613579944551541, 1e-14);

        dr = 5.9338074302227188048671087;
        dd = -1.1784870613579944551540570;
        Icrs2g(dr, dd, out dl, out db);
        SofaValidate.EqualsApprox(dl, 5.5850536063818546461558, 1e-14);
        SofaValidate.EqualsApprox(db, -0.7853981633974483096157, 1e-14);
    }

#endif
    // end self test code
}       // end partial class Sofa
}       // end namespace HirosePS