﻿// Geodetic.cs
// .NET adapters for IAU SOFA routines in the GeodeticGeocentric category.
// Copyright 2017 Paul S. Hirose.

// revision notes
#if false
2014-08-31  Began work.
2014-09-03  Complete, and self test passes.
2014-10-03  Ensured array parameters are checked for correct dimensions.
2014-10-06  No 2-part JDs, no XML comments missing for parameters. Self tests call my methods,
not SOFA.

2017-09-10 SofaJpl 2.1.
2017-09-17 Spun off the FundamentalArgs methods into their own file.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {

    /// <summary>
    /// Earth ellipsoid.
    /// </summary>
    public enum Ellipsoid {
        /// <summary>
        /// WGS84
        /// </summary>
        WGS84 = 1,
        /// <summary>
        /// GRS80
        /// </summary>
        GRS80,
        /// <summary>
        /// WGS72
        /// </summary>
        WGS72 };


    /* Astronomy/GeodeticGeocentric */


    // DECLSPEC int iauEform(int n, double *a, double *f);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauEform(int n, out double a, out double f);

    /// <summary>
    /// Earth reference ellipsoids.
    /// </summary>
    /// <param name="ellipsoid">enumeration to identify the ellipsoid</param>
    /// <param name="a">equatorial radius (meters)</param>
    /// <param name="f">flattening (ca. 0.00335)</param>
    /// <returns>0 = OK</returns>
    public static int Eform(Ellipsoid ellipsoid, out double a, out double f) {
        return iauEform((int) ellipsoid, out a, out f);
    }


    // DECLSPEC int iauGc2gd(int n, double xyz[3],
    //              double *elong, double *phi, double *height);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauGc2gd(int n, double[] xyz,
        out double elong, out double phi, out double height);

    /// <summary>
    /// Transform geocentric coordinates to geodetic using the specified reference ellipsoid.
    /// </summary>
    /// <param name="ellipsoid">ellipsoid identifier (WGS84, GRS80, or WGS72)</param>
    /// <param name="xyz">geocentric vector (meters)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="height">height above ellipsoid (meters)</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="xyz"/> is not dimensioned [3]
    /// </exception>
    public static int Gc2gd(Ellipsoid ellipsoid, double[] xyz,
        out double elong, out double phi, out double height)
    {
        SofaValidate.VerifyDouble1x3(xyz);
        return iauGc2gd((int) ellipsoid, xyz, out elong, out phi, out height);
    }

    
    // DECLSPEC int iauGc2gde(ouble a, double f, double xyz[3],
    //               double *elong, double *phi, double *height);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauGc2gde(double a, double f, double[] xyz,
        out double elong, out double phi, out double height);

    /// <summary>
    /// Transform geocentric coordinates to geodetic for a reference ellipsoid of specified form.
    /// </summary>
    /// <param name="a">equatorial radius</param>
    /// <param name="f">flattening (approx. 0.00335)</param>
    /// <param name="xyz">geocentric vector</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="height">height above ellipsoid</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="xyz"/> is not dimensioned [3]
    /// </exception>
    /// <remarks>The equatorial radius and the geocentric vector must use the same unit (e.g.,
    /// meters), which is the unit of the returned height.</remarks>
    public static int Gc2gde(double a, double f, double[] xyz,
        out double elong, out double phi, out double height)
    {
        SofaValidate.VerifyDouble1x3(xyz);
        return iauGc2gde(a, f, xyz, out elong, out phi, out height);
    }


    // DECLSPEC int iauGd2gc(int n, double elong, double phi, double height,
    //              double xyz[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauGd2gc(int n, double elong, double phi, double height, double[] xyz);

    /// <summary>
    /// Transform geodetic coordinates to geocentric using the specified reference ellipsoid.
    /// </summary>
    /// <param name="ellipsoid">ellipsoid identifier (WGS84, GRS80, or WGS72)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="height">height above ellipsoid (meters)</param>
    /// <param name="xyz">geocentric vector (meters)</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="xyz"/> is not dimensioned [3]
    /// </exception>
    public static int Gd2gc(Ellipsoid ellipsoid, double elong, double phi, double height,
        double[] xyz)
    {
        SofaValidate.VerifyDouble1x3(xyz);
        return iauGd2gc((int)ellipsoid, elong, phi, height, xyz);
    }


    // DECLSPEC int iauGd2gce(double a, double f,
    //               double elong, double phi, double height, double xyz[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauGd2gce(double a, double f, double elong, double phi, double height,
        double[] xyz);

    /// <summary>
    /// Transform geodetic coordinates to geocentric for a reference ellipsoid of specified form.
    /// </summary>
    /// <param name="a">equatorial radius</param>
    /// <param name="f">flattening (approx. 0.00335)</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="phi">geodetic latitude (radians)</param>
    /// <param name="height">height above ellipsoid</param>
    /// <param name="xyz">geocentric vector</param>
    /// <returns>0 = OK</returns>
    /// <exception cref="ArgumentException">if <paramref name="xyz"/> is not dimensioned [3]
    /// </exception>
    /// <remarks>Equatorial radius and the height above ellipsoid must be given in the same units
    /// (e.g., meters), which determines the units of the returned geocentric vector</remarks>
    public static int Gd2gce(double a, double f, double elong, double phi, double height,
        double[] xyz) 
    {
        SofaValidate.VerifyDouble1x3(xyz);
        return iauGd2gce(a, f, elong, phi, height, xyz);
    }


    // self test code
#if DEBUG

    /// <summary>
    /// Self test the geodetic methods.
    /// </summary>
    public static void TestGeodetic() {
        double a, f;
        double e, p, h;
        double[] xyz;
        int j;

        /* Astronomy/GeodeticGeocentric */

        // DECLSPEC int iauEform(int n, double *a, double *f);
        j = Eform(0, out a, out f);
        SofaValidate.EqualsApprox(j, -1, 0);
        j = Eform(Ellipsoid.WGS84, out a, out f);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(a, 6378137.0, 1e-10);
        SofaValidate.EqualsApprox(f, 0.0033528106647474807, 1e-18);
        j = Eform(Ellipsoid.GRS80, out a, out f);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(a, 6378137.0, 1e-10);
        SofaValidate.EqualsApprox(f, 0.0033528106811823189, 1e-18);
        j = Eform(Ellipsoid.WGS72, out a, out f);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(a, 6378135.0, 1e-10);
        SofaValidate.EqualsApprox(f, 0.0033527794541675049, 1e-18);
        // Test for ellipsoid 4 omitted since the illegal value is impossible with an enum.

        // DECLSPEC int iauGc2gd(int n, double xyz[3],
        //              double *elong, double *phi, double *height);
        xyz = new double[] {2e6, 3e6, 5.244e6};
        j = Gc2gd(0, xyz, out e, out p, out h);
        SofaValidate.EqualsApprox(j, -1, 0);
        j = Gc2gd(Ellipsoid.WGS84, xyz, out e, out p, out h);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(e, 0.98279372324732907, 1e-14);
        SofaValidate.EqualsApprox(p, 0.97160184819075459, 1e-14);
        SofaValidate.EqualsApprox(h, 331.41724614260599, 1e-8);
        j = Gc2gd(Ellipsoid.GRS80, xyz, out e, out p, out h);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(e, 0.98279372324732907, 1e-14);
        SofaValidate.EqualsApprox(p, 0.97160184820607853, 1e-14);
        SofaValidate.EqualsApprox(h, 331.41731754844348, 1e-8);
        j = Gc2gd(Ellipsoid.WGS72, xyz, out e, out p, out h);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(e, 0.98279372324732907, 1e-14);
        SofaValidate.EqualsApprox(p, 0.97160181811015119, 1e-14);
        SofaValidate.EqualsApprox(h, 333.27707261303181, 1e-8);

        // DECLSPEC int iauGc2gde(double a, double f, double xyz[3],
        //               double *elong, double *phi, double *height);
        a = 6378136.0;
        f = 0.0033528;
        xyz = new double[] {2e6, 3e6, 5.244e6};
        j = Gc2gde(a, f, xyz, out e, out p, out h);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(e, 0.98279372324732907, 1e-14);
        SofaValidate.EqualsApprox(p, 0.97160183775704115, 1e-14);
        SofaValidate.EqualsApprox(h, 332.36862495764397, 1e-8);

        // DECLSPEC int iauGd2gc(int n, double elong, double phi, double height,
        //              double xyz[3]);
        e = 3.1;
        p = -0.5;
        h = 2500.0;
        j = Gd2gc(0, e, p, h, xyz);
        SofaValidate.EqualsApprox(j, -1, 0);
        j = Gd2gc(Ellipsoid.WGS84, e, p, h, xyz);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(xyz,
            new double[] {-5599000.5577049947, 233011.67223479203, -3040909.4706983363}, 1e-7);
        j = Gd2gc(Ellipsoid.GRS80, e, p, h, xyz);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(xyz,
            new double[] {-5599000.5577260984, 233011.6722356703, -3040909.4706095476}, 1e-7);
        j = Gd2gc(Ellipsoid.WGS72, e, p, h, xyz);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(xyz,
            new double[] {-5598998.7626301490, 233011.5975297822, -3040908.6861467111}, 1e-7);

        // DECLSPEC int iauGd2gce(double a, double f,
        //               double elong, double phi, double height, double xyz[3]);
        a = 6378136.0;
        f = 0.0033528;
        e = 3.1;
        p = -0.5;
        h = 2500.0;
        j = Gd2gce(a, f, e, p, h, xyz);
        SofaValidate.EqualsApprox(j, 0, 0);
        SofaValidate.EqualsApprox(xyz,
            new double[] {-5598999.6665116328, 233011.63514630572, -3040909.0517314132}, 1e-7);

    }
#endif  // end self test code
}
}