﻿// Precession.cs
// .NET adapters for IAU SOFA routines in the PrecNutPolar category.
// Copyright 2017 Paul S. Hirose.

// revision notes
#if false
2014-09-03  Began work.
2014-09-20  Finished, and self test OK. Final desk check pending.
2014-10-03  Ensured all arrays are checked for correct dimensions.
2014-10-06  XML comments have no blank fields. Self tests call my routines, not SOFA.

2017-09-10 Building SofaJpl 2.1.
2017-09-12 Added array dimension check to C2ibpn(). Omission was probably an oversight.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {
    
    /* Astronomy/PrecNutPolar */

    // DECLSPEC void iauBi00(double *dpsibi, double *depsbi, double *dra);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauBi00(out double dpsibi, out double depsbi, out double dra);

    /// <summary>
    /// Frame bias components of IAU 2000 precession-nutation models (part of MHB2000 with
    /// additions).
    /// </summary>
    /// <param name="dpsibi">longitude correction (radians)</param>
    /// <param name="depsbi">obliquity correction (radians)</param>
    /// <param name="dra">ICRS RA of the J2000.0 mean equinox (radians)</param>
    public static void Bi00(out double dpsibi, out double depsbi, out double dra) {
         iauBi00(out dpsibi, out depsbi, out dra);
    }


    // DECLSPEC void iauBp00(double date1, double date2,
    //              double rb[3][3], double rp[3][3], double rbp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauBp00(double date1, double date2,
        double[,] rb, double[,] rp, double[,] rbp);

    /// <summary>
    /// Frame bias and precession, IAU 2000.
    /// </summary>
    /// <param name="date1">TT Julian Date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian Date part 2</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <exception cref="ArgumentException">if any array is not dimensioned [3,3]</exception>
    public static void Bp00(double date1, double date2, double[,] rb, double[,] rp, double[,] rbp)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        iauBp00(date1, date2, rb, rp, rbp);
    }


    // DECLSPEC void iauBp06(double date1, double date2,
    //              double rb[3][3], double rp[3][3], double rbp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauBp06(double date1, double date2,
        double[,] rb, double[,] rp, double[,] rbp);

    /// <summary>
    /// Frame bias and precession, IAU 2006.
    /// </summary>
    /// <param name="date1">TT Julian Date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian Date part 2</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <exception cref="ArgumentException">if any array is not dimensioned [3,3]</exception>
    public static void Bp06(double date1, double date2,
        double[,] rb, double[,] rp, double[,] rbp)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        iauBp06(date1, date2, rb, rp, rbp);
    }


    // DECLSPEC void iauBpn2xy(double rbpn[3][3], double *x, double *y);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauBpn2xy(double[,] rbpn, out double x, out double y);

    /// <summary>
    /// Extract from the bias-precession-nutation matrix the X,Y coordinates of the Celestial
    /// Intermediate Pole.
    /// </summary>
    /// <param name="rbpn">celestial-to-true matrix</param>
    /// <param name="x">x coordinate, Celestial Intermediate Pole unit vector</param>
    /// <param name="y">y coordinate, Celestial Intermediate Pole unit vector</param>
    /// <exception cref="ArgumentException">if <paramref name="rbpn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Bpn2xy(double[,] rbpn, out double x, out double y) {
        SofaValidate.VerifyDouble3x3(rbpn);
        iauBpn2xy(rbpn, out x, out y);
    }


    // DECLSPEC void iauC2i00a(double date1, double date2, double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2i00a(double date1, double date2, double[,] rc2i);

    /// <summary>
    /// Form the celestial to intermediate matrix for a given date using the IAU 2000A
    /// precession nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/>is not dimensioned [3,3]
    /// </exception>
    public static void C2i00a(double date1, double date2, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2i00a(date1, date2, rc2i);
    }


    // DECLSPEC void iauC2i00b(double date1, double date2, double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2i00b(double date1, double date2, double[,] rc2i);

    /// <summary>
    /// Form the celestial to intermediate matrix for a given date using the IAU 2000B
    /// precession nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/>is not dimensioned [3,3]
    /// </exception>
    public static void C2i00b(double date1, double date2, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2i00b(date1, date2, rc2i);
    }


    // DECLSPEC void iauC2i06a(double date1, double date2, double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2i06a(double date1, double date2, double[,] rc2i);

    /// <summary>
    /// Form the celestial to intermediate matrix for a given date using the IAU 2006 precession
    /// and IAU 2000A nutation models.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/>is not dimensioned [3,3]
    /// </exception>
    public static void C2i06a(double date1, double date2, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2i06a(date1, date2, rc2i);
    }


    // DECLSPEC void iauC2ibpn(double date1, double date2, double rbpn[3][3],
    //               double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2ibpn(double date1, double date2, double[,] rbpn, double[,] rc2i);

    /// <summary>
    /// Form the celestial-to-intermediate matrix for a given date given the bias precession
    /// nutation matrix. IAU 2000.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rbpn">celestial to true matrix</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/> or
    /// <paramref name="rbpn"/>is not dimensioned [3,3]
    /// </exception>
    /// <remarks>Although its name does not include "00", this method is in fact specific to the
    /// IAU 2000 models.</remarks>
    public static void C2ibpn(double date1, double date2, double[,] rbpn, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rbpn);
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2ibpn(date1, date2, rbpn, rc2i);
    }

    // DECLSPEC void iauC2ixy(double date1, double date2, double x, double y,
    //              double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2ixy(double date1, double date2, double x, double y, double[,] rc2i);

    /// <summary>
    /// Form the celestial to intermediate matrix for a given date when the CIP X,Y is known.
    /// IAU 2000.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">Celestial Intermediate Pole unit vector x</param>
    /// <param name="y">Celestial Intermediate Pole unit vector y</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/>not dimensioned [3,3]
    /// </exception>
    /// <remarks>This function is specific to the IAU 2000 models.</remarks>
    public static void C2ixy(double date1, double date2, double x, double y, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2ixy(date1, date2, x, y, rc2i);
    }


    // DECLSPEC void iauC2ixys(double x, double y, double s, double rc2i[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2ixys(double x, double y, double s, double[,] rc2i);

    /// <summary>
    /// Form the celestial to intermediate matrix given the CIP X,Y and the CIO locator s.
    /// </summary>
    /// <param name="x">Celestial Intermediate Pole unit vector x</param>
    /// <param name="y">Celestial Intermediate Pole unit vector y</param>
    /// <param name="s">CIO locator s</param>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2i"/>not dimensioned [3,3]
    /// </exception>
    public static void C2ixys(double x, double y, double s, double[,] rc2i) {
        SofaValidate.VerifyDouble3x3(rc2i);
        iauC2ixys(x, y, s, rc2i);
    }


    // DECLSPEC void iauC2t00a(double tta, double ttb, double uta, double utb,
    //               double xp, double yp, double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2t00a(double tta, double ttb, double uta, double utb, double xp,
        double yp, double[,] rc2t);

    /// <summary>
    /// Form the celestial to terrestrial matrix given TT, UT1, and the polar motion, using the
    /// IAU 2000A nutation model.
    /// </summary>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for max precision)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2t"/>not dimensioned [3,3]
    /// </exception>
    public static void C2t00a(double tta, double ttb, double uta, double utb, double xp,
        double yp, double[,] rc2t)
    {
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2t00a(tta, ttb, uta, utb, xp, yp, rc2t);
    }

    
    // DECLSPEC void iauC2t00b(double tta, double ttb, double uta, double utb,
    //               double xp, double yp, double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2t00b(double tta, double ttb, double uta, double utb, double xp,
        double yp, double[,] rc2t);

    /// <summary>
    /// Form the celestial to terrestrial matrix given TT, UT1, and the polar motion, using the
    /// IAU 2000B nutation model.
    /// </summary>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for max precision)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rc2t"/>not dimensioned [3,3]
    /// </exception>
    public static void C2t00b(double tta, double ttb, double uta, double utb, double xp,
        double yp, double[,] rc2t)
    {
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2t00b(tta, ttb, uta, utb, xp, yp, rc2t);
    }


    // DECLSPEC void iauC2t06a(double tta, double ttb, double uta, double utb,
    //               double xp, double yp, double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2t06a(double tta, double ttb, double uta, double utb,
        double xp, double yp, double[,] rc2t);

    /// <summary>
    /// Form the celestial to terrestrial matrix given the date, the UT1 and the polar motion,
    /// using the IAU 2006 precession and IAU 2000A nutation models.
    /// </summary>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for max precision)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rc2t"/> is not dimensioned [3,3]
    /// </exception>
    public static void C2t06a(double tta, double ttb, double uta, double utb,
        double xp, double yp, double[,] rc2t) 
    {
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2t06a(tta, ttb, uta, utb, xp, yp, rc2t);
    }


    // DECLSPEC void iauC2tcio(double rc2i[3][3], double era, double rpom[3][3],
    //               double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2tcio(double[,] rc2i, double era, double[,] rpom, double[,] rc2t);

    /// <summary>
    /// Assemble the celestial to terrestrial matrix from CIO-based components (the celestial to
    /// intermediate matrix, the Earth Rotation Angle and the polar motion matrix).
    /// </summary>
    /// <param name="rc2i">celestial to intermediate matrix</param>
    /// <param name="era">Earth rotation angle (radians)</param>
    /// <param name="rpom">polar motion matrix</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void C2tcio(double[,] rc2i, double era, double[,] rpom, double[,] rc2t) {
        SofaValidate.VerifyDouble3x3(rc2i);
        SofaValidate.VerifyDouble3x3(rpom);
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2tcio(rc2i, era, rpom, rc2t);
    }


    // DECLSPEC void iauC2teqx(double rbpn[3][3], double gst, double rpom[3][3],
    //               double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2teqx(double[,] rbpn, double gst, double[,] rpom, double[,] rc2t);

    /// <summary>
    /// Assemble the celestial to terrestrial matrix from equinox-based components (the celestial
    /// to true matrix, the Greenwich Apparent Sidereal Time and the polar motion matrix).
    /// </summary>
    /// <param name="rbpn">celestial to true matrix</param>
    /// <param name="gst">Greenwich (apparent) Sidereal Time (radians)</param>
    /// <param name="rpom">polar motion matrix</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void C2teqx(double[,] rbpn, double gst, double[,] rpom, double[,] rc2t) {
        SofaValidate.VerifyDouble3x3(rbpn);
        SofaValidate.VerifyDouble3x3(rpom);
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2teqx(rbpn, gst, rpom, rc2t);
    }

 
    // DECLSPEC void iauC2tpe(double tta, double ttb, double uta, double utb,
    //              double dpsi, double deps, double xp, double yp,
    //              double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2tpe(double tta, double ttb, double uta, double utb,
        double dpsi, double deps, double xp, double yp, double[,] rc2t);

    /// <summary>
    /// Form the celestial to terrestrial matrix given the date, the UT1, the nutation and the
    /// polar motion.  IAU 2000.
    /// </summary>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for max precision)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rc2t"/> is not dimensioned [3,3]
    /// </exception>
    /// <remarks>Specific to the IAU 2000 models.</remarks>
    public static void C2tpe(double tta, double ttb, double uta, double utb,
        double dpsi, double deps, double xp, double yp, double[,] rc2t)
    {
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2tpe(tta, ttb, uta, utb, dpsi, deps, xp, yp, rc2t);
    }

    // DECLSPEC void iauC2txy(double tta, double ttb, double uta, double utb,
    //              double x, double y, double xp, double yp,
    //              double rc2t[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2txy(double tta, double ttb, double uta, double utb,
        double x, double y, double xp, double yp, double[,] rc2t);

    /// <summary>
    /// Form the celestial to terrestrial matrix given the date, the UT1, the CIP coordinates and
    /// the polar motion.  IAU 2000.
    /// </summary>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for max precision)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="x">CIP unit vector x coordinate in the GCRS</param>
    /// <param name="y">CIP unit vector y coordinate in the GCRS</param>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="rc2t">celestial to terrestrial matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rc2t"/> is not dimensioned [3,3]
    /// </exception>
    /// <remarks>This function is specific to the IAU 2000 models.</remarks>
    public static void C2txy(double tta, double ttb, double uta, double utb,
        double x, double y, double xp, double yp, double[,] rc2t)
    {
        SofaValidate.VerifyDouble3x3(rc2t);
        iauC2txy(tta, ttb, uta, utb, x, y, xp, yp, rc2t);
    }


    // DECLSPEC double iauEo06a(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauEo06a(double date1, double date2);

    /// <summary>
    /// Equation of the origins, IAU 2006 precession and IAU 2000A nutation.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>equation of the origins in radians</returns>
    public static double Eo06a(double date1, double date2) {
        return iauEo06a(date1, date2);
    }


    // DECLSPEC double iauEors(double rnpb[3][3], double s);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauEors(double[,] rnpb, double s);

    /// <summary>
    /// Equation of the origins, given the classical NPB matrix and the quantity s.
    /// </summary>
    /// <param name="rnpb">classical nutation x precession x bias matrix</param>
    /// <param name="s">the quantity s (the CIO locator)</param>
    /// <returns>equation of the origins in radians</returns>
    /// <exception cref="ArgumentException"> if <paramref name="rnpb"/> is not dimensioned [3,3]
    /// </exception>
    public static double Eors(double[,] rnpb, double s) {
        SofaValidate.VerifyDouble3x3(rnpb);
        return iauEors(rnpb, s);
    }


    // DECLSPEC void iauFw2m(double gamb, double phib, double psi, double eps,
    //             double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauFw2m(double gamb, double phib, double psi, double eps, double[,] r);

    /// <summary>
    /// Form rotation matrix given the Fukushima-Williams angles.
    /// </summary>
    /// <param name="gamb">F-W angle gamma_bar (radians)</param>
    /// <param name="phib">F-W angle phi_bar (radians)</param>
    /// <param name="psi">F-W angle psi (radians)</param>
    /// <param name="eps">F-W angle epsilon (radians)</param>
    /// <param name="r">rotation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="r"/> is not dimensioned [3,3]
    /// </exception>
    public static void Fw2m(double gamb, double phib, double psi, double eps, double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauFw2m(gamb, phib, psi, eps, r);
    }


    // DECLSPEC void iauFw2xy(double gamb, double phib, double psi, double eps,
    //              double *x, double *y);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauFw2xy(double gamb, double phib, double psi, double eps,
        out double x, out double y);

    /// <summary>
    /// CIP X,Y given Fukushima-Williams bias-precession-nutation angles.
    /// </summary>
    /// <param name="gamb">F-W angle gamma_bar (radians)</param>
    /// <param name="phib">F-W angle phi_bar (radians)</param>
    /// <param name="psi">F-W angle psi (radians)</param>
    /// <param name="eps">F-W angle epsilon (radians)</param>
    /// <param name="x">CIP unit vector X</param>
    /// <param name="y">CIP unit vector Y</param>
    public static void Fw2xy(double gamb, double phib, double psi, double eps,
        out double x, out double y)
    {
        iauFw2xy(gamb, phib, psi, eps, out x, out y);
    }

    
    // DECLSPEC void iauLtp(double epj, double rp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLtp(double epj, double[,] rp);

    /// <summary>
    /// Precession matrix, J2000 system to mean equator and equinox of date, with a long-term model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="rp">precession matrix, J2000.0 to mean equator and equinox of date</param>
    /// <exception cref="ArgumentException">if <paramref name="rp"/> does not have
    /// dimensions [3,3]</exception>
    public static void Ltp(double epj, double[,] rp) {
        SofaValidate.VerifyDouble3x3(rp);
        iauLtp(epj, rp);
    }


    // DECLSPEC void iauLtpb(double epj, double rpb[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLtpb(double epj, double[,] rpb);

    /// <summary>
    /// Precession matrix, ICRS to mean equator and equinox of date, with a long term model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="rpb">precession-bias matrix, ICRS to mean of date</param>
    /// <exception cref="ArgumentException">if <paramref name="rpb"/> does not have
    /// dimensions [3,3].</exception>
    public static void Ltpb(double epj, double[,] rpb) {
        SofaValidate.VerifyDouble3x3(rpb);
        iauLtpb(epj, rpb);
    }


    // DECLSPEC void iauLtpecl(double epj, double vec[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLtpecl(double epj, double[] vec);

    /// <summary>
    /// Pole of the ecliptic, from a long term model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="vec">unit vector to the ecliptic pole in the J2000.0 system</param>
    /// <exception cref="ArgumentException">if <paramref name="vec"/> does not have dimension [3].
    /// </exception>
    public static void Ltpecl(double epj, double[] vec) {
        SofaValidate.VerifyDouble1x3(vec);
        iauLtpecl(epj, vec);
    }


    // DECLSPEC void iauLtpequ(double epj, double veq[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauLtpequ(double epj, double[] veq);

    /// <summary>
    /// Pole of the mean equator, from a long term model.
    /// </summary>
    /// <param name="epj">Julian epoch (TT)</param>
    /// <param name="veq">unit vector to the pole of the mean equator in the J2000.0 system</param>
    /// <exception cref="ArgumentException">if <paramref name="veq"/> does not have
    /// dimension [3].</exception>
    public static void Ltpequ(double epj, double[] veq) {
        SofaValidate.VerifyDouble1x3(veq);
        iauLtpequ(epj, veq);
    }


    // DECLSPEC void iauNum00a(double date1, double date2, double rmatn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNum00a(double date1, double date2, double[,] rmatn);

    /// <summary>
    /// Form the matrix of nutation for a given date, IAU 2000A model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rmatn">nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatn"/> not dimensioned [3,3]
    /// </exception>
    public static void Num00a(double date1, double date2, double[,] rmatn) {
        SofaValidate.VerifyDouble3x3(rmatn);
        iauNum00a(date1, date2, rmatn);
    }


    // DECLSPEC void iauNum00b(double date1, double date2, double rmatn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNum00b(double date1, double date2, double[,] rmatn);

    /// <summary>
    /// Form the matrix of nutation for a given date, IAU 2000B model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rmatn">nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatn"/> not dimensioned [3,3]
    /// </exception>
    public static void Num00b(double date1, double date2, double[,] rmatn) {
        SofaValidate.VerifyDouble3x3(rmatn);
        iauNum00b(date1, date2, rmatn);
    }


    // DECLSPEC void iauNum06a(double date1, double date2, double rmatn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNum06a(double date1, double date2, double[,] rmatn);

    /// <summary>
    /// Form the matrix of nutation for a given date, IAU 2006/2000A model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rmatn">nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Num06a(double date1, double date2, double[,] rmatn) {
        SofaValidate.VerifyDouble3x3(rmatn);
        iauNum06a(date1, date2, rmatn);
    }


    // DECLSPEC void iauNumat(double epsa, double dpsi, double deps, double rmatn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNumat(double epsa, double dpsi, double deps, double[,] rmatn);

    /// <summary>
    /// Form the matrix of nutation.
    /// </summary>
    /// <param name="epsa">mean obliquity of date</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    /// <param name="rmatn">nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Numat(double epsa, double dpsi, double deps, double[,] rmatn) {
        SofaValidate.VerifyDouble3x3(rmatn);
        iauNumat(epsa, dpsi, deps, rmatn);
    }


    // DECLSPEC void iauNut00a(double date1, double date2, double *dpsi, double *deps);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNut00a(double date1, double date2, out double dpsi, out double deps);

    /// <summary>
    /// Nutation, IAU 2000A model (MHB2000 luni-solar and planetary nutation with free core nutation
    /// omitted).
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    public static void Nut00a(double date1, double date2, out double dpsi, out double deps) {
        iauNut00a(date1, date2, out dpsi, out deps);
    }


    // DECLSPEC void iauNut00b(double date1, double date2, double *dpsi, double *deps);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNut00b(double date1, double date2, out double dpsi, out double deps);

    /// <summary>
    /// Nutation, IAU 2000B model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    public static void Nut00b(double date1, double date2, out double dpsi, out double deps) {
        iauNut00b(date1, date2, out dpsi, out deps);
    }


    // DECLSPEC void iauNut06a(double date1, double date2, double *dpsi, double *deps);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNut06a(double date1, double date2, out double dpsi, out double deps);

    /// <summary>
    /// IAU 2000A nutation with adjustments to match the IAU 2006 precession.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    public static void Nut06a(double date1, double date2, out double dpsi, out double deps) {
        iauNut06a(date1, date2, out dpsi, out deps);
    }


    // DECLSPEC void iauNut80(double date1, double date2, double *dpsi, double *deps);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNut80(double date1, double date2, out double dpsi, out double deps);

    /// <summary>
    /// Nutation, IAU 1980 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    public static void Nut80(double date1, double date2, out double dpsi, out double deps) {
        iauNut80(date1, date2, out dpsi, out deps);
    }


    // DECLSPEC void iauNutm80(double date1, double date2, double rmatn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauNutm80(double date1, double date2, double[,] rmatn);

    /// <summary>
    /// Form the matrix of nutation for a given date, IAU 1980 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rmatn">nutation matrix</param>
    /// <exception cref="ArgumentException">if <paramref name="rmatn"/> is not dimensioned [3,3].
    /// </exception>
    public static void Nutm80(double date1, double date2, double[,] rmatn) {
        SofaValidate.VerifyDouble3x3(rmatn);
        iauNutm80(date1, date2, rmatn);
    }


    // DECLSPEC double iauObl06(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauObl06(double date1, double date2);

    /// <summary>
    /// Mean obliquity of the ecliptic, IAU 2006 precession model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>mean obliquity of the ecliptic (radians)</returns>
    public static double Obl06(double date1, double date2) {
        return iauObl06(date1, date2);
    }


    // DECLSPEC double iauObl80(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauObl80(double date1, double date2);

    /// <summary>
    /// Mean obliquity of the ecliptic, IAU 1980 precession model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>mean obliquity of the ecliptic (radians)</returns>
    public static double Obl80(double date1, double date2) {
        return iauObl80(date1, date2);
    }


    // DECLSPEC void iauP06e(double date1, double date2,
    //             double *eps0, double *psia, double *oma, double *bpa,
    //             double *bqa, double *pia, double *bpia,
    //             double *epsa, double *chia, double *za, double *zetaa,
    //             double *thetaa, double *pa,
    //             double *gam, double *phi, double *psi);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauP06e(double date1, double date2, out double eps0, out double psia,
        out double oma, out double bpa, out double bqa, out double pia, out double bpia,
        out double epsa, out double chia, out double za, out double zetaa, out double thetaa,
        out double pa, out double gam, out double phi, out double psi);

    /// <summary>
    /// Precession angles, IAU 2006, equinox based.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="eps0">epsilon_0 (all angles are radians)</param>
    /// <param name="psia">psi_A</param>
    /// <param name="oma">omega_A</param>
    /// <param name="bpa">P_A</param>
    /// <param name="bqa">Q_A</param>
    /// <param name="pia">pi_A</param>
    /// <param name="bpia">Pi_A</param>
    /// <param name="epsa">obliquity epsilon_A</param>
    /// <param name="chia">chi_A</param>
    /// <param name="za">z_A</param>
    /// <param name="zetaa">zeta_A</param>
    /// <param name="thetaa">theta_A</param>
    /// <param name="pa">p_A</param>
    /// <param name="gam">F-W angle gamma_J2000</param>
    /// <param name="phi">F-W angle phi_J2000</param>
    /// <param name="psi">F-W angle psi_J2000</param>
    public static void P06e(double date1, double date2, out double eps0, out double psia,
        out double oma, out double bpa, out double bqa, out double pia, out double bpia,
        out double epsa, out double chia, out double za, out double zetaa, out double thetaa,
        out double pa, out double gam, out double phi, out double psi)
    {
        iauP06e(date1, date2, out eps0, out psia, out oma, out bpa, out bqa, out pia, out bpia,
            out epsa, out chia, out za, out zetaa, out thetaa, out pa, out gam, out phi, out psi);
    }


    // DECLSPEC void iauPb06(double date1, double date2,
    //             double *bzeta, double *bz, double *btheta);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPb06(double date1, double date2,
        out double bzeta, out double bz, out double btheta);

    /// <summary>
    /// Form the three Euler angles which implement general precession from epoch J2000.0, using the
    /// IAU 2006 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="bzeta">1st rotation: radians cw around z</param>
    /// <param name="bz">3rd rotation: radians cw around z</param>
    /// <param name="btheta">2nd rotation: radians ccw around y</param>
    /// <remarks>Frame bias (the offset between ICRS and mean J2000.0) is included.</remarks>
    public static void Pb06(double date1, double date2,
        out double bzeta, out double bz, out double btheta)
    {
        iauPb06(date1, date2, out bzeta, out bz, out btheta);
    }


    // DECLSPEC void iauPfw06(double date1, double date2,
    //              double *gamb, double *phib, double *psib, double *epsa);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPfw06(double date1, double date2,
        out double gamb, out double phib, out double psib, out double epsa);

    /// <summary>
    /// Precession angles, IAU 2006 (Fukushima-Williams 4-angle formulation).
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="gamb">F-W angle gamma_bar (radians)</param>
    /// <param name="phib">F-W angle phi_bar (radians)</param>
    /// <param name="psib">F-W angle psi_bar (radians)</param>
    /// <param name="epsa">F-W angle epsilon_A (radians)</param>
    public static void Pfw06(double date1, double date2,
        out double gamb, out double phib, out double psib, out double epsa)
    {
        iauPfw06(date1, date2, out gamb, out phib, out psib, out epsa);
    }


    // DECLSPEC void iauPmat00(double date1, double date2, double rbp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPmat00(double date1, double date2, double[,] rbp);

    /// <summary>
    /// Precession matrix (including frame bias) from GCRS to a specified date, IAU 2000 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rbp"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pmat00(double date1, double date2, double[,] rbp) {
        SofaValidate.VerifyDouble3x3(rbp);
        iauPmat00(date1, date2, rbp);
    }


    // DECLSPEC void iauPmat06(double date1, double date2, double rbp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPmat06(double date1, double date2, double[,] rbp);

    /// <summary>
    /// Precession matrix (including frame bias) from GCRS to a specified date, IAU 2006 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rbp"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pmat06(double date1, double date2, double[,] rbp) {
        SofaValidate.VerifyDouble3x3(rbp);
        iauPmat06(date1, date2, rbp);
    }


    // DECLSPEC void iauPmat76(double date1, double date2, double rmatp[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPmat76(double date1, double date2, double[,] rmatp);

    /// <summary>
    /// Precession matrix from J2000.0 to a specified date, IAU 1976 model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rmatp">precession matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatp"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pmat76(double date1, double date2, double[,] rmatp) {
        SofaValidate.VerifyDouble3x3(rmatp);
        iauPmat76(date1, date2, rmatp);
    }


    // DECLSPEC void iauPn00(double date1, double date2, double dpsi, double deps,
    //             double *epsa,
    //             double rb[3][3], double rp[3][3], double rbp[3][3],
    //             double rn[3][3], double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn00(double date1, double date2, double dpsi, double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn);

    /// <summary>
    /// Precession-nutation, IAU 2000 model:  a multi-purpose function, supporting classical
    /// (equinox-based) use directly and CIO-based use indirectly.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    /// <param name="epsa">mean obliquity</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <param name="rn">nutation matrix</param>
    /// <param name="rbpn">GCRS-to-true matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void Pn00(double date1, double date2, double dpsi, double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        SofaValidate.VerifyDouble3x3(rn);
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPn00(date1, date2, dpsi, deps, out epsa, rb, rp, rbp, rn, rbpn);
    }


    // DECLSPEC void iauPn00a(double date1, double date2,
    //              double *dpsi, double *deps, double *epsa,
    //              double rb[3][3], double rp[3][3], double rbp[3][3],
    //              double rn[3][3], double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn00a(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn);

    /// <summary>
    /// Precession-nutation, IAU 2000A model:  a multi-purpose function, supporting classical
    /// (equinox-based) use directly and CIO-based use indirectly.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">>nutation in obliquity (radians)</param>
    /// <param name="epsa">mean obliquity</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <param name="rn">nutation matrix</param>
    /// <param name="rbpn">GCRS-to-true matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void Pn00a(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        SofaValidate.VerifyDouble3x3(rn);
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPn00a(date1, date2, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
    }


    // DECLSPEC void iauPn00b(double date1, double date2,
    //              double *dpsi, double *deps, double *epsa,
    //              double rb[3][3], double rp[3][3], double rbp[3][3],
    //              double rn[3][3], double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn00b(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn);

    /// <summary>
    /// Precession-nutation, IAU 2000B model:  a multi-purpose function, supporting classical
    /// (equinox-based) use directly and CIO-based use indirectly.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">>nutation in obliquity (radians)</param>
    /// <param name="epsa">mean obliquity</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <param name="rn">nutation matrix</param>
    /// <param name="rbpn">GCRS-to-true matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void Pn00b(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        SofaValidate.VerifyDouble3x3(rn);
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPn00b(date1, date2, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
    }


    // DECLSPEC void iauPn06(double date1, double date2, double dpsi, double deps,
    //             double *epsa,
    //             double rb[3][3], double rp[3][3], double rbp[3][3],
    //             double rn[3][3], double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn06(double date1, double date2, double dpsi, double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn);

    /// <summary>
    /// Precession-nutation, IAU 2006 model:  a multi-purpose function, supporting classical
    /// (equinox-based) use directly and CIO-based use indirectly.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">nutation in obliquity (radians)</param>
    /// <param name="epsa">mean obliquity</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <param name="rn">nutation matrix</param>
    /// <param name="rbpn">GCRS-to-true matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void Pn06(double date1, double date2, double dpsi, double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        SofaValidate.VerifyDouble3x3(rn);
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPn06(date1, date2, dpsi, deps, out epsa, rb, rp, rbp, rn, rbpn);
    }


    // DECLSPEC void iauPn06a(double date1, double date2,
    //              double *dpsi, double *deps, double *epsa,
    //              double rb[3][3], double rp[3][3], double rbp[3][3],
    //              double rn[3][3], double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn06a(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn);

    /// <summary>
    /// Precession-nutation, IAU 2006/2000A models:  a multi-purpose function, supporting classical
    /// (equinox-based) use directly and CIO-based use indirectly.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <param name="deps">>nutation in obliquity (radians)</param>
    /// <param name="epsa">mean obliquity</param>
    /// <param name="rb">frame bias matrix</param>
    /// <param name="rp">precession matrix</param>
    /// <param name="rbp">bias-precession matrix</param>
    /// <param name="rn">nutation matrix</param>
    /// <param name="rbpn">GCRS-to-true matrix</param>
    /// <exception cref="ArgumentException"> if any matrix is not dimensioned [3,3]</exception>
    public static void Pn06a(double date1, double date2, out double dpsi, out double deps,
        out double epsa, double[,] rb, double[,] rp, double[,] rbp, double[,] rn, double[,] rbpn)
    {
        SofaValidate.VerifyDouble3x3(rb);
        SofaValidate.VerifyDouble3x3(rp);
        SofaValidate.VerifyDouble3x3(rbp);
        SofaValidate.VerifyDouble3x3(rn);
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPn06a(date1, date2, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
    }


    // DECLSPEC void iauPnm00a(double date1, double date2, double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPnm00a(double date1, double date2, double[,] rbpn);

    /// <summary>
    /// Form the matrix of precession-nutation for a given date (including frame bias),
    /// equinox-based, IAU 2000A model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rbpn">bias precession nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rbpn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pnm00a(double date1, double date2, double[,] rbpn) {
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPnm00a(date1, date2, rbpn);
    }

    
    // DECLSPEC void iauPnm00b(double date1, double date2, double rbpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPnm00b(double date1, double date2, double[,] rbpn);

    /// <summary>
    /// Form the matrix of precession-nutation for a given date (including frame bias),
    /// equinox-based, IAU 2000B model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rbpn">bias precession nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rbpn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pnm00b(double date1, double date2, double[,] rbpn) {
        SofaValidate.VerifyDouble3x3(rbpn);
        iauPnm00b(date1, date2, rbpn);
    }


    // DECLSPEC void iauPnm06a(double date1, double date2, double rnpb[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPnm06a(double date1, double date2, double[,] rnpb);

    /// <summary>
    /// Form the matrix of precession-nutation for a given date (including frame bias),
    /// IAU 2006 precession and IAU 2000A nutation models.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="rnpb">bias precession nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rnpb"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pnm06a(double date1, double date2, double[,] rnpb) {
        SofaValidate.VerifyDouble3x3(rnpb);
        iauPnm06a(date1, date2, rnpb);
    }


    // DECLSPEC void iauPnm80(double date1, double date2, double rmatpn[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPnm80(double date1, double date2, double[,] rmatpn);

    /// <summary>
    /// Form the matrix of precession nutation for a given date, IAU 1976 precession model, IAU 1980
    /// nutation model.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="rmatpn">combined precession nutation matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rmatpn"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pnm80(double date1, double date2, double[,] rmatpn) {
        SofaValidate.VerifyDouble3x3(rmatpn);
        iauPnm80(date1, date2, rmatpn);
    }


    // DECLSPEC void iauPom00(double xp, double yp, double sp, double rpom[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPom00(double xp, double yp, double sp, double[,] rpom);

    /// <summary>
    /// Form the matrix of polar motion for a given date, IAU 2000.
    /// </summary>
    /// <param name="xp">polar motion x (radians)</param>
    /// <param name="yp">polar motion y (radians)</param>
    /// <param name="sp">TIO locator s' (radians)</param>
    /// <param name="rpom">polar motion matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="rpom"/> is not dimensioned [3,3]
    /// </exception>
    public static void Pom00(double xp, double yp, double sp, double[,] rpom) {
        SofaValidate.VerifyDouble3x3(rpom);
        iauPom00(xp, yp, sp, rpom);
    }


    // DECLSPEC void iauPr00(double date1, double date2, double *dpsipr, double *depspr);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPr00(double date1, double date2, out double dpsipr, out double depspr);

    /// <summary>
    /// Precession-rate part of the IAU 2000 precession-nutation models (part of MHB2000).
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="dpsipr">correction in longitude</param>
    /// <param name="depspr">correction in obliquity</param>
    public static void Pr00(double date1, double date2, out double dpsipr, out double depspr) {
        iauPr00(date1, date2, out dpsipr, out depspr);
    }


    // DECLSPEC void iauPrec76(double ep01, double ep02, double ep11, double ep12,
    //               double *zeta, double *z, double *theta);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPrec76(double ep01, double ep02, double ep11, double ep12,
        out double zeta, out double z, out double theta);

    /// <summary>
    /// Form the three Euler angles which implement general precession between two dates, using the
    /// IAU 1976 model
    /// </summary>
    /// <param name="ep01">initial TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="ep02">initial TDB Julian date part 2</param>
    /// <param name="ep11">final TDB Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="ep12">final TDB Julian date part 2</param>
    /// <param name="zeta">1st rotation: radians cw around z</param>
    /// <param name="z">3rd rotation: radians cw around z</param>
    /// <param name="theta">2nd rotation: radians ccw around y</param>
    public static void Prec76(double ep01, double ep02, double ep11, double ep12,
        out double zeta, out double z, out double theta) {
        iauPrec76(ep01, ep02, ep11, ep12, out zeta, out z, out theta);
    }


    // DECLSPEC double iauS00(double date1, double date2, double x, double y);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauS00(double date1, double date2, double x, double y);

    /// <summary>
    /// The CIO locator s, positioning the Celestial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole, given the CIP's X,Y coordinates.  Compatible with IAU 2000A
    /// precession nutation.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    /// <returns>CIO locator s (radians)</returns>
    public static double S00(double date1, double date2, double x, double y) {
        return iauS00(date1, date2, x, y);
    }


    // DECLSPEC double iauS00a(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauS00a(double date1, double date2);

    /// <summary>
    /// The CIO locator s, positioning the Celestial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole, using the IAU 2000A precession nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>CIO locator s (radians)</returns>
    public static double S00a(double date1, double date2) {
        return iauS00a(date1, date2);
    }


    // DECLSPEC double iauS00b(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauS00b(double date1, double date2);

    /// <summary>
    /// The CIO locator s, positioning the Celestial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole, using the IAU 2000B precession nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>CIO locator s (radians)</returns>
    public static double S00b(double date1, double date2) {
        return iauS00b(date1, date2);
    }


    // DECLSPEC double iauS06(double date1, double date2, double x, double y);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauS06(double date1, double date2, double x, double y);

    /// <summary>
    /// The CIO locator s, positioning the Celestial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole, given the CIP's X,Y coordinates.  Compatible with IAU
    /// 2006/2000A precession nutation.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    /// <returns>CIO locator s (radians)</returns>
    public static double S06(double date1, double date2, double x, double y) {
        return iauS06(date1, date2, x, y);
    }


    // DECLSPEC double iauS06a(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauS06a(double date1, double date2);

    /// <summary>
    /// The CIO locator s, positioning the Celestial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole, using the IAU 2006 precession and IAU 2000A nutation models.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>CIO locator s (radians)</returns>
    public static double S06a(double date1, double date2) {
        return iauS06a(date1, date2);
    }


    // DECLSPEC double iauSp00(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauSp00(double date1, double date2);

    /// <summary>
    /// The TIO locator s', positioning the Terrestrial Intermediate Origin on the equator of the
    /// Celestial Intermediate Pole.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <returns>TIO locator s' (radians)</returns>
    public static double Sp00(double date1, double date2) {
        return iauSp00(date1, date2);
    }


    // DECLSPEC void iauXy06(double date1, double date2, double *x, double *y);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauXy06(double date1, double date2, out double x, out double y);

    /// <summary>
    /// X,Y coordinates of celestial intermediate pole from series based on IAU 2006 precession and
    /// IAU 2000A nutation.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    public static void Xy06(double date1, double date2, out double x, out double y) {
        iauXy06(date1, date2, out x, out y);
    }


    // DECLSPEC void iauXys00a(double date1, double date2,
    //               double *x, double *y, double *s);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauXys00a(double date1, double date2, out double x, out double y,
        out double s);

    /// <summary>
    /// For a given TT date, compute the X,Y coordinates of the Celestial Intermediate Pole and the
    /// CIO locator s, using the IAU 2000A precession-nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    /// <param name="s">CIO locator s (radians)</param>
    public static void Xys00a(double date1, double date2, out double x, out double y,
        out double s)
    {
        iauXys00a(date1, date2, out x, out y, out s);
    }


    // DECLSPEC void iauXys00b(double date1, double date2,
    //               double *x, double *y, double *s);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauXys00b(double date1, double date2, out double x, out double y,
        out double s);

    /// <summary>
    /// For a given TT date, compute the X,Y coordinates of the Celestial Intermediate Pole and the
    /// CIO locator s, using the IAU 2000B precession-nutation model.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    /// <param name="s">CIO locator s (radians)</param>
    public static void Xys00b(double date1, double date2, out double x, out double y,
        out double s)
    {
        iauXys00b(date1, date2, out x, out y, out s);
    }


    // DECLSPEC void iauXys06a(double date1, double date2,
    //               double *x, double *y, double *s);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauXys06a(double date1, double date2, out double x, out double y,
        out double s);

    /// <summary>
    /// For a given TT date, compute the X,Y coordinates of the Celestial Intermediate Pole and the
    /// CIO locator s, using the IAU 2006 precession and IAU 2000A nutation models.
    /// </summary>
    /// <param name="date1">TT Julian date part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT Julian date part 2</param>
    /// <param name="x">CIP unit vector x</param>
    /// <param name="y">CIP unit vector y</param>
    /// <param name="s">CIO locator s (radians)</param>
    public static void Xys06a(double date1, double date2, out double x, out double y,
        out double s)
    {
        iauXys06a(date1, date2, out x, out y, out s);
    }



    // Self test this part of the class.
#if DEBUG

    /// <summary>
    /// Self test precession methods.
    /// </summary>
    public static void TestPrecession() {

        double[,] rb, rp, rn, rbp, rbpn, rnpb, rc2i, rc2t, rpom, r, rmatn, rmatp, rmatpn;
        double[] vec;
        double dpsibi, depsbi, dra, x, y, s;
        double tta, uta, ttb, utb, xp, yp, sp, era, gst, deps, dpsi, eo;
        double gamb, phib, psi, eps, epsa;
        double eps0, psia, oma, bpa, bqa, pia, bpia, chia, za, zetaa, thetaa, pa, gam, phi;
        double bzeta, bz, btheta, psib, dpsipr, depspr;
        double ep01, ep02, ep11, ep12, zeta, z, theta;
        double epj;

        rb = new double[3, 3];
        rp = new double[3, 3];
        rn = new double[3, 3];
        rbp = new double[3, 3];
        rbpn = new double[3, 3];
        rc2i = new double[3, 3];
        rc2t = new double[3, 3];
        rpom = new double[3, 3];
        rnpb = new double[3, 3];
        r = new double[3, 3];
        rmatn = new double[3, 3];
        rmatp = new double[3, 3];
        rmatpn = new double[3, 3];
        vec = new double[3];

        // DECLSPEC void iauBi00(double *dpsibi, double *depsbi, double *dra);
        Bi00(out dpsibi, out depsbi, out dra);
        SofaValidate.EqualsApprox(dpsibi, -0.2025309152835086613e-6, 1e-12);
        SofaValidate.EqualsApprox(depsbi, -0.3306041454222147847e-7, 1e-12);
        SofaValidate.EqualsApprox(dra, -0.7078279744199225506e-7, 1e-12);

        // DECLSPEC void iauBp00(double date1, double date2,
        //              double rb[3][3], double rp[3][3], double rbp[3][3]);
        Bp00(2400000.5, 50123.9999, rb, rp, rbp);
        SofaValidate.EqualsApprox(rb, new double[,] {
            {0.9999999999999942498, -0.7078279744199196626e-7, 0.8056217146976134152e-7},
            {0.7078279477857337206e-7, 0.9999999999999969484, 0.3306041454222136517e-7},
            {-0.8056217380986972157e-7, -0.3306040883980552500e-7, 0.9999999999999962084}},
            1e-12, 1e-16);
        SofaValidate.EqualsApprox(rp, new double[,] {
            {0.9999995504864048241, 0.8696113836207084411e-3, 0.3778928813389333402e-3},
            {-0.8696113818227265968e-3, 0.9999996218879365258, -0.1690679263009242066e-6},
            {-0.3778928854764695214e-3, -0.1595521004195286491e-6, 0.9999999285984682756}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9999995505175087260, 0.8695405883617884705e-3, 0.3779734722239007105e-3},
            {-0.8695405990410863719e-3, 0.9999996219494925900, -0.1360775820404982209e-6},
            {-0.3779734476558184991e-3, -0.1925857585832024058e-6, 0.9999999285680153377}},
            1e-12, 1e-14);

        // DECLSPEC void iauBp06(double date1, double date2,
        //              double rb[3][3], double rp[3][3], double rbp[3][3]);
        Bp06(2400000.5, 50123.9999, rb, rp, rbp);
        SofaValidate.EqualsApprox(rb, new double[,] {
            {0.9999999999999942497, -0.7078368960971557145e-7, 0.8056213977613185606e-7},
            {0.7078368694637674333e-7, 0.9999999999999969484, 0.3305943742989134124e-7},
            {-0.8056214211620056792e-7, -0.3305943172740586950e-7, 0.9999999999999962084}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rp, new double[,] {
            {0.9999995504864960278, 0.8696112578855404832e-3, 0.3778929293341390127e-3},
            {-0.8696112560510186244e-3, 0.9999996218880458820, -0.1691646168941896285e-6},
            {-0.3778929335557603418e-3, -0.1594554040786495076e-6, 0.9999999285984501222}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9999995505176007047, 0.8695404617348208406e-3, 0.3779735201865589104e-3},
            {-0.8695404723772031414e-3, 0.9999996219496027161, -0.1361752497080270143e-6},
            {-0.3779734957034089490e-3, -0.1924880847894457113e-6, 0.9999999285679971958}},
            1e-12, 1e-14);

        // DECLSPEC void iauBpn2xy(double rbpn[3][3], double *x, double *y);
        rbpn = new double[,] {
            {9.999962358680738e-1, -2.516417057665452e-3, -1.093569785342370e-3},
            {2.516462370370876e-3, 9.999968329010883e-1, 4.006159587358310e-5},
            {1.093465510215479e-3, -4.281337229063151e-5, 9.999994012499173e-1}};
        Bpn2xy(rbpn, out x, out y);
        SofaValidate.EqualsApprox(x,  1.093465510215479e-3, 1e-12);
        SofaValidate.EqualsApprox(y, -4.281337229063151e-5, 1e-12);
   
        // DECLSPEC void iauC2i00a(double date1, double date2, double rc2i[3][3]);
        C2i00a(2400000.5, 53736.0, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999998323037165557, 0.5581526348992140183e-9, -0.5791308477073443415e-3},
            {-0.2384266227870752452e-7, 0.9999999991917405258, -0.4020594955028209745e-4},
            {0.5791308472168152904e-3, 0.4020595661591500259e-4, 0.9999998314954572304}}, 1e-12);

        // DECLSPEC void iauC2i00b(double date1, double date2, double rc2i[3][3]);
        C2i00b(2400000.5, 53736.0, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999998323040954356, 0.5581526349131823372e-9, -0.5791301934855394005e-3},
            {-0.2384239285499175543e-7, 0.9999999991917574043, -0.4020552974819030066e-4},
            {0.5791301929950208873e-3, 0.4020553681373720832e-4, 0.9999998314958529887}}, 1e-12);

        // DECLSPEC void iauC2i06a(double date1, double date2, double rc2i[3][3]);
        C2i06a(2400000.5, 53736.0, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999998323037159379, 0.5581121329587613787e-9, -0.5791308487740529749e-3},
            {-0.2384253169452306581e-7, 0.9999999991917467827, -0.4020579392895682558e-4},
            {0.5791308482835292617e-3, 0.4020580099454020310e-4, 0.9999998314954628695}}, 1e-12);

        // DECLSPEC void iauC2ibpn(double date1, double date2, double rbpn[3][3],
        //               double rc2i[3][3]);
        rbpn = new double[,] {
            {9.999962358680738e-1, -2.516417057665452e-3, -1.093569785342370e-3},
            {2.516462370370876e-3, 9.999968329010883e-1, 4.006159587358310e-5},
            {1.093465510215479e-3, -4.281337229063151e-5, 9.999994012499173e-1}};
        C2ibpn(2400000.5, 50123.9999, rbpn, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999994021664089977, -0.3869195948017503664e-8, -0.1093465511383285076e-2},
            {0.5068413965715446111e-7, 0.9999999990835075686, 0.4281334246452708915e-4},
            {0.1093465510215479000e-2, -0.4281337229063151000e-4, 0.9999994012499173103}}, 1e-12);
   
        // DECLSPEC void iauC2ixy(double date1, double date2, double x, double y,
        //              double rc2i[3][3]);
        x = 0.5791308486706011000e-3;
        y = 0.4020579816732961219e-4;
        C2ixy(2400000.5, 53736, x, y, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999998323037157138, 0.5581526349032241205e-9, -0.5791308491611263745e-3},
            {-0.2384257057469842953e-7, 0.9999999991917468964, -0.4020579110172324363e-4},
            {0.5791308486706011000e-3, 0.4020579816732961219e-4, 0.9999998314954627590}}, 1e-12);

        // DECLSPEC void iauC2ixys(double x, double y, double s, double rc2i[3][3]);
        x = 0.5791308486706011000e-3;
        y = 0.4020579816732961219e-4;
        s = -0.1220040848472271978e-7;
        C2ixys(x, y, s, rc2i);
        SofaValidate.EqualsApprox(rc2i, new double[,] {
            {0.9999998323037157138, 0.5581984869168499149e-9, -0.5791308491611282180e-3},
            {-0.2384261642670440317e-7, 0.9999999991917468964, -0.4020579110169668931e-4},
            {0.5791308486706011000e-3, 0.4020579816732961219e-4, 0.9999998314954627590}}, 1e-12);

        // DECLSPEC void iauC2t00a(double tta, double ttb, double uta, double utb,
        //               double xp, double yp, double rc2t[3][3]);
        tta = 2400000.5;
        uta = 2400000.5;
        ttb = 53736.0;
        utb = 53736.0;
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        C2t00a(tta, ttb, uta, utb, xp, yp, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128307182668, 0.9834769806938457836, 0.6555535638688341725e-4},
            {-0.9834768134135984552, -0.1810332203649520727, 0.5749801116141056317e-3},
            {0.5773474014081406921e-3, 0.3961832391770163647e-4, 0.9999998325501692289}}, 1e-12);

        // DECLSPEC void iauC2t00b(double tta, double ttb, double uta, double utb,
        //               double xp, double yp, double rc2t[3][3]);
        tta = 2400000.5;
        uta = 2400000.5;
        ttb = 53736.0;
        utb = 53736.0;
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        C2t00b(tta, ttb, uta, utb, xp, yp, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128439678965, 0.9834769806913872359, 0.6555565082458415611e-4},
            {-0.9834768134115435923, -0.1810332203784001946, 0.5749793922030017230e-3},
            {0.5773467471863534901e-3, 0.3961790411549945020e-4, 0.9999998325505635738}}, 1e-12);
   

        // DECLSPEC void iauC2t06a(double tta, double ttb, double uta, double utb,
        //               double xp, double yp, double rc2t[3][3]);
        tta = 2400000.5;
        uta = 2400000.5;
        ttb = 53736.0;
        utb = 53736.0;
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        C2t06a(tta, ttb, uta, utb, xp, yp, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128305897282, 0.9834769806938592296, 0.6555550962998436505e-4},
            {-0.9834768134136214897, -0.1810332203649130832, 0.5749800844905594110e-3},
            {0.5773474024748545878e-3, 0.3961816829632690581e-4, 0.9999998325501747785}}, 1e-12);

        // DECLSPEC void iauC2tcio(double rc2i[3][3], double era, double rpom[3][3],
        //               double rc2t[3][3]);
        rc2i = new double[,] {
            {0.9999998323037164738, 0.5581526271714303683e-9, -0.5791308477073443903e-3},
            {-0.2384266227524722273e-7, 0.9999999991917404296, -0.4020594955030704125e-4},
            {0.5791308472168153320e-3, 0.4020595661593994396e-4, 0.9999998314954572365}};
        era = 1.75283325530307;
        rpom = new double[,] {
            {0.9999999999999674705, -0.1367174580728847031e-10, 0.2550602379999972723e-6},
            {0.1414624947957029721e-10, 0.9999999999982694954, -0.1860359246998866338e-5},
            {-0.2550602379741215275e-6, 0.1860359247002413923e-5, 0.9999999999982369658}};
        C2tcio(rc2i, era, rpom, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128307110439, 0.9834769806938470149, 0.6555535638685466874e-4},
            {-0.9834768134135996657, -0.1810332203649448367, 0.5749801116141106528e-3},
            {0.5773474014081407076e-3, 0.3961832391772658944e-4, 0.9999998325501691969}}, 1e-12);

       // DECLSPEC void iauC2teqx(double rbpn[3][3], double gst, double rpom[3][3],
        //               double rc2t[3][3]);
        rbpn = new double[,] {
            {0.9999989440476103608, -0.1332881761240011518e-2, -0.5790767434730085097e-3},
            {0.1332858254308954453e-2, 0.9999991109044505944, -0.4097782710401555759e-4},
            {0.5791308472168153320e-3, 0.4020595661593994396e-4, 0.9999998314954572365}};
        gst = 1.754166138040730516;
        rpom = new double[,] {
            {0.9999999999999674705, -0.1367174580728847031e-10,  0.2550602379999972723e-6},
            {0.1414624947957029721e-10, 0.9999999999982694954, -0.1860359246998866338e-5},
            {-0.2550602379741215275e-6, 0.1860359247002413923e-5, 0.9999999999982369658}};
        C2teqx(rbpn, gst, rpom, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128528685730, 0.9834769806897685071, 0.6555535639982634449e-4},
            {-0.9834768134095211257, -0.1810332203871023800, 0.5749801116126438962e-3},
            {0.5773474014081539467e-3, 0.3961832391768640871e-4, 0.9999998325501691969}}, 1e-12);

        // DECLSPEC void iauC2tpe(double tta, double ttb, double uta, double utb,
        //              double dpsi, double deps, double xp, double yp,
        //              double rc2t[3][3]);
        tta = 2400000.5;
        uta = 2400000.5;
        ttb = 53736.0;
        utb = 53736.0;
        deps = 0.4090789763356509900;
        dpsi = -0.9630909107115582393e-5;
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        C2tpe(tta, ttb, uta, utb, dpsi, deps, xp, yp, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1813677995763029394, 0.9023482206891683275, -0.3909902938641085751},
            {-0.9834147641476804807, -0.1659883635434995121, 0.7309763898042819705e-1},
            {0.1059685430673215247e-2, 0.3977631855605078674, 0.9174875068792735362}}, 1e-12);

        // DECLSPEC void iauC2txy(double tta, double ttb, double uta, double utb,
        //              double x, double y, double xp, double yp,
        //              double rc2t[3][3]);
        tta = 2400000.5;
        uta = 2400000.5;
        ttb = 53736.0;
        utb = 53736.0;
        x = 0.5791308486706011000e-3;
        y = 0.4020579816732961219e-4;
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        C2txy(tta, ttb, uta, utb, x, y, xp, yp, rc2t);
        SofaValidate.EqualsApprox(rc2t, new double[,] {
            {-0.1810332128306279253, 0.9834769806938520084, 0.6555551248057665829e-4},
            {-0.9834768134136142314, -0.1810332203649529312, 0.5749800843594139912e-3},
            {0.5773474028619264494e-3, 0.3961816546911624260e-4, 0.9999998325501746670}}, 1e-12);

        // DECLSPEC double iauEo06a(double date1, double date2);
        eo = Eo06a(2400000.5, 53736.0);
        SofaValidate.EqualsApprox(eo, -0.1332882371941833644e-2, 1e-15);

        // DECLSPEC double iauEors(double rnpb[3][3], double s);
        rnpb = new double[,] {
            {0.9999989440476103608, -0.1332881761240011518e-2, -0.5790767434730085097e-3},
            {0.1332858254308954453e-2, 0.9999991109044505944, -0.4097782710401555759e-4},
            {0.5791308472168153320e-3, 0.4020595661593994396e-4, 0.9999998314954572365}};
        s = -0.1220040848472271978e-7;
        eo = Eors(rnpb, s);
        SofaValidate.EqualsApprox(eo, -0.1332882715130744606e-2, 1e-14);

        // DECLSPEC void iauFw2m(double gamb, double phib, double psi, double eps,
        //             double r[3][3]);
        gamb = -0.2243387670997992368e-5;
        phib = 0.4091014602391312982;
        psi = -0.9501954178013015092e-3;
        eps = 0.4091014316587367472;
        Fw2m(gamb, phib, psi, eps, r);
        SofaValidate.EqualsApprox(r, new double[,] {
            {0.9999995505176007047, 0.8695404617348192957e-3, 0.3779735201865582571e-3},
            {-0.8695404723772016038e-3, 0.9999996219496027161, -0.1361752496887100026e-6},
            {-0.3779734957034082790e-3, -0.1924880848087615651e-6, 0.9999999285679971958}}, 1e-12);

        // DECLSPEC void iauFw2xy(double gamb, double phib, double psi, double eps,
        //              double *x, double *y);
        gamb = -0.2243387670997992368e-5;
        phib = 0.4091014602391312982;
        psi = -0.9501954178013015092e-3;
        eps = 0.4091014316587367472;
        Fw2xy(gamb, phib, psi, eps, out x, out y);
        SofaValidate.EqualsApprox(x, -0.3779734957034082790e-3, 1e-14);
        SofaValidate.EqualsApprox(y, -0.1924880848087615651e-6, 1e-14);

        // DECLSPEC void iauLtp(double epj, double rp[3][3]);
        epj = 1666.666;
        Ltp(epj, rp);
        SofaValidate.EqualsApprox(rp, new double[,] {
            {0.9967044141159213819, 0.7437801893193210840e-1, 0.3237624409345603401e-1},
            {-0.7437802731819618167e-1, 0.9972293894454533070, -0.1205768842723593346e-2},
            {-0.3237622482766575399e-1, -0.1206286039697609008e-2, 0.9994750246704010914}
        }, 1e-14);

        // DECLSPEC void iauLtpb(double epj, double rpb[3][3]);
        epj = 1666.666;
        Ltpb(epj, rbp);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9967044167723271851, 0.7437794731203340345e-1, 0.3237632684841625547e-1},
            {-0.7437795663437177152e-1, 0.9972293947500013666, -0.1205741865911243235e-2},
            {-0.3237630543224664992e-1, -0.1206316791076485295e-2, 0.9994750220222438819},
        }, 1e-14);

        // DECLSPEC void iauLtpecl(double epj, double vec[3]);
        epj = -1500.0;
        Ltpecl(epj, vec);
        SofaValidate.EqualsApprox(vec, new double[3]
            { 0.4768625676477096525e-3, -0.4052259533091875112, 0.9142164401096448012 }, 1e-14);

        // DECLSPEC void iauLtpequ(double epj, double veq[3]);
        epj = -2500.0;
        Ltpequ(epj, vec);
        SofaValidate.EqualsApprox(vec, new double[]
            { -0.3586652560237326659, -0.1996978910771128475, 0.9118552442250819624 }, 1e-14);

        // DECLSPEC void iauNum00a(double date1, double date2, double rmatn[3][3]);
        Num00a(2400000.5, 53736.0, rmatn);
        SofaValidate.EqualsApprox(rmatn, new double[,] {
            {0.9999999999536227949, 0.8836238544090873336e-5, 0.3830835237722400669e-5},
            {-0.8836082880798569274e-5, 0.9999999991354655028, -0.4063240865362499850e-4},
            {-0.3831194272065995866e-5, 0.4063237480216291775e-4, 0.9999999991671660338}}, 1e-12);

        // DECLSPEC void iauNum00b(double date1, double date2, double rmatn[3][3]);
        Num00b(2400000.5, 53736, rmatn);
        SofaValidate.EqualsApprox(rmatn, new double[,] {
            {0.9999999999536069682, 0.8837746144871248011e-5, 0.3831488838252202945e-5},
            {-0.8837590456632304720e-5, 0.9999999991354692733, -0.4063198798559591654e-4},
            {-0.3831847930134941271e-5, 0.4063195412258168380e-4, 0.9999999991671806225}}, 1e-12);

        // DECLSPEC void iauNum06a(double date1, double date2, double rmatn[3][3]);
        Num06a(2400000.5, 53736, rmatn);
        SofaValidate.EqualsApprox(rmatn, new double[,] {
            {0.9999999999536227668, 0.8836241998111535233e-5, 0.3830834608415287707e-5},
            {-0.8836086334870740138e-5, 0.9999999991354657474, -0.4063240188248455065e-4},
            {-0.3831193642839398128e-5, 0.4063236803101479770e-4, 0.9999999991671663114}}, 1e-12);

        // DECLSPEC void iauNumat(double epsa, double dpsi, double deps, double rmatn[3][3]);
        epsa = 0.4090789763356509900;
        dpsi = -0.9630909107115582393e-5;
        deps = 0.4063239174001678826e-4;
        Numat(epsa, dpsi, deps, rmatn);
        SofaValidate.EqualsApprox(rmatn, new double[,] {
            {0.9999999999536227949, 0.8836239320236250577e-5, 0.3830833447458251908e-5},
            {-0.8836083657016688588e-5, 0.9999999991354654959, -0.4063240865361857698e-4},
            {-0.3831192481833385226e-5, 0.4063237480216934159e-4, 0.9999999991671660407}}, 1e-12);

        // DECLSPEC void iauNut00a(double date1, double date2, double *dpsi, double *deps);
        Nut00a(2400000.5, 53736.0, out dpsi, out deps);
        SofaValidate.EqualsApprox(dpsi, -0.9630909107115518431e-5, 1e-13);
        SofaValidate.EqualsApprox(deps, 0.4063239174001678710e-4, 1e-13);

        // DECLSPEC void iauNut00b(double date1, double date2, double *dpsi, double *deps);
        Nut00b(2400000.5, 53736.0, out dpsi, out deps);
        SofaValidate.EqualsApprox(dpsi, -0.9632552291148362783e-5, 1e-13);
        SofaValidate.EqualsApprox(deps, 0.4063197106621159367e-4, 1e-13);

        // DECLSPEC void iauNut06a(double date1, double date2, double *dpsi, double *deps);
        Nut06a(2400000.5, 53736.0, out dpsi, out deps);
        SofaValidate.EqualsApprox(dpsi, -0.9630912025820308797e-5, 1e-13);
        SofaValidate.EqualsApprox(deps, 0.4063238496887249798e-4, 1e-13);

        // DECLSPEC void iauNut80(double date1, double date2, double *dpsi, double *deps);
        Nut80(2400000.5, 53736.0, out dpsi, out deps);
        SofaValidate.EqualsApprox(dpsi, -0.9643658353226563966e-5, 1e-13);
        SofaValidate.EqualsApprox(deps, 0.4060051006879713322e-4, 1e-13);

        // DECLSPEC void iauNutm80(double date1, double date2, double rmatn[3][3]);
        Nutm80(2400000.5, 53736.0, rmatn);
        SofaValidate.EqualsApprox(rmatn, new double[,] {
            {0.9999999999534999268, 0.8847935789636432161e-5, 0.3835906502164019142e-5},
            {-0.8847780042583435924e-5, 0.9999999991366569963, -0.4060052702727130809e-4},
            {-0.3836265729708478796e-5, 0.4060049308612638555e-4, 0.9999999991684415129}}, 1e-12);

        // DECLSPEC double iauObl06(double date1, double date2);
        SofaValidate.EqualsApprox(Obl06(2400000.5, 54388.0), 0.4090749229387258204, 1e-14);

        // DECLSPEC double iauObl80(double date1, double date2);
        SofaValidate.EqualsApprox(Obl80(2400000.5, 54388.0), 0.4090751347643816218, 1e-14);

        // DECLSPEC void iauP06e(double date1, double date2,
        //             double *eps0, double *psia, double *oma, double *bpa,
        //             double *bqa, double *pia, double *bpia,
        //             double *epsa, double *chia, double *za, double *zetaa,
        //             double *thetaa, double *pa,
        //             double *gam, double *phi, double *psi);
        P06e(2400000.5, 52541.0, out eps0, out psia, out oma, out bpa,
           out bqa, out pia, out bpia, out epsa, out chia, out za,
           out zetaa, out thetaa, out pa, out gam, out phi, out psi);
        SofaValidate.EqualsApprox(eps0, 0.4090926006005828715, 1e-14);
        SofaValidate.EqualsApprox(psia, 0.6664369630191613431e-3, 1e-14);
        SofaValidate.EqualsApprox(oma, 0.4090925973783255982, 1e-14);
        SofaValidate.EqualsApprox(bpa, 0.5561149371265209445e-6, 1e-14);
        SofaValidate.EqualsApprox(bqa, -0.6191517193290621270e-5, 1e-14);
        SofaValidate.EqualsApprox(pia, 0.6216441751884382923e-5, 1e-14);
        SofaValidate.EqualsApprox(bpia, 3.052014180023779882, 1e-14);
        SofaValidate.EqualsApprox(epsa, 0.4090864054922431688, 1e-14);
        SofaValidate.EqualsApprox(chia, 0.1387703379530915364e-5, 1e-14);
        SofaValidate.EqualsApprox(za, 0.2921789846651790546e-3, 1e-14);
        SofaValidate.EqualsApprox(zetaa, 0.3178773290332009310e-3, 1e-14);
        SofaValidate.EqualsApprox(thetaa, 0.2650932701657497181e-3, 1e-14);
        SofaValidate.EqualsApprox(pa, 0.6651637681381016344e-3, 1e-14);
        SofaValidate.EqualsApprox(gam, 0.1398077115963754987e-5, 1e-14);
        SofaValidate.EqualsApprox(phi, 0.4090864090837462602, 1e-14);
        SofaValidate.EqualsApprox(psi, 0.6664464807480920325e-3, 1e-14);

        // DECLSPEC void iauPb06(double date1, double date2,
        //             double *bzeta, double *bz, double *btheta);
        Pb06(2400000.5, 50123.9999, out bzeta, out bz, out btheta);
        SofaValidate.EqualsApprox(bzeta, -0.5092634016326478238e-3, 1e-12);
        SofaValidate.EqualsApprox(bz, -0.3602772060566044413e-3, 1e-12);
        SofaValidate.EqualsApprox(btheta, -0.3779735537167811177e-3, 1e-12);

        // DECLSPEC void iauPfw06(double date1, double date2,
        //              double *gamb, double *phib, double *psib, double *epsa);
        Pfw06(2400000.5, 50123.9999, out gamb, out phib, out psib, out epsa);
        SofaValidate.EqualsApprox(gamb, -0.2243387670997995690e-5, 1e-16);
        SofaValidate.EqualsApprox(phib, 0.4091014602391312808, 1e-12);
        SofaValidate.EqualsApprox(psib, -0.9501954178013031895e-3, 1e-14);
        SofaValidate.EqualsApprox(epsa, 0.4091014316587367491, 1e-12);

        // DECLSPEC void iauPmat00(double date1, double date2, double rbp[3][3]);
        Pmat00(2400000.5, 50123.9999, rbp);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9999995505175087260, 0.8695405883617884705e-3, 0.3779734722239007105e-3},
            {-0.8695405990410863719e-3, 0.9999996219494925900, -0.1360775820404982209e-6},
            {-0.3779734476558184991e-3, -0.1925857585832024058e-6, 0.9999999285680153377}},
            1e-12, 1e-14);

        // DECLSPEC void iauPmat06(double date1, double date2, double rbp[3][3]);
        Pmat06(2400000.5, 50123.9999, rbp);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9999995505176007047, 0.8695404617348208406e-3, 0.3779735201865589104e-3},
            {-0.8695404723772031414e-3, 0.9999996219496027161, -0.1361752497080270143e-6},
            {-0.3779734957034089490e-3, -0.1924880847894457113e-6, 0.9999999285679971958}},
            1e-12, 1e-14);

        // DECLSPEC void iauPmat76(double date1, double date2, double rmatp[3][3]);
        Pmat76(2400000.5, 50123.9999, rmatp);
        SofaValidate.EqualsApprox(rmatp, new double[,] {
            {0.9999995504328350733, 0.8696632209480960785e-3, 0.3779153474959888345e-3},
            {-0.8696632209485112192e-3, 0.9999996218428560614, -0.1643284776111886407e-6},
            {-0.3779153474950335077e-3, -0.1643306746147366896e-6, 0.9999999285899790119}},
            1e-12, 1e-14);

        // DECLSPEC void iauPn00(double date1, double date2, double dpsi, double deps,
        //             double *epsa,
        //             double rb[3][3], double rp[3][3], double rbp[3][3],
        //             double rn[3][3], double rbpn[3][3]);
        dpsi = -0.9632552291149335877e-5;
        deps = 0.4063197106621141414e-4;
        Pn00(2400000.5, 53736.0, dpsi, deps, out epsa, rb, rp, rbp, rn, rbpn);
        SofaValidate.EqualsApprox(epsa, 0.4090791789404229916, 1e-12);
        SofaValidate.EqualsApprox(rb, new double[,] {
            {0.9999999999999942498, -0.7078279744199196626e-7, 0.8056217146976134152e-7},
            {0.7078279477857337206e-7, 0.9999999999999969484, 0.3306041454222136517e-7},
            {-0.8056217380986972157e-7, -0.3306040883980552500e-7, 0.9999999999999962084}},
            1e-12, 1e-18);
        SofaValidate.EqualsApprox(rp, new double[,] {
            {0.9999989300532289018, -0.1341647226791824349e-2, -0.5829880927190296547e-3},
            {0.1341647231069759008e-2, 0.9999990999908750433, -0.3837444441583715468e-6},
            {0.5829880828740957684e-3, -0.3984203267708834759e-6, 0.9999998300623538046}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,] {
            {0.9999989300052243993, -0.1341717990239703727e-2, -0.5829075749891684053e-3},
            {0.1341718013831739992e-2, 0.9999990998959191343, -0.3505759733565421170e-6},
            {0.5829075206857717883e-3, -0.4315219955198608970e-6, 0.9999998301093036269}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rn, new double[,] {
            {0.9999999999536069682, 0.8837746144872140812e-5, 0.3831488838252590008e-5},
            {-0.8837590456633197506e-5, 0.9999999991354692733, -0.4063198798559573702e-4},
            {-0.3831847930135328368e-5, 0.4063195412258150427e-4, 0.9999999991671806225}},
            1e-12, 1e-16);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999989440499982806, -0.1332880253640848301e-2, -0.5790760898731087295e-3},
            {0.1332856746979948745e-2, 0.9999991109064768883, -0.4097740555723063806e-4},
            {0.5791301929950205000e-3, 0.4020553681373702931e-4, 0.9999998314958529887}},
            1e-12, 1e-14);

        // DECLSPEC void iauPn00a(double date1, double date2,
        //              double *dpsi, double *deps, double *epsa,
        //              double rb[3][3], double rp[3][3], double rbp[3][3],
        //              double rn[3][3], double rbpn[3][3]);
        Pn00a(2400000.5, 53736.0, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
        SofaValidate.EqualsApprox(dpsi, -0.9630909107115518431e-5, 1e-12);
        SofaValidate.EqualsApprox(deps, 0.4063239174001678710e-4, 1e-12);
        SofaValidate.EqualsApprox(epsa, 0.4090791789404229916, 1e-12);
        SofaValidate.EqualsApprox(rb, new double[,] {
            {0.9999999999999942498, -0.7078279744199196626e-7, 0.8056217146976134152e-7},
            {0.7078279477857337206e-7, 0.9999999999999969484, 0.3306041454222136517e-7},
            {-0.8056217380986972157e-7, -0.3306040883980552500e-7, 0.9999999999999962084}},
            1e-12, 1e-16);
        SofaValidate.EqualsApprox(rp, new double[,]{
            {0.9999989300532289018, -0.1341647226791824349e-2, -0.5829880927190296547e-3},
            {0.1341647231069759008e-2, 0.9999990999908750433, -0.3837444441583715468e-6},
            {0.5829880828740957684e-3, -0.3984203267708834759e-6, 0.9999998300623538046}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,]{
            {0.9999989300052243993, -0.1341717990239703727e-2, -0.5829075749891684053e-3},
            {0.1341718013831739992e-2, 0.9999990998959191343, -0.3505759733565421170e-6},
            {0.5829075206857717883e-3, -0.4315219955198608970e-6, 0.9999998301093036269}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rn, new double[,]{
            {0.9999999999536227949, 0.8836238544090873336e-5, 0.3830835237722400669e-5},
            {-0.8836082880798569274e-5, 0.9999999991354655028, -0.4063240865362499850e-4},
            {-0.3831194272065995866e-5, 0.4063237480216291775e-4, 0.9999999991671660338}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999989440476103435, -0.1332881761240011763e-2, -0.5790767434730085751e-3},
            {0.1332858254308954658e-2, 0.9999991109044505577, -0.4097782710396580452e-4},
            {0.5791308472168152904e-3, 0.4020595661591500259e-4, 0.9999998314954572304}},
            1e-12, 1e-14);

        // DECLSPEC void iauPn00b(double date1, double date2,
        //              double *dpsi, double *deps, double *epsa,
        //              double rb[3][3], double rp[3][3], double rbp[3][3],
        //              double rn[3][3], double rbpn[3][3]);
        Pn00b(2400000.5, 53736.0, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
        SofaValidate.EqualsApprox(dpsi, -0.9632552291148362783e-5, 1e-12);
        SofaValidate.EqualsApprox(deps, 0.4063197106621159367e-4, 1e-12);
        SofaValidate.EqualsApprox(epsa, 0.4090791789404229916, 1e-12);
        SofaValidate.EqualsApprox(rb, new double[,] {
            {0.9999999999999942498, -0.7078279744199196626e-7, 0.8056217146976134152e-7},
            {0.7078279477857337206e-7, 0.9999999999999969484, 0.3306041454222136517e-7},
            {-0.8056217380986972157e-7, -0.3306040883980552500e-7, 0.9999999999999962084}},
            1e-12, 1e-16);
        SofaValidate.EqualsApprox(rp, new double[,]{
            {0.9999989300532289018, -0.1341647226791824349e-2, -0.5829880927190296547e-3},
            {0.1341647231069759008e-2, 0.9999990999908750433, -0.3837444441583715468e-6},
            {0.5829880828740957684e-3, -0.3984203267708834759e-6, 0.9999998300623538046}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,]{
            {0.9999989300052243993, -0.1341717990239703727e-2, -0.5829075749891684053e-3},
            {0.1341718013831739992e-2, 0.9999990998959191343, -0.3505759733565421170e-6},
            {0.5829075206857717883e-3, -0.4315219955198608970e-6, 0.9999998301093036269}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rn, new double[,]{
            {0.9999999999536069682, 0.8837746144871248011e-5, 0.3831488838252202945e-5},
            {-0.8837590456632304720e-5, 0.9999999991354692733, -0.4063198798559591654e-4},
            {-0.3831847930134941271e-5, 0.4063195412258168380e-4, 0.9999999991671806225}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999989440499982806, -0.1332880253640849194e-2, -0.5790760898731091166e-3},
            {0.1332856746979949638e-2,  0.9999991109064768883, -0.4097740555723081811e-4},
            {0.5791301929950208873e-3, 0.4020553681373720832e-4, 0.9999998314958529887}},
            1e-12, 1e-14);

        // DECLSPEC void iauPn06(double date1, double date2, double dpsi, double deps,
        //             double *epsa,
        //             double rb[3][3], double rp[3][3], double rbp[3][3],
        //             double rn[3][3], double rbpn[3][3]);
        dpsi = -0.9632552291149335877e-5;
        deps = 0.4063197106621141414e-4;
        Pn06(2400000.5, 53736.0, dpsi, deps, out epsa, rb, rp, rbp, rn, rbpn);
        SofaValidate.EqualsApprox(epsa, 0.4090789763356509926, 1e-12);
        SofaValidate.EqualsApprox(rb, new double[,]{
            {0.9999999999999942497, -0.7078368960971557145e-7, 0.8056213977613185606e-7},
            {0.7078368694637674333e-7, 0.9999999999999969484, 0.3305943742989134124e-7},
            {-0.8056214211620056792e-7, -0.3305943172740586950e-7, 0.9999999999999962084}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rp, new double[,]{
            {0.9999989300536854831, -0.1341646886204443795e-2, -0.5829880933488627759e-3},
            {0.1341646890569782183e-2, 0.9999990999913319321, -0.3835944216374477457e-6},
            {0.5829880833027867368e-3, -0.3985701514686976112e-6, 0.9999998300623534950}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,]{
            {0.9999989300056797893, -0.1341717650545059598e-2, -0.5829075756493728856e-3},
            {0.1341717674223918101e-2, 0.9999990998963748448, -0.3504269280170069029e-6},
            {0.5829075211461454599e-3, -0.4316708436255949093e-6, 0.9999998301093032943}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rn, new double[,]{
            {0.9999999999536069682, 0.8837746921149881914e-5, 0.3831487047682968703e-5},
            {-0.8837591232983692340e-5, 0.9999999991354692664, -0.4063198798558931215e-4},
            {-0.3831846139597250235e-5, 0.4063195412258792914e-4, 0.9999999991671806293}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999989440504506688, -0.1332879913170492655e-2, -0.5790760923225655753e-3},
            {0.1332856406595754748e-2, 0.9999991109069366795, -0.4097725651142641812e-4},
            {0.5791301952321296716e-3, 0.4020538796195230577e-4, 0.9999998314958576778}},
            1e-12, 1e-14);

        // DECLSPEC void iauPn06a(double date1, double date2,
        //              double *dpsi, double *deps, double *epsa,
        //              double rb[3][3], double rp[3][3], double rbp[3][3],
        //              double rn[3][3], double rbpn[3][3]);
        Pn06a(2400000.5, 53736.0, out dpsi, out deps, out epsa, rb, rp, rbp, rn, rbpn);
        SofaValidate.EqualsApprox(dpsi, -0.9630912025820308797e-5, 1e-12);
        SofaValidate.EqualsApprox(deps, 0.4063238496887249798e-4, 1e-12);
        SofaValidate.EqualsApprox(epsa, 0.4090789763356509926, 1e-12);
        SofaValidate.EqualsApprox(rb, new double[,]{
            {0.9999999999999942497, -0.7078368960971557145e-7, 0.8056213977613185606e-7},
            {0.7078368694637674333e-7, 0.9999999999999969484, 0.3305943742989134124e-7},
            {-0.8056214211620056792e-7, -0.3305943172740586950e-7, 0.9999999999999962084}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rp, new double[,]{
            {0.9999989300536854831, -0.1341646886204443795e-2, -0.5829880933488627759e-3},
            {0.1341646890569782183e-2, 0.9999990999913319321, -0.3835944216374477457e-6},
            {0.5829880833027867368e-3, -0.3985701514686976112e-6, 0.9999998300623534950}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbp, new double[,]{
            {0.9999989300056797893, -0.1341717650545059598e-2, -0.5829075756493728856e-3},
            {0.1341717674223918101e-2, 0.9999990998963748448, -0.3504269280170069029e-6},
            {0.5829075211461454599e-3, -0.4316708436255949093e-6, 0.9999998301093032943}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rn, new double[,]{
            {0.9999999999536227668, 0.8836241998111535233e-5, 0.3830834608415287707e-5},
            {-0.8836086334870740138e-5, 0.9999999991354657474, -0.4063240188248455065e-4},
            {-0.3831193642839398128e-5, 0.4063236803101479770e-4, 0.9999999991671663114}},
            1e-12, 1e-14);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999989440480669738, -0.1332881418091915973e-2, -0.5790767447612042565e-3},
            {0.1332857911250989133e-2, 0.9999991109049141908, -0.4097767128546784878e-4},
            {0.5791308482835292617e-3, 0.4020580099454020310e-4, 0.9999998314954628695}},
            1e-12, 1e-14);

        // DECLSPEC void iauPnm00a(double date1, double date2, double rbpn[3][3]);
        Pnm00a(2400000.5, 50123.9999, rbpn);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999995832793134257, 0.8372384254137809439e-3, 0.3639684306407150645e-3},
            {-0.8372535226570394543e-3, 0.9999996486491582471, 0.4132915262664072381e-4},
            {-0.3639337004054317729e-3, -0.4163386925461775873e-4, 0.9999999329094390695}},
            1e-12, 1e-14);

        // DECLSPEC void iauPnm00b(double date1, double date2, double rbpn[3][3]);
        Pnm00b(2400000.5, 50123.9999, rbpn);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999995832776208280, 0.8372401264429654837e-3, 0.3639691681450271771e-3},
            {-0.8372552234147137424e-3, 0.9999996486477686123, 0.4132832190946052890e-4},
            {-0.3639344385341866407e-3, -0.4163303977421522785e-4, 0.9999999329092049734}},
            1e-12, 1e-14);

        // DECLSPEC void iauPnm06a(double date1, double date2, double rnpb[3][3]);
        Pnm06a(2400000.5, 50123.9999, rbpn);
        SofaValidate.EqualsApprox(rbpn, new double[,]{
            {0.9999995832794205484, 0.8372382772630962111e-3, 0.3639684771140623099e-3},
            {-0.8372533744743683605e-3, 0.9999996486492861646, 0.4132905944611019498e-4},
            {-0.3639337469629464969e-3, -0.4163377605910663999e-4, 0.9999999329094260057}},
            1e-12, 1e-14);

        // DECLSPEC void iauPnm80(double date1, double date2, double rmatpn[3][3]);
        Pnm80(2400000.5, 50123.9999, rmatpn);
        SofaValidate.EqualsApprox(rmatpn, new double[,]{
            {0.9999995831934611169, 0.8373654045728124011e-3, 0.3639121916933106191e-3},
            {-0.8373804896118301316e-3, 0.9999996485439674092, 0.4130202510421549752e-4},
            {-0.3638774789072144473e-3, -0.4160674085851722359e-4, 0.9999999329310274805}},
            1e-12, 1e-14);

        // DECLSPEC void iauPom00(double xp, double yp, double sp, double rpom[3][3]);
        xp = 2.55060238e-7;
        yp = 1.860359247e-6;
        sp = -0.1367174580728891460e-10;
        Pom00(xp, yp, sp, rpom);
        SofaValidate.EqualsApprox(rpom, new double[,]{
            {0.9999999999999674721, -0.1367174580728846989e-10, 0.2550602379999972345e-6},
            {0.1414624947957029801e-10, 0.9999999999982695317, -0.1860359246998866389e-5},
            {-0.2550602379741215021e-6, 0.1860359247002414021e-5, 0.9999999999982370039}},
            1e-12, 1e-16);

        // DECLSPEC void iauPr00(double date1, double date2, double *dpsipr, double *depspr);
        Pr00(2400000.5, 53736, out dpsipr, out depspr);
        SofaValidate.EqualsApprox(dpsipr, -0.8716465172668347629e-7, 1e-22);
        SofaValidate.EqualsApprox(depspr, -0.7342018386722813087e-8, 1e-22);

        // DECLSPEC void iauPrec76(double ep01, double ep02, double ep11, double ep12,
        //               double *zeta, double *z, double *theta);
        ep01 = 2400000.5;
        ep02 = 33282.0;
        ep11 = 2400000.5;
        ep12 = 51544.0;
        Prec76(ep01, ep02, ep11, ep12, out zeta, out z, out theta);
        SofaValidate.EqualsApprox(zeta, 0.5588961642000161243e-2, 1e-12);
        SofaValidate.EqualsApprox(z, 0.5589922365870680624e-2, 1e-12);
        SofaValidate.EqualsApprox(theta, 0.4858945471687296760e-2, 1e-12);

        // DECLSPEC double iauS00(double date1, double date2, double x, double y);
        x = 0.5791308486706011000e-3;
        y = 0.4020579816732961219e-4;
        s = S00(2400000.5, 53736.0, x, y);
        SofaValidate.EqualsApprox(s, -0.1220036263270905693e-7, 1e-18);

        // DECLSPEC double iauS00a(double date1, double date2);
        SofaValidate.EqualsApprox(S00a(2400000.5, 52541.0), -0.1340684448919163584e-7, 1e-18);

        // DECLSPEC double iauS00b(double date1, double date2);
        s = S00b(2400000.5, 52541.0);
        SofaValidate.EqualsApprox(s, -0.1340695782951026584e-7, 1e-18);

        // DECLSPEC double iauS06(double date1, double date2, double x, double y);
        x = 0.5791308486706011000e-3;
        y = 0.4020579816732961219e-4;
        s = S06(2400000.5, 53736.0, x, y);
        SofaValidate.EqualsApprox(s, -0.1220032213076463117e-7, 1e-18);

        // DECLSPEC double iauS06a(double date1, double date2);
        s = S06a(2400000.5, 52541.0);
        SofaValidate.EqualsApprox(s, -0.1340680437291812383e-7, 1e-18);

        // DECLSPEC double iauSp00(double date1, double date2);
        SofaValidate.EqualsApprox(Sp00(2400000.5, 52541.0), -0.6216698469981019309e-11, 1e-12);

        // DECLSPEC void iauXy06(double date1, double date2, double *x, double *y);
        Xy06(2400000.5, 53736.0, out x, out y);
        SofaValidate.EqualsApprox(x, 0.5791308486706010975e-3, 1e-15);
        SofaValidate.EqualsApprox(y, 0.4020579816732958141e-4, 1e-16);

        // DECLSPEC void iauXys00a(double date1, double date2,
        //               double *x, double *y, double *s);
        Xys00a(2400000.5, 53736.0, out x, out y, out s);
        SofaValidate.EqualsApprox(x, 0.5791308472168152904e-3, 1e-14);
        SofaValidate.EqualsApprox(y, 0.4020595661591500259e-4, 1e-15);
        SofaValidate.EqualsApprox(s, -0.1220040848471549623e-7, 1e-18);

        // DECLSPEC void iauXys00b(double date1, double date2,
        //               double *x, double *y, double *s);
        Xys00b(2400000.5, 53736.0, out x, out y, out s);
        SofaValidate.EqualsApprox(x, 0.5791301929950208873e-3, 1e-14);
        SofaValidate.EqualsApprox(y, 0.4020553681373720832e-4, 1e-15);
        SofaValidate.EqualsApprox(s, -0.1220027377285083189e-7, 1e-18);

        // DECLSPEC void iauXys06a(double date1, double date2,
        //               double *x, double *y, double *s);
        Xys06a(2400000.5, 53736.0, out x, out y, out s);
        SofaValidate.EqualsApprox(x, 0.5791308482835292617e-3, 1e-14);
        SofaValidate.EqualsApprox(y, 0.4020580099454020310e-4, 1e-15);
        SofaValidate.EqualsApprox(s, -0.1220032294164579896e-7, 1e-18);
    }
#endif      // end self test code
}   // end partial class Sofa
}   // end namespace HirosePS