﻿// RotationAndTime.cs
// .NET adapters for the IAU SOFA RotationAndTime functions
// Copyright Paul S. Hirose 2017

// revision history
#if false
2017-09-17 Created. Originally this was part of another file, but now in a file of its own.
#endif

using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {
    /* Astronomy/RotationAndTime */


    // DECLSPEC double iauEe00(double date1, double date2, double epsa, double dpsi);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEe00(double date1, double date2, double epsa, double dpsi);

    /// <summary>
    /// The equation of the equinoxes, compatible with IAU 2000 resolutions, given the nutation in
    /// longitude and the mean obliquity.
    /// </summary>
    /// <param name="date1">TT JD part 1 (2451545.0 for max resolution)</param>
    /// <param name="date2">TT JD part 2</param>
    /// <param name="epsa">mean obliquity (radians)</param>
    /// <param name="dpsi">nutation in longitude (radians)</param>
    /// <returns>equation of the equinoxes (radians)</returns>
    public static double Ee00(double date1, double date2, double epsa, double dpsi) {
        return iauEe00(date1, date2, epsa, dpsi);
    }


    // DECLSPEC double iauEe00a(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEe00a(double date1, double date2);

    /// <summary>
    /// Equation of the equinoxes, IAU 2000A.
    /// </summary>
    /// <param name="date1">TT JD part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT JD part 2</param>
    /// <returns>equation of the equinoxes (radians)</returns>
    public static double Ee00a(double date1, double date2) {
        return iauEe00a(date1, date2);
    }


    // DECLSPEC double iauEe00b(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEe00b(double date1, double date2);

    /// <summary>
    /// Equation of the equinoxes, IAU 2000B.
    /// </summary>
    /// <param name="date1">TT JD part 1 (2451545.0 for optimum resolution)</param>
    /// <param name="date2">TT JD part 2</param>
    /// <returns>equation of the equinoxes (radians)</returns>
    public static double Ee00b(double date1, double date2) {
        return iauEe00b(date1, date2);
    }


    // DECLSPEC double iauEe06a(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEe06a(double date1, double date2);

    /// <summary>
    /// Equation of the equinoxes, compatible with IAU 2000 resolutions and
    /// IAU 2006/2000A precession nutation.
    /// </summary>
    /// <param name="date1">TT JD part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TT JD part 2</param>
    /// <returns>equation of the equinoxes (radians)</returns>
    public static double Ee06a(double date1, double date2) {
        return iauEe06a(date1, date2);
    }


    // DECLSPEC double iauEect00(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEect00(double date1, double date2);

    /// <summary>
    ///  Equation of the equinoxes complementary terms, consistent with IAU 2000 resolutions.
    /// </summary>
    /// <param name="date1">TT JD part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TT JD part 2</param>
    /// <returns>complementary terms</returns>
    public static double Eect00(double date1, double date2) {
        return iauEect00(date1, date2);
    }


    // DECLSPEC double iauEqeq94(double date1, double date2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEqeq94(double date1, double date2);

    /// <summary>
    /// Equation of the equinoxes, IAU 1994 model.
    /// </summary>
    /// <param name="date1">TDB date part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TDB date part 2</param>
    /// <returns>equation of the equinoxes (radians)</returns>
    public static double Eqeq94(double date1, double date2) {
        return iauEqeq94(date1, date2);
    }


    // DECLSPEC double iauEra00(double dj1, double dj2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauEra00(double dj1, double dj2);

    /// <summary>
    /// Earth rotation angle (IAU 2000 model).
    /// </summary>
    /// <param name="dj1">UT1 JD part 1 (at 0 h for best resolution)</param>
    /// <param name="dj2">UT1 JD part 2</param>
    /// <returns>Earth rotation angle (radians), range 0-2pi</returns>
    public static double Era00(double dj1, double dj2) {
        return iauEra00(dj1, dj2);
    }


    // DECLSPEC double iauGmst00(double uta, double utb, double tta, double ttb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGmst00(double uta, double utb, double tta, double ttb);

    /// <summary>
    /// Greenwich mean sidereal time (model consistent with IAU 2000 resolutions).
    /// </summary>
    /// <param name="uta">UT1 part 1 (at 0 h for best resolution)</param>
    /// <param name="utb">UT1 part 2</param>
    /// <param name="tta">TT part 1</param>
    /// <param name="ttb">TT part 2</param>
    /// <returns>Greenwich mean sidereal time (radians)</returns>
    public static double Gmst00(double uta, double utb, double tta, double ttb) {
        return iauGmst00(uta, utb, tta, ttb);
    }


    // DECLSPEC double iauGmst06(double uta, double utb, double tta, double ttb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGmst06(double uta, double utb, double tta, double ttb);

    /// <summary>
    /// Greenwich mean sidereal time (consistent with IAU 2006 precession).
    /// </summary>
    /// <param name="uta">>UT1 part 1 (at 0 h for best resolution)</param>
    /// <param name="utb">>UT1 part 2</param>
    /// <param name="tta">TT part 1</param>
    /// <param name="ttb">TT part 2</param>
    /// <returns>Greenwich mean sidereal time (radians)</returns>
    public static double Gmst06(double uta, double utb, double tta, double ttb) {
        return iauGmst06(uta, utb, tta, ttb);
    }


    // DECLSPEC double iauGmst82(double dj1, double dj2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGmst82(double dj1, double dj2);

    /// <summary>
    /// Universal Time to Greenwich mean sidereal time (IAU 1982 model).
    /// </summary>
    /// <param name="dj1">UT1 Julian Date part 1 (at 0h for best accuracy)</param>
    /// <param name="dj2">UT1 Julian Date part 2</param>
    /// <returns>Greenwich mean sidereal time (radians)</returns>
    public static double Gmst82(double dj1, double dj2) {
        return iauGmst82(dj1, dj2);
    }

    // DECLSPEC double iauGst00a(double uta, double utb, double tta, double ttb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGst00a(double uta, double utb, double tta, double ttb);

    /// <summary>
    /// Greenwich apparent sidereal time (consistent with IAU 2000 resolutions).
    /// </summary>
    /// <param name="uta">UT1 Julian date part 1 (at 0h for best accuracy)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <returns>Greenwich apparent sidereal time (radians)</returns>
    public static double Gst00a(double uta, double utb, double tta, double ttb) {
        return iauGst00a(uta, utb, tta, ttb);
    }


    // DECLSPEC double iauGst00b(double uta, double utb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGst00b(double uta, double utb);

    /// <summary>
    /// Greenwich apparent sidereal time (consistent with IAU 2000 resolutions but using the
    /// truncated nutation model IAU 2000B).
    /// </summary>
    /// <param name="uta">UT1 Julian date part 1 (at 0h for best accuracy)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <returns>Greenwich apparent sidereal time (radians)</returns>
    public static double Gst00b(double uta, double utb) {
        return iauGst00b(uta, utb);
    }


    // DECLSPEC double iauGst06(double uta, double utb, double tta, double ttb, double rnpb[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGst06(double uta, double utb, double tta, double ttb, double[,] rnpb);

    /// <summary>
    /// Greenwich apparent sidereal time, IAU 2006, given the NPB matrix.
    /// </summary>
    /// <param name="uta">UT1 Julian date part 1 (at 0 h for best accuracy)</param>
    /// <param name="utb">UT1 Julian date part 2></param>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <param name="rnpb">nutation x precession x bias matrix</param>
    /// <returns>Greenwich apparent sidereal time (radians)</returns>
    /// <exception cref="ArgumentException">if <paramref name="rnpb"/>does not have
    /// dimensions [3,3].</exception>
    public static double Gst06(double uta, double utb, double tta, double ttb, double[,] rnpb) {
        SofaValidate.VerifyDouble3x3(rnpb);
        return iauGst06(uta, utb, tta, ttb, rnpb);
    }


    // DECLSPEC double iauGst06a(double uta, double utb, double tta, double ttb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGst06a(double uta, double utb, double tta, double ttb);

    /// <summary>
    /// Greenwich apparent sidereal time (consistent with IAU 2000 and 2006 resolutions).
    /// </summary>
    /// <param name="uta">UT1 Julian date part 1 (at 0h for best accuracy)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <param name="tta">TT Julian date part 1</param>
    /// <param name="ttb">TT Julian date part 2</param>
    /// <returns>Greenwich apparent sidereal time (radians)</returns>
    public static double Gst06a(double uta, double utb, double tta, double ttb) {
        return iauGst06a(uta, utb, tta, ttb);
    }


    // DECLSPEC double iauGst94(double uta, double utb);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauGst94(double uta, double utb);

    /// <summary>
    /// Greenwich apparent sidereal time (consistent with IAU 1982/94 resolutions).
    /// </summary>
    /// <param name="uta">UT1 Julian date part 1 (at 0h for best accuracy)</param>
    /// <param name="utb">UT1 Julian date part 2</param>
    /// <returns>Greenwich apparent sidereal time (radians)</returns>
    public static double Gst94(double uta, double utb) {
        return iauGst94(uta, utb);
    }



    // Test the methods in this file.
#if DEBUG

    /// <summary>
    /// Self test the IAU SOFA RotationAndTime methods.
    /// </summary>
    public static void TestRotationAndTime(){
        double[,] rnpb;

        /* Astronomy/RotationAndTime */

        // DECLSPEC double iauEe00(double date1, double date2, double epsa, double dpsi);
        SofaValidate.EqualsApprox(Ee00(2400000.5, 53736.0, 0.4090789763356509900,
            -0.9630909107115582393e-5), -0.8834193235367965479e-5, 1e-18);

        // DECLSPEC double iauEe00a(double date1, double date2);
        SofaValidate.EqualsApprox(Ee00a(2400000.5, 53736.0), -0.8834192459222588227e-5, 1e-18);

        // DECLSPEC double iauEe00b(double date1, double date2);
        SofaValidate.EqualsApprox(Ee00b(2400000.5, 53736.0), -0.8835700060003032831e-5, 1e-18);

        // DECLSPEC double iauEe06a(double date1, double date2);
        SofaValidate.EqualsApprox(Ee06a(2400000.5, 53736.0), -0.8834195072043790156e-5, 1e-15);

        // DECLSPEC double iauEect00(double date1, double date2);
        SofaValidate.EqualsApprox(Eect00(2400000.5, 53736.0), 0.2046085004885125264e-8, 1e-20);

        // DECLSPEC double iauEqeq94(double date1, double date2);
        SofaValidate.EqualsApprox(Eqeq94(2400000.5, 41234.0), 0.5357758254609256894e-4, 1e-17);

        // DECLSPEC double iauEra00(double dj1, double dj2);
        SofaValidate.EqualsApprox(Era00(2400000.5, 54388.0), 0.4022837240028158102, 1e-12);

        // DECLSPEC double iauGmst00(double uta, double utb, double tta, double ttb);
        SofaValidate.EqualsApprox(Gmst00(2400000.5, 53736.0, 2400000.5, 53736.0),
            1.754174972210740592, 1e-12);

        // DECLSPEC double iauGmst06(double uta, double utb, double tta, double ttb);
        SofaValidate.EqualsApprox(Gmst06(2400000.5, 53736.0, 2400000.5, 53736.0),
            1.754174971870091203, 1e-12);

        // DECLSPEC double iauGmst82(double dj1, double dj2);
        SofaValidate.EqualsApprox(Gmst82(2400000.5, 53736.0), 1.754174981860675096, 1e-12);

        // DECLSPEC double iauGst00a(double uta, double utb, double tta, double ttb);
        SofaValidate.EqualsApprox(Gst00a(2400000.5, 53736.0, 2400000.5, 53736.0),
            1.754166138018281369, 1e-12);

        // DECLSPEC double iauGst00b(double uta, double utb);
        SofaValidate.EqualsApprox(Gst00b(2400000.5, 53736.0), 1.754166136510680589, 1e-12);

        // DECLSPEC double iauGst06(double uta, double utb, double tta, double ttb,
        //                double rnpb[3][3]);
        rnpb = new double[,]{
            {0.9999989440476103608, -0.1332881761240011518e-2, -0.5790767434730085097e-3},
            {0.1332858254308954453e-2, 0.9999991109044505944, -0.4097782710401555759e-4},
            { 0.5791308472168153320e-3, 0.4020595661593994396e-4, 0.9999998314954572365}
        };
        SofaValidate.EqualsApprox(Gst06(2400000.5, 53736.0, 2400000.5, 53736.0, rnpb),
            1.754166138018167568, 1e-12);

        // DECLSPEC double iauGst06a(double uta, double utb, double tta, double ttb);
        SofaValidate.EqualsApprox(Gst06a(2400000.5, 53736.0, 2400000.5, 53736.0),
            1.754166137675019159, 1e-12);

        // DECLSPEC double iauGst94(double uta, double utb);
        SofaValidate.EqualsApprox(Gst94(2400000.5, 53736.0), 1.754166136020645203, 1e-12);

    }   // end self test
#endif
}   // end partial class Sofa
}   // end namespace HirosePS
