﻿// SofaAstrom.cs
// IAU SOFA Library struct "iauASTROM", translated to the .NET Framework CLR.
// Copyright 2017 by Paul S. Hirose

// revision history
#if false
2014-07-27  Began work.
2014-08-06  Added attributes so the class can interoperate with the C++ SOFA routines.
2014-08-08  Added XML comments to public fields, and a Set method for the BPN matrix.
2014-10-03  SetBpn() wasn't checking its array argument for correct dimensions.

2017-09-10 Building SofaJpl 2.1.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {

/// <summary>
/// Star-independent astrometry parameters
/// </summary>
[StructLayout(LayoutKind.Sequential)]
public class SofaAstrom {

    // double pmt;        /* PM time interval (SSB, Julian years) */
    /// <summary>
    /// PM time interval (SSB, Julian years)
    /// </summary>
    public double Pmt;

    // double eb[3];      /* SSB to observer (vector, au) */
    /* (Vectors eb, eh, em and v are all with respect to BCRS axes.) */
    /// <summary>
    /// SSB to observer vector (au, in the BCRS)
    /// </summary>
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
    public double[] Eb;

    // double eh[3];      /* Sun to observer (unit vector) */
    /// <summary>
    /// Sun to observer vector (unit vector in the BCRS)
    /// </summary>
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
    public double[] Eh;

    // double em;         /* distance from Sun to observer (au) */
    /// <summary>
    /// distance from Sun to observer (au)
    /// </summary>
    public double Em;

    // double v[3];       /* barycentric observer velocity (vector, c) */
    /// <summary>
    /// barycentric observer velocity vector (c, in the BCRS)
    /// </summary>
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
    public double[] V;

    // double bm1;        /* sqrt(1-|v|^2): reciprocal of Lorenz factor */
    /// <summary>
    /// sqrt(1-|v|^2): reciprocal of Lorenz factor
    /// </summary>
    public double Bm1;

    // double bpn[3][3];  /* bias-precession-nutation matrix */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 9)]
    private double[] _bpn;

    // double along;      /* longitude + s' + dERA(DUT) (radians) */
    /// <summary>
    /// longitude + s' + dERA(DUT) (radians)
    /// </summary>
    public double Along;

    // double phi;        /* geodetic latitude (radians) */
    /// <summary>
    /// geodetic latitude (radians)
    /// </summary>
    public double Phi;

    // double xpl;        /* polar motion xp wrt local meridian (radians) */
    /// <summary>
    /// polar motion xp wrt local meridian (radians)
    /// </summary>
    public double Xpl;

    // double ypl;        /* polar motion yp wrt local meridian (radians) */
    /// <summary>
    /// polar motion yp wrt local meridian (radians)
    /// </summary>
    public double Ypl;

    // double sphi;       /* sine of geodetic latitude */
    /// <summary>
    /// sine of geodetic latitude
    /// </summary>
    public double Sphi;

    // double cphi;       /* cosine of geodetic latitude */
    /// <summary>
    /// cosine of geodetic latitude
    /// </summary>
    public double Cphi;

    // double diurab;     /* magnitude of diurnal aberration vector */
    /// <summary>
    /// magnitude of diurnal aberration vector
    /// </summary>
    public double Diurab;

    // double eral;       /* "local" Earth rotation angle (radians) */
    /// <summary>
    /// "local" Earth rotation angle (radians)
    /// </summary>
    public double Eral;

    // double refa;       /* refraction constant A (radians) */
    /// <summary>
    /// refraction constant A (radians)
    /// </summary>
    public double Refa;

    // double refb;       /* refraction constant B (radians) */
    /// <summary>
    /// refraction constant B (radians)
    /// </summary>
    public double Refb;


    /// <summary>
    /// Construct a new instance with all fields zero.
    /// </summary>
    public SofaAstrom() {
        Eb = new double[3];
        Eh = new double[3];
        V = new double[3];
        _bpn = new double[9];
    }


    /// <summary>
    /// Get the bias precession nutation matrix.
    /// </summary>
    /// <param name="bpn">output: bias precession nutation matrix</param>
    /// <exception cref="ArgumentException">if argument is null or not dimensioned [3,3]
    /// </exception>
    public void GetBpn(double[,] bpn) {
        SofaValidate.VerifyDouble3x3(bpn);
        bpn[0, 0] = _bpn[0];    bpn[0, 1] = _bpn[1];    bpn[0, 2] = _bpn[2];
        bpn[1, 0] = _bpn[3];    bpn[1, 1] = _bpn[4];    bpn[1, 2] = _bpn[5];
        bpn[2, 0] = _bpn[6];    bpn[2, 1] = _bpn[7];    bpn[2, 2] = _bpn[8];
    }


    /// <summary>
    /// Set the bias precession nutation matrix.
    /// </summary>
    /// <param name="bpn">Input: bias precession nutation matrix</param>
    /// <exception cref="ArgumentException">if argument is null or not dimensioned [3,3]
    /// </exception>
    public void SetBpn(double[,] bpn) {
        SofaValidate.VerifyDouble3x3(bpn);
        _bpn[0] = bpn[0, 0]; _bpn[1] = bpn[0, 1]; _bpn[2] = bpn[0, 2];
        _bpn[3] = bpn[1, 0];    _bpn[4] = bpn[1, 1];    _bpn[5] = bpn[1, 2];
        _bpn[6] = bpn[2, 0];    _bpn[7] = bpn[2, 1];    _bpn[8] = bpn[2, 2];
    }

}   // end class SofaAstrom
}       // end namespace HirosePS
