﻿// SofaLDBody.cs
// IAU SOFA Library struct "iauLDBODY", translated to the .NET Framework CLR.
// Copyright 2017 Paul S. Hirose

// revision history
#if false
2014-08-08  Created.
2014-08-16  Changed from class to struct due to trouble passing an array of instances to the
unmanaged SOFA code.
2014-10-03  Added XML comment for exception when array has wrong dimensions.

2017-09-10 Building SofaJpl 2.1.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {

/// <summary>
/// Body parameters for light deflection
/// </summary>
[StructLayout(LayoutKind.Sequential)]
public struct SofaLDBody {

   // double bm;         /* mass of the body (solar masses) */
    /// <summary>
    /// mass of the body (solar masses)
    /// </summary>
    public double Bm;

   // double dl;         /* deflection limiter (radians^2/2) */
    /// <summary>
    /// deflection limiter (radians^2/2)
    /// </summary>
    public double Dl;

   // double pv[2][3];   /* barycentric PV of the body (au, au/day) */
    double pv00, pv01, pv02, pv10, pv11, pv12;


    /// <summary>
    /// Construct a new instance, given mass, deflection limiter, and
    /// barycentric position / velocity.
    /// </summary>
    /// <param name="bm">mass of the body (solar masses)</param>
    /// <param name="dl">deflection limiter (radians^2/2)</param>
    /// <param name="pv">barycentric position / velocity of the body (au, au/day)</param>
    /// <exception cref="ArgumentException">if <paramref name="pv"/> is not dimensioned [2,3]
    /// </exception>
    public SofaLDBody(double bm, double dl, double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        Bm = bm;
        Dl = dl;
        pv00 = pv[0, 0];    pv01 = pv[0, 1];    pv02 = pv[0, 2];
        pv10 = pv[1, 0];    pv11 = pv[1, 1];    pv12 = pv[1, 2];
    }


    /// <summary>
    /// Get the barycentric position and velocity of the body.
    /// </summary>
    /// <param name="pv">barycentric PV (au, au/day)</param>
    /// <exception cref="ArgumentException">if argument is null or not dimensioned [2,3]
    /// </exception>
    public void GetPv(double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        pv[0, 0] = pv00;  pv[0, 1] = pv01;  pv[0, 2] = pv02;
        pv[1, 0] = pv10;  pv[1, 1] = pv11;  pv[1, 2] = pv12;
    }


    /// <summary>
    /// Set the barycentric position and velocity of the body.
    /// </summary>
    /// <param name="pv">barycentric PV (au, au/day)</param>
    /// <exception cref="ArgumentException">if argument is null or not dimensioned [2,3]
    /// </exception>
    public void SetPv(double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        pv00 = pv[0, 0]; pv01 = pv[0, 1]; pv02 = pv[0, 2];
        pv10 = pv[1, 0]; pv11 = pv[1, 1]; pv12 = pv[1, 2];
    }

}   // end struct SofaDbody
}   // end namespace HirosePS
