﻿// SofaValidate.cs
// Static class to validate methods and arguments in the Sofa class.
// Copyright 2017 by Paul S. Hirose

// revision history
#if false
2014-08-07  Created from code formerly embedded in the Sofa class. Runs OK.
2014-09-03  Added EqualsApprox overload with separate tolerance for diagonal elements of a
square matrix. Test is good.
2015-08-27  Modified TestConstants() to a void method which throws an exception if error.

2017-09-10 Building SofaJpl 2.1.
#endif

using System;
using System.Runtime.InteropServices;


namespace HirosePS {
static class SofaValidate {
    /// <summary>
    /// Verify two values don't differ by more than a given tolerance.
    /// </summary>
    /// <param name="a">left hand value</param>
    /// <param name="b">right hand value</param>
    /// <param name="tolerance">max allowable difference (absolute value)</param>
    /// <exception cref="Exception"> if tolerance is exceeded.</exception>
    /// <exception cref="ArgumentException"> if tolerance is negative</exception>
    public static void EqualsApprox(double a, double b, double tolerance) {
        if (tolerance < 0.0)
            throw new ArgumentException("Negative tolerance not allowed");
        double error = Math.Abs(a - b);
        if (error > tolerance)
            throw new Exception("out of tolerance");
    }


    /// <summary>
    /// Verify all elements of two arrays agree within a given tolerance.
    /// </summary>
    /// <param name="a">one array</param>
    /// <param name="b">other array</param>
    /// <param name="tolerance">max allowable difference (absolute value)</param>
    /// <exception cref="Exception"> if tolerance is exceeded.</exception>
    /// <exception cref="ArgumentException">if either array is null, or arrays have different
    /// length.</exception>
    public static void EqualsApprox(double[] a, double[] b, double tolerance) {
        if (a == null || b == null)
            throw new ArgumentException("Null argument.");
        if (a.Length != b.Length)
            throw new ArgumentException("Arrays have different length.");
        for (int i = 0; i < a.Length; ++i)
            EqualsApprox(a[i], b[i], tolerance);
    }


    /// <summary>
    /// Verify all elements of two arrays agree within a given tolerance.
    /// </summary>
    /// <param name="a">one array</param>
    /// <param name="b">other array</param>
    /// <param name="tolerance">max allowable difference (absolute value)</param>
    /// <exception cref="Exception"> if tolerance is exceeded.</exception>
    /// <exception cref="ArgumentException">if either array is null, or arrays have different
    /// dimensions.</exception>
    public static void EqualsApprox(double[,] a, double[,] b, double tolerance) {
        if (a == null || b == null)
            throw new ArgumentException("Neither array can be null.");
        int rows = b.GetLength(0);
        int columns = b.GetLength(1);
        if (a.GetLength(0) != rows || a.GetLength(1) != columns)
            throw new ArgumentException("Array dimensions aren't equal.");
        for (int row = 0; row < rows; ++row)
            for (int column = 0; column < columns; ++column)
                EqualsApprox(a[row, column], b[row, column], tolerance);
    }


    /// <summary>
    /// Verify two arrays agree, with separate tolerances for the diagonal, or top and bottom rows
    /// if matrices are [2,3].
    /// </summary>
    /// <param name="a">one array</param>
    /// <param name="b">other array</param>
    /// <param name="tolerance1">max allowable difference (absolute value) for diagonal or top row
    /// </param>
    /// <param name="tolerance2">max allowable difference (absolute value) not on the
    /// diagonal, or bottom row if 2x3, otherwise all elements</param>
    /// <exception cref="Exception"> if tolerance is exceeded.</exception>
    /// <exception cref="ArgumentException">if either array is null, or arrays have different
    /// dimensions.</exception>
    public static void EqualsApprox(double[,] a, double[,] b, double tolerance1, double tolerance2)
    {
        if (a == null || b == null)
            throw new ArgumentException("Neither array can be null.");
        int rows = b.GetLength(0);
        int columns = b.GetLength(1);
        if (a.GetLength(0) != rows || a.GetLength(1) != columns)
            throw new ArgumentException("Array dimensions aren't equal.");
        if (rows == columns)
            // matrices are square
            for (int row = 0; row < rows; ++row)
                for (int column = 0; column < columns; ++column)
                    if (row == column)
                        // current element is on the diagonal
                        EqualsApprox(a[row, column], b[row, column], tolerance1);
                    else
                        EqualsApprox(a[row, column], b[row, column], tolerance2);
        else if (rows == 2 && columns == 3)
            for (int row = 0; row < rows; ++row)
                for (int column = 0; column < columns; ++column)
                    if (row == 0)
                        EqualsApprox(a[row, column], b[row, column], tolerance1);
                    else
                        EqualsApprox(a[row, column], b[row, column], tolerance2);
        else
            // matrices are neither square nor 2x3, so tolerance2 applies to all elements
            for (int row = 0; row < rows; ++row)
                for (int column = 0; column < columns; ++column)
                    EqualsApprox(a[row, column], b[row, column], tolerance2);

    }


    /// <summary>
    /// Verify an int array is not null and has length 4.
    /// </summary>
    /// <param name="a">the array to check</param>
    /// <exception cref="ArgumentException">if null or wrong length.</exception>
    public static void VerifyInt1x4(int[] a) {
        if (a == null)
            throw new ArgumentException("Array must not be null.");
        if (a.Length != 4)
            throw new ArgumentException("Array must have length 4.");
    }


    /// <summary>
    /// Verify a double array is not null and has length 3.
    /// </summary>
    /// <param name="a">the array to check</param>
    /// <exception cref="ArgumentException">if null or wrong length.</exception>
    public static void VerifyDouble1x3(double[] a) {
        if (a == null)
            throw new ArgumentException("Array must not be null.");
        if (a.Length != 3)
            throw new ArgumentException("Array must have length 3.");
    }


    /// <summary>
    /// Verify a double array is not null and has dimensions 2 by 3.
    /// </summary>
    /// <param name="a">the array to check</param>
    /// <exception cref="ArgumentException">if null or wrong dimensions.</exception>
    public static void VerifyDouble2x3(double[,] a) {
        if (a == null)
            throw new ArgumentException("Array must not be null.");
        if (a.GetLength(0) != 2 || a.GetLength(1) != 3)
            throw new ArgumentException("Array dimensions must be [2,3]");
    }


    /// <summary>
    /// Verify a double array is not null and has dimensions 3 by 3.
    /// </summary>
    /// <param name="a">the array to check</param>
    /// <exception cref="ArgumentException">if null or wrong dimensions.</exception>
    public static void VerifyDouble3x3(double[,] a) {
        if (a == null)
            throw new ArgumentException("Array must not be null.");
        if (a.GetLength(0) != 3 || a.GetLength(1) != 3)
            throw new ArgumentException("Array dimensions must be [3,3]");
    }


    // replaced by modified version below
#if false
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int TestConstants(sbyte* dest, int len);

    /// <summary>
    /// Verify the #defined constants in SOFA header.
    /// </summary>
    public static void TestSofaConstants() {
        int len = 100;
        sbyte[] bytes = new sbyte[len];
        int error = 0;
        fixed (sbyte* bytes_ptr = bytes) {
            error = TestConstants(bytes_ptr, len);
            if (error != 0) {
                Console.WriteLine("Error = {0}", new string(bytes_ptr));
                throw new Exception();
            }
        }
    }
#endif


    // modified version, itself replaced by a direct call from the main program
#if false
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void TestConstants();

    /// <summary>
    /// Verify the #defined constants in SOFA header.
    /// </summary>
    public static void TestSofaConstants() {
        TestConstants();
    }
#endif


    public static void TestTest() {
        double[] vec1, vec2;
        double[,] pvvec1, rnpb;
        double[,] correctPvvec = new double[,] { { 0.0, 0.1, 0.2 }, { 1.0, 1.1, 1.2 } };
        // all elements .01 greater than nominal
        double[,] pvvecPlus01 = new double[,] { { 0.01, 0.11, 0.21 }, { 1.01, 1.11, 1.21 } };
        double[,] correctRmat =
            new double[,] { { 0.0, 0.1, 0.2 }, { 1.0, 1.1, 1.2 }, { 2.0, 2.1, 2.2 } };
        // all elements on diagonal .01 greater than nominal, others .001 greater
        double[,] rmatPlus01Plus001 = new double[,] {
            { 0.01, 0.101, 0.201 }, { 1.001, 1.11, 1.201 }, { 2.001, 2.101, 2.21 } };
        double[,] correctThreeByTwo = { { 0.0, 0.1 }, { 1.0, 1.1 }, { 2.0, 2.1 } };
        double[,] threeByTwoPlus01 = { { 0.01, 0.11 }, { 1.01, 1.11 }, { 2.01, 2.11 } };
        int[] ymdf;
        bool error;


        // equalsApprox(), scalar overload

        EqualsApprox(1.5, 1.55, .1);        // should not throw exception here
        error = false;
        try {
            EqualsApprox(1.5, 1.7, .1);     // should throw exception
            error = true;                   // Should not reach this statement.
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;            // EqualsApprox() failed to throw expected exception
        }


        // EqualsApprox(), array overload

        // Verify an exception is thrown if the arrays are null.
        vec1 = null;
        vec2 = null;
        error = false;
        try {
            EqualsApprox(vec1, vec2, 0.0);      // should throw exception
            error = false;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }

        // Verify an exception if one array is null.
        vec1 = new double[] { 1.0, 2.0, 3.0 };
        try {
            EqualsApprox(vec1, vec2, 0.0);      // should throw exception
            error = false;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }

        // Verify an exception if other array is null.
        vec2 = vec1;
        vec1 = null;
        try {
            EqualsApprox(vec1, vec2, 0.0);      // should throw exception
            error = false;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }

        // Verify an exception if arrays are differnt lengths.
        vec1 = new double[] { 1.0, 2.0 };
        vec2 = new double[] { 1.0, 2.0, 3.0 };
        try {
            EqualsApprox(vec1, vec2, 0.0);      // should throw exception
            error = false;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }


        // Verify no exception if discrepancies are within tolerance.
        vec1 = new double[] { 1.0, 2.0, 2.91 };
        vec2 = new double[] { 1.0, 2.0, 3.0 };
        EqualsApprox(vec1, vec2, .1);


        // Verify an exception if discrepancies are out of tolerance.
        vec1 = new double[] { 1.0, 2.0, 2.8 };
        vec2 = new double[] { 1.0, 2.0, 3.0 };
        try {
            EqualsApprox(vec1, vec2, .1);      // should throw exception
            error = false;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }


        // EqualsApprox(double[,], double[,], tolerance)
        // First array is null. Confirm exception.
        error = false;
        try {
            EqualsApprox(null, correctPvvec, .1);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Second array is null. Confirm exception.
        try {
            EqualsApprox(correctPvvec, null, .1);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays have different dimensions. Confirm exception.
        try {
            EqualsApprox(correctPvvec, correctRmat, 0.0);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays out of tolerance. Confirm exception.
        pvvec1 = new double[,] { { 0.0, 0.1, 0.2 }, { 1.0, 1.1, 1.3 } };
        try {
            EqualsApprox(pvvec1, correctPvvec, .01);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays within tolerance. Confirm no exception.
        pvvec1 = new double[,] { { 0.01, 0.11, 0.21 }, { 1.01, 1.11, 1.21 } };
        EqualsApprox(pvvec1, correctPvvec, .02);


        // EqualsApprox(double[,], double[,], tolerance1, tolerance2)
        // First array is null. Confirm exception.
        error = false;
        try {
            EqualsApprox(null, correctPvvec, .1, .01);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Second array is null. Confirm exception.
        try {
            EqualsApprox(correctPvvec, null, .1, .01);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays have different dimensions. Confirm exception.
        try {
            EqualsApprox(correctPvvec, correctRmat, .1, .01);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays not square, not 2x3, and in tolerance. Confirm tolerance1 has no effect.
        EqualsApprox(threeByTwoPlus01,correctThreeByTwo, 0.0, .02);
        // Arrays not square and not 2x3. Confirm tolerance2 has control.
        try {
            EqualsApprox(threeByTwoPlus01, correctThreeByTwo, 0.0, .005);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays are square and a diagonal element is out of tolerance.
        try {
            EqualsApprox(rmatPlus01Plus001, correctRmat, .005, .002);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays are square and a non-diagonal element is out of tolerance.
        try {
            EqualsApprox(rmatPlus01Plus001, correctRmat, .02, .0005);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // Arrays are square and within tolerance.
        EqualsApprox(rmatPlus01Plus001, correctRmat, .02, .002);
        // 2x3 array, top row out of tolerance
        try {
            EqualsApprox(pvvecPlus01, correctPvvec, .005, .02);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // 2x3 array, bottom row out of tolerance
        try {
            EqualsApprox(pvvecPlus01, correctPvvec, .02, .005);
            error = true;
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;
        }
        // 2x3 array, in tolerance
        EqualsApprox(pvvecPlus01, correctPvvec, .02, .02);

        // VerifyInt1x4()

        // test for null array
        ymdf = null;
        error = false;
        try {
            VerifyInt1x4(ymdf);     // should throw exception
            error = true;           // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;            // expected exception not thrown
        }

        // Test for wrong array length.
        ymdf = new int[3];
        error = false;
        try {
            VerifyInt1x4(ymdf);     // should throw exception
            error = true;           // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;            // expected exception not thrown
        }

        // Test correct array length.
        ymdf = new int[4];
        VerifyInt1x4(ymdf);


        // VerifyDouble1x3()

        // test for null array
        vec1 = null;
        error = false;
        try {
            VerifyDouble1x3(vec1);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test for wrong array length.
        vec1 = new double[2];
        error = false;
        try {
            VerifyDouble1x3(vec1);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test correct array length.
        vec1 = new double[3];
        VerifyDouble1x3(vec1);


        // VerifyDouble2x3()

        // test for null array
        pvvec1 = null;
        error = false;
        try {
            VerifyDouble2x3(pvvec1);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test for wrong number of rows.
        pvvec1 = new double[3, 3];
        error = false;
        try {
            VerifyDouble2x3(pvvec1);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test for wrong number of columns.
        pvvec1 = new double[2, 2];
        error = false;
        try {
            VerifyDouble2x3(pvvec1);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test correct array dimensions.
        pvvec1 = new double[2, 3];
        VerifyDouble2x3(pvvec1);


        // VerifyDouble3x3()

        // test for null array
        rnpb = null;
        error = false;
        try {
            VerifyDouble3x3(rnpb);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test for wrong number of rows.
        rnpb = new double[2, 3];
        error = false;
        try {
            VerifyDouble3x3(rnpb);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test for wrong number of columns.
        rnpb = new double[3, 2];
        error = false;
        try {
            VerifyDouble3x3(rnpb);      // should throw exception
            error = true;               // should not reach this statement
            throw new Exception();
        }
        catch (Exception e) {
            if (error)
                throw e;                // expected exception not thrown
        }

        // Test correct array dimensions.
        rnpb = new double[3, 3];
        VerifyDouble3x3(rnpb);

    }
}   // end class SofaValidate
}   // end namespace HirosePS
