﻿// Timescales.cs
// .NET adapters for the IAU SOFA Timescales routines.
// Copyright Paul S. Hirose, 2017

// revision notes
#if false
2014-07-05  Began work.
2014-07-07  Cal2jd() written and tested OK.
2014-07-21  All methods written and self test OK. Still need to review XML comments for errors.
2014-07-25  Finished.
2014-08-31  One test called the SOFA routine instead of mine.
2014-10-03  Verified dimension checks for all array arguments.
2014-10-06  Verified no blank fields in XML comments. Optimum JD format is noted when it's
significant. Arrays tested for correct dimensions. Self tests call my methods, not SOFA.
2014-10-07  Excluded UTC methods from the release build.

2017-09-10 Building SofaJpl 2.1.
2017-09-17 Moved the Calendars and RotationAndTime methods into their own files. Minor
improvements to XML comments.
#endif

// Anything with UTC is excluded from the production build so I don't have to rebuild the product
// every time there's a leap second. My classes provide their own UTC facilities.


using System;
using System.Runtime.InteropServices;


namespace HirosePS
{
public static partial class Sofa
{
    /* Astronomy/Timescales */


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauD2dtf(const char *scale, int ndp, double d1, double d2,
    // int *iy, int *im, int *id, int ihmsf[4]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauD2dtf(byte[] scale, int ndp, double d1, double d2,
        out int iy, out int im, out int id, int[] ihmsf);

    /// <summary>
    /// Format for output a 2-part Julian Date (or in the case of UTC a quasi-JD form that includes
    /// special provision for leap seconds).
    /// </summary>
    /// <param name="scale">only "UTC" is recognized</param>
    /// <param name="ndp">number of decimal places in the seconds field</param>
    /// <param name="d1">Julian date part 1</param>
    /// <param name="d2">Julian date part 2</param>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day</param>
    /// <param name="ihmsf">hours, minutes, seconds, fraction</param>
    /// <returns>status; 0 = OK</returns>
    /// <exception cref="ArgumentException">if<paramref name="ihmsf"/>length is not 4.</exception>
    public static int D2dtf(string scale, int ndp, double d1, double d2,
    out int iy, out int im, out int id, int[] ihmsf) {
        SofaValidate.VerifyInt1x4(ihmsf);
        System.Text.ASCIIEncoding ascii = new System.Text.ASCIIEncoding();
        byte[] bytes = ascii.GetBytes(scale);
        return iauD2dtf(bytes, ndp, d1, d2, out iy, out im, out id, ihmsf);
    }
#endif


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauDat(int iy, int im, int id, double fd, double *deltat);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauDat(int iy, int im, int id, double fd, out double deltat);

    /// <summary>
    /// For a given UTC date, calculate delta(AT) = TAI-UTC.
    /// </summary>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day</param>
    /// <param name="fd">fractional day</param>
    /// <param name="deltat">TAI minus UTC, seconds</param>
    /// <returns>status; 0 = OK</returns>
    public static int Dat(int iy, int im, int id, double fd, out double deltat) {
        return iauDat(iy, im, id, fd, out deltat);
    }
#endif


    // DECLSPEC double iauDtdb(double date1, double date2, double ut, double elong, double u,
    // double v);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern double iauDtdb(double date1, double date2, double ut, double elong, double u,
    double v);

    /// <summary>
    /// An approximation to TDB-TT, the difference between barycentric dynamical time and
    /// terrestrial time, for an observer on the Earth.
    /// </summary>
    /// <param name="date1">TDB Julian date part 1 (2451545.0 for best resolution)</param>
    /// <param name="date2">TDB Julian date part 2</param>
    /// <param name="ut">UT1, fraction of one day</param>
    /// <param name="elong">east longitude (radians)</param>
    /// <param name="u">distance from Earth spin axis (km)</param>
    /// <param name="v">distance north of equatorial plane (km)</param>
    /// <returns>TDB-TT (seconds)</returns>
    /// <remarks>Although the date is, formally, barycentric dynamical time (TDB), terrestrial
    /// dynamical time (TT) can be used with no practical effect on accuracy.</remarks>
    public static double Dtdb(double date1, double date2, double ut, double elong, double u,
    double v) {
        return iauDtdb(date1, date2, ut, elong, u, v);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauDtf2d(const char *scale, int iy, int im, int id,
    //             int ihr, int imn, double sec, double *d1, double *d2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauDtf2d(byte[] scale, int iy, int im, int id,
    int ihr, int imn, double sec, out double d1, out double d2);

    /// <summary>
    /// Encode date and time fields into 2-part Julian Date.
    /// </summary>
    /// <param name="scale">Only "UTC" is significant.</param>
    /// <param name="iy">year</param>
    /// <param name="im">month</param>
    /// <param name="id">day</param>
    /// <param name="ihr">hour</param>
    /// <param name="imn">minute</param>
    /// <param name="sec">second</param>
    /// <param name="d1">Julian date part 1</param>
    /// <param name="d2">Julian date part 2</param>
    /// <returns>0 = OK</returns>
    public static int Dtf2d(string scale, int iy, int im, int id, int ihr, int imn, double sec,
        out double d1, out double d2) {
        System.Text.ASCIIEncoding ascii = new System.Text.ASCIIEncoding();
        byte[] bytes = ascii.GetBytes(scale);
        return iauDtf2d(bytes, iy, im, id, ihr, imn, sec, out d1, out d2);
    }
#endif


    // DECLSPEC int iauTaitt(double tai1, double tai2, double *tt1, double *tt2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    static extern int iauTaitt(double tai1, double tai2, out double tt1, out double tt2);

    /// <summary>
    /// Convert International Atomic Time (TAI) to Terrestrial Time (TT).
    /// </summary>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Taitt(double tai1, double tai2, out double tt1, out double tt2) {
        return iauTaitt(tai1, tai2, out tt1, out tt2);
    }


    // DECLSPEC int iauTaiut1(double tai1, double tai2, double dta,
    //              double *ut11, double *ut12);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTaiut1(double tai1, double tai2, double dta,
        out double ut11, out double ut12);

    /// <summary>
    /// Convert International Atomic Time (TAI) to Universal Time (UT1).
    /// </summary>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <param name="dta">UT1-TAI (seconds)</param>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Taiut1(double tai1, double tai2, double dta,
        out double ut11, out double ut12) {
            return iauTaiut1(tai1, tai2, dta, out ut11, out  ut12);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauTaiutc(double tai1, double tai2, double *utc1, double *utc2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTaiutc(double tai1, double tai2, out double utc1, out double utc2);

    /// <summary>
    /// Convert International Atomic Time (TAI) to Coordinated Universal Time (UTC).
    /// </summary>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <param name="utc1">UTC Julian date part 1</param>
    /// <param name="utc2">UTC Julian date part 2</param>
    /// <returns>0 = OK</returns>
    public static int Taiutc(double tai1, double tai2, out double utc1, out double utc2) {
        return iauTaiutc(tai1, tai2, out utc1, out utc2);
    }
#endif

    
    // DECLSPEC int iauTcbtdb(double tcb1, double tcb2, double *tdb1, double *tdb2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTcbtdb(double tcb1, double tcb2, out double tdb1, out double tdb2);

    /// <summary>
    /// Convert Barycentric Coordinate Time (TCB) to Barycentric Dynamical Time (TDB).
    /// </summary>
    /// <param name="tcb1">TCB Julian date part 1</param>
    /// <param name="tcb2">TCB Julian date part 2</param>
    /// <param name="tdb1">TDB Julian date part 1</param>
    /// <param name="tdb2">TDB Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tcbtdb(double tcb1, double tcb2, out double tdb1, out double tdb2) {
        return iauTcbtdb(tcb1, tcb2, out tdb1, out tdb2);
    }

    
    // DECLSPEC int iauTcgtt(double tcg1, double tcg2, double *tt1, double *tt2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTcgtt(double tcg1, double tcg2, out double tt1, out double tt2);

    /// <summary>
    /// Convert Geocentric Coordinate Time (TCG) to Terrestrial Time (TT).
    /// </summary>
    /// <param name="tcg1">TCG Julian date part 1</param>
    /// <param name="tcg2">TCG Julian date part 2</param>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tcgtt(double tcg1, double tcg2, out double tt1, out double tt2) {
        return iauTcgtt(tcg1, tcg2, out tt1, out tt2);
    }


    // DECLSPEC int iauTdbtcb(double tdb1, double tdb2, double *tcb1, double *tcb2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTdbtcb(double tdb1, double tdb2, out double tcb1, out double tcb2);

    /// <summary>
    /// Convert Barycentric Dynamical Time (TDB) to Barycentric Coordinate Time (TCB).
    /// </summary>
    /// <param name="tdb1">TDB Julian date part 1</param>
    /// <param name="tdb2">TDB Julian date part 2</param>
    /// <param name="tcb1">TCB Julian date part 1</param>
    /// <param name="tcb2">TCB Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tdbtcb(double tdb1, double tdb2, out double tcb1, out double tcb2) {
        return iauTdbtcb(tdb1, tdb2, out tcb1, out  tcb2);
    }


    // DECLSPEC int iauTdbtt(double tdb1, double tdb2, double dtr,
    //             double *tt1, double *tt2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTdbtt(double tdb1, double tdb2, double dtr,
        out double tt1, out double tt2);

    /// <summary>
    /// Convert Barycentric Dynamical Time (TDB) to Terrestrial Time (TT).
    /// </summary>
    /// <param name="tdb1">TDB Julian date part 1</param>
    /// <param name="tdb2">TDB Julian date part 2</param>
    /// <param name="dtr">TDB-TT in seconds</param>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tdbtt(double tdb1, double tdb2, double dtr,
        out double tt1, out double tt2) {
            return iauTdbtt(tdb1, tdb2, dtr, out  tt1, out  tt2);
    }


    // DECLSPEC int iauTttai(double tt1, double tt2, double *tai1, double *tai2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTttai(double tt1, double tt2, out double tai1, out double tai2);

    /// <summary>
    /// Convert Terrestrial Time (TT) to International Atomic Time (TAI).
    /// </summary>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tttai(double tt1, double tt2, out double tai1, out double tai2) {
        return iauTttai(tt1, tt2, out tai1, out tai2);
    }


    // DECLSPEC int iauTttcg(double tt1, double tt2, double *tcg1, double *tcg2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTttcg(double tt1, double tt2, out double tcg1, out double tcg2);

    /// <summary>
    /// Convert Terrestrial Time (TT) to Geocentric Coordinate Time (TCG).
    /// </summary>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <param name="tcg1">TCG Julian date part 1</param>
    /// <param name="tcg2">TCG Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tttcg(double tt1, double tt2, out double tcg1, out double tcg2) {
        return iauTttcg(tt1, tt2, out tcg1, out tcg2);
    }


    // DECLSPEC int iauTttdb(double tt1, double tt2, double dtr,
    //             double *tdb1, double *tdb2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTttdb(double tt1, double tt2, double dtr,
        out double tdb1, out double tdb2);

    /// <summary>
    /// Convert Terrestrial Time, TT, to Barycentric Dynamical Time, TDB.
    /// </summary>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <param name="dtr">TDB-TT in seconds</param>
    /// <param name="tdb1">TDB Julian date part 1</param>
    /// <param name="tdb2">TDB Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Tttdb(double tt1, double tt2, double dtr,
        out double tdb1, out double tdb2) {
            return iauTttdb(tt1, tt2, dtr, out tdb1, out tdb2);
    }


    // DECLSPEC int iauTtut1(double tt1, double tt2, double dt,
    //             double *ut11, double *ut12);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauTtut1(double tt1, double tt2, double dt,
        out double ut11, out double ut12);

    /// <summary>
    /// Convert Terrestrial Time (TT) to Universal Time (UT1).
    /// </summary>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <param name="dt">TT-UT1 in seconds</param>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Ttut1(double tt1, double tt2, double dt,
        out double ut11, out double ut12) {
            return iauTtut1(tt1, tt2, dt, out ut11, out ut12);
    }


    // DECLSPEC int iauUt1tai(double ut11, double ut12, double dta,
    //              double *tai1, double *tai2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauUt1tai(double ut11, double ut12, double dta,
    out double tai1, out double tai2);

    /// <summary>
    /// Convert Universal Time (UT1) to International Atomic Time (TAI).
    /// </summary>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2</param>
    /// <param name="dta">UT1-TAI in seconds</param>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Ut1tai(double ut11, double ut12, double dta,
    out double tai1, out double tai2) {
        return iauUt1tai(ut11, ut12, dta, out tai1, out tai2);
    }


    // DECLSPEC int iauUt1tt(double ut11, double ut12, double dt,
    //             double *tt1, double *tt2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauUt1tt(double ut11, double ut12, double dt,
    out double tt1, out double tt2);

    /// <summary>
    /// Convert Universal Time (UT1) to Terrestrial Time (TT).
    /// </summary>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2></param>
    /// <param name="dt">TT-UT1 in seconds</param>
    /// <param name="tt1">TT Julian date part 1</param>
    /// <param name="tt2">TT Julian date part 2</param>
    /// <returns>always 0</returns>
    public static int Ut1tt(double ut11, double ut12, double dt,
    out double tt1, out double tt2) {
        return iauUt1tt(ut11, ut12, dt, out tt1, out tt2);
    }


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauUt1utc(double ut11, double ut12, double dut1,
    //              double *utc1, double *utc2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauUt1utc(double ut11, double ut12, double dut1,
    out double utc1, out double utc2);

    /// <summary>
    /// Convert Universal Time (UT1) to Coordinated Universal Time (UTC).
    /// </summary>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2</param>
    /// <param name="dut1">UT1-UTC in seconds</param>
    /// <param name="utc1">UTC Julian date part 1</param>
    /// <param name="utc2">UTC Julian date part 2</param>
    /// <returns>0 = OK</returns>
    public static int Ut1utc(double ut11, double ut12, double dut1,
    out double utc1, out double utc2) {
        return iauUt1utc(ut11, ut12, dut1, out utc1, out utc2);
    }
#endif


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauUtctai(double utc1, double utc2, double *tai1, double *tai2);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauUtctai(double utc1, double utc2, out double tai1, out double tai2);

    /// <summary>
    /// Convert Coordinated Universal Time (UTC) to International Atomic Time (TAI).
    /// </summary>
    /// <param name="utc1">UTC Julian date part 1</param>
    /// <param name="utc2">UTC Julian date part 2</param>
    /// <param name="tai1">TAI Julian date part 1</param>
    /// <param name="tai2">TAI Julian date part 2</param>
    /// <returns>0 = OK</returns>
    public static int Utctai(double utc1, double utc2, out double tai1, out double tai2) {
        return iauUtctai(utc1, utc2, out tai1, out tai2);
    }
#endif


    // Anything with UTC is excluded from the production build.
#if DEBUG
    // DECLSPEC int iauUtcut1(double utc1, double utc2, double dut1,
    //              double *ut11, double *ut12);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static int iauUtcut1(double utc1, double utc2, double dut1,
    out double ut11, out double ut12);

    /// <summary>
    /// Convert Coordinated Universal Time (UTC) to Universal Time (UT1).
    /// </summary>
    /// <param name="utc1">UTC Julian date part 1</param>
    /// <param name="utc2">UTC Julian date part 2</param>
    /// <param name="dut1">UT1-UTC in seconds</param>
    /// <param name="ut11">UT1 Julian date part 1</param>
    /// <param name="ut12">UT1 Julian date part 2</param>
    /// <returns>0 = OK</returns>
    public static int Utcut1(double utc1, double utc2, double dut1,
    out double ut11, out double ut12) {
        return iauUtcut1(utc1, utc2, dut1, out ut11, out ut12);
    }
#endif



    // Test the methods in this file.
#if DEBUG

    /// <summary>
    /// Self test the time scale methods.
    /// </summary>
    public static void TestTimescales(){
        double jd1, jd2, deltat;
        int y, m, d;
        int[] hmsf;

        /* Astronomy/Timescales */
        
        // DECLSPEC int iauD2dtf(const char *scale, int ndp, double d1, double d2,
        //             int *iy, int *im, int *id, int ihmsf[4]);
        hmsf = new int[4];
        SofaValidate.EqualsApprox(D2dtf("UTC", 5, 2400000.5, 49533.99999,
            out y, out m, out d, hmsf), 0, 0);
        SofaValidate.EqualsApprox(y, 1994, 0);
        SofaValidate.EqualsApprox(m, 6, 0);
        SofaValidate.EqualsApprox(d, 30, 0);
        SofaValidate.EqualsApprox(hmsf[0], 23, 0);
        SofaValidate.EqualsApprox(hmsf[1], 59, 0);
        SofaValidate.EqualsApprox(hmsf[2], 60, 0);
        SofaValidate.EqualsApprox(hmsf[3], 13599, 0);

        // DECLSPEC int iauDat(int iy, int im, int id, double fd, double *deltat);
        SofaValidate.EqualsApprox(Dat(2003, 6, 1, 0.0, out deltat), 0, 0);
        SofaValidate.EqualsApprox(deltat, 32.0, 0.0);
        SofaValidate.EqualsApprox(Dat(2008, 1, 17, 0.0, out deltat), 0, 0);
        SofaValidate.EqualsApprox(deltat, 33.0, 0.0);

        // DECLSPEC double iauDtdb(double date1, double date2,
        //               double ut, double elong, double u, double v);
        SofaValidate.EqualsApprox(Dtdb(2448939.5, 0.123, 0.76543, 5.0123, 5525.242, 3190.0),
            -0.1280368005936998991e-2, 1e-15);

        // DECLSPEC int iauDtf2d(const char *scale, int iy, int im, int id,
        //             int ihr, int imn, double sec, double *d1, double *d2);
        SofaValidate.EqualsApprox(Dtf2d("UTC", 1994, 6, 30, 23, 59, 60.13599, out jd1, out jd2),
            0, 0);
        SofaValidate.EqualsApprox(jd1 + jd2, 2449534.49999, 1e-6);

        // DECLSPEC int iauTaitt(double tai1, double tai2, double *tt1, double *tt2);
        SofaValidate.EqualsApprox(Taitt(2453750.5, 0.892482639, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.892855139, 1e-12);

        // DECLSPEC int iauTaiut1(double tai1, double tai2, double dta,
        //              double *ut11, double *ut12);
        SofaValidate.EqualsApprox(Taiut1(2453750.5, 0.892482639, -32.6659, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8921045614537037037, 1e-12);

        // DECLSPEC int iauTaiutc(double tai1, double tai2, double *utc1, double *utc2);
        SofaValidate.EqualsApprox(Taiutc(2453750.5, 0.892482639, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8921006945555555556, 1e-12);

        // DECLSPEC int iauTcbtdb(double tcb1, double tcb2, double *tdb1, double *tdb2);
        SofaValidate.EqualsApprox(Tcbtdb(2453750.5, 0.893019599, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8928551362746343397, 1e-12);

        // DECLSPEC int iauTcgtt(double tcg1, double tcg2, double *tt1, double *tt2);
        SofaValidate.EqualsApprox(Tcgtt(2453750.5, 0.892862531, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8928551387488816828, 1e-12);

        // DECLSPEC int iauTdbtcb(double tdb1, double tdb2, double *tcb1, double *tcb2);
        SofaValidate.EqualsApprox(Tdbtcb(2453750.5, 0.892855137, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8930195997253656716, 1e-12);

        // DECLSPEC int iauTdbtt(double tdb1, double tdb2, double dtr,
        //             double *tt1, double *tt2);
        SofaValidate.EqualsApprox(Tdbtt(2453750.5, 0.892855137, -0.000201, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8928551393263888889, 1e-12);

        // DECLSPEC int iauTttai(double tt1, double tt2, double *tai1, double *tai2);
        SofaValidate.EqualsApprox(Tttai(2453750.5, 0.892482639, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.892110139, 1e-12);

        // DECLSPEC int iauTttcg(double tt1, double tt2, double *tcg1, double *tcg2);
        SofaValidate.EqualsApprox(Tttcg(2453750.5, 0.892482639, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8924900312508587113, 1e-12);

        // DECLSPEC int iauTttdb(double tt1, double tt2, double dtr,
        //             double *tdb1, double *tdb2);
        SofaValidate.EqualsApprox(Tttdb(2453750.5, 0.892855139, -0.000201, out jd1, out jd2),
            0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8928551366736111111, 1e-12);

        // DECLSPEC int iauTtut1(double tt1, double tt2, double dt,
        //             double *ut11, double *ut12);
        SofaValidate.EqualsApprox(Ttut1(2453750.5, 0.892855139, 64.8499, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8921045614537037037, 1e-12);

        // DECLSPEC int iauUt1tai(double ut11, double ut12, double dta,
        //              double *tai1, double *tai2);
        SofaValidate.EqualsApprox(Ut1tai(2453750.5, 0.892104561, -32.6659, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8924826385462962963, 1e-12);

        // DECLSPEC int iauUt1tt(double ut11, double ut12, double dt,
        //             double *tt1, double *tt2);
        SofaValidate.EqualsApprox(Ut1tt(2453750.5, 0.892104561, 64.8499, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8928551385462962963, 1e-12);

        // DECLSPEC int iauUt1utc(double ut11, double ut12, double dut1,
        //              double *utc1, double *utc2);
        SofaValidate.EqualsApprox(Ut1utc(2453750.5, 0.892104561, 0.3341, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8921006941018518519, 1e-12);

        // DECLSPEC int iauUtctai(double utc1, double utc2, double *tai1, double *tai2);
        SofaValidate.EqualsApprox(Utctai(2453750.5, 0.892100694, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8924826384444444444, 1e-12);

        // DECLSPEC int iauUtcut1(double utc1, double utc2, double dut1,
        //              double *ut11, double *ut12);
        SofaValidate.EqualsApprox(Utcut1(2453750.5, 0.892100694, 0.3341, out jd1, out jd2), 0, 0);
        SofaValidate.EqualsApprox(jd1, 2453750.5, 1e-6);
        SofaValidate.EqualsApprox(jd2, 0.8921045608981481481, 1e-12);
    }   // end self test
#endif
}   // end partial class Sofa
}   // end namespace HirosePS
