﻿// Vector.cs
// .NET adapters for IAU SOFA routines in the VectorMatrix category.
// Copyright 2017 by Paul S. Hirose.


// revision notes
#if false
2014-09-21  Began work.
2014-10-02  All methods coded, and self test good. Final desk check pending.
2014-10-03  Ensured all array arguments are checked for correct dimensions.
2014-10-07  No parameter XML comment is blank. Array parameters verified for correct dimensions, and
comments indicate exception may be thrown. Self tests call my methods, not SOFA.

2017-09-10 Building SofaJpl 2.1.
2017-09-12 Added D2dtf(), which was probably omitted by accident in previous SofaJpl.
2017-09-17 The methods in the AngleOps group have been moved to their own file.
#endif


using System;
using System.Runtime.InteropServices;


namespace HirosePS {
public static partial class Sofa {

    /* VectorMatrix/BuildRotations */


    // DECLSPEC void iauRx(double phi, double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRx(double phi, double[,] r);

    /// <summary>
    /// Rotate an r-matrix about the x-axis.
    /// </summary>
    /// <param name="phi">angle (radians)</param>
    /// <param name="r">r-matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="r"/> does not have dimensions [3,3]
    /// </exception>
    /// <remarks>Calling this function with positive phi incorporates in the supplied r-matrix r an
    /// additional rotation, about the x-axis, anticlockwise as seen looking towards the origin from
    /// positive x.</remarks>
    public static void Rx(double phi, double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauRx(phi, r);
    }

    // DECLSPEC void iauRy(double theta, double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRy(double theta, double[,] r);

    /// <summary>
    /// Rotate an r-matrix about the y-axis.
    /// </summary>
    /// <param name="theta">angle (radians)</param>
    /// <param name="r">r-matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="r"/> does not have dimensions [3,3]
    /// </exception>
    /// <remarks>Calling this function with positive theta incorporates in the supplied r-matrix r
    /// an additional rotation, about the y-axis, anticlockwise as seen looking towards the origin
    /// from positive y.</remarks>
    public static void Ry(double theta, double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauRy(theta, r);
    }


    // DECLSPEC void iauRz(double psi, double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRz(double psi, double[,] r);

    /// <summary>
    /// Rotate an r-matrix about the z-axis.
    /// </summary>
    /// <param name="psi">angle (radians)</param>
    /// <param name="r">r-matrix</param>
    /// <exception cref="ArgumentException"> if <paramref name="r"/> does not have dimensions [3,3]
    /// </exception>
    /// <remarks>Calling this function with positive psi incorporates in the supplied r-matrix r an
    /// additional rotation, about the z-axis, anticlockwise as seen looking towards the origin from
    /// positive z.</remarks>
    public static void Rz(double psi, double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauRz(psi, r);
    }


    /* VectorMatrix/CopyExtendExtract */


    // DECLSPEC void iauCp(double p[3], double c[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauCp(double[] p, double[] c);

    /// <summary>
    /// Copy a p-vector.
    /// </summary>
    /// <param name="p">p-vector to be copied></param>
    /// <param name="c">copy</param>
    /// <exception cref="ArgumentException"> unless both arguments have dimension [3]</exception>
    public static void Cp(double[] p, double[] c) {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(c);
        iauCp(p, c);
    }

    
    // DECLSPEC void iauCpv(double pv[2][3], double c[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauCpv(double[,] pv, double[,] c);

    /// <summary>
    /// Copy a position/velocity vector.
    /// </summary>
    /// <param name="pv">position/velocity vector to be copied</param>
    /// <param name="c">copy</param>
    /// <exception cref="ArgumentException"> unless both arguments have dimensions [2,3]</exception>
    public static void Cpv(double[,] pv, double[,] c) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(c);
        iauCpv(pv,c);
    }


    // DECLSPEC void iauCr(double r[3][3], double c[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauCr(double[,] r, double[,] c);

    /// <summary>
    /// Copy an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix to be copied</param>
    /// <param name="c">copy</param>
    /// <exception cref="ArgumentException"> unless both arguments have dimensions [3,3]</exception>
    public static void Cr(double[,] r, double[,] c) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble3x3(c);
        iauCr(r, c);
    }


    // DECLSPEC void iauP2pv(double p[3], double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauP2pv(double[] p, double[,] pv);

    /// <summary>
    /// Extend a p-vector to a pv-vector by appending a zero velocity.
    /// </summary>
    /// <param name="p">p-vector</param>
    /// <param name="pv">pv-vector</param>
    /// <exception cref="ArgumentException"> unless arguments have dimensions [3] and [2,3]
    /// </exception>
    public static void P2pv(double[] p, double[,] pv) {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble2x3(pv);
        iauP2pv(p, pv);
    }


    // DECLSPEC void iauPv2p(double pv[2][3], double p[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPv2p(double[,] pv, double[] p);

    /// <summary>
    /// Discard velocity component of a pv-vector.
    /// </summary>
    /// <param name="pv">pv-vector</param>
    /// <param name="p">p-vector</param>
    /// <exception cref="ArgumentException"> unless arguments have dimensions [2,3] and [3]
    /// </exception>
    public static void Pv2p(double[,] pv, double[] p) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble1x3(p);
        iauPv2p(pv, p);
    }


    /* VectorMatrix/Initialization */


    // DECLSPEC void iauIr(double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauIr(double[,] r);

    /// <summary>
    /// Initialize a rotation matrix to the identity matrix.
    /// </summary>
    /// <param name="r">rotation matrix</param>
    /// <exception cref="ArgumentException"> if argument does not have dimensions [3,3]</exception>
    public static void Ir(double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauIr(r);
    }


    // DECLSPEC void iauZp(double p[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauZp(double[] p);

    /// <summary>
    /// Zero a p-vector.
    /// </summary>
    /// <param name="p">p-vector</param>
    /// <exception cref="ArgumentException"> if argument does not have dimension [3]</exception>
    public static void Zp(double[] p) {
        SofaValidate.VerifyDouble1x3(p);
        iauZp(p);
    }


    // DECLSPEC void iauZpv(double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauZpv(double[,] pv);

    /// <summary>
    /// Zero a pv-vector.
    /// </summary>
    /// <param name="pv">pv-vector</param>
    /// <exception cref="ArgumentException"> if argument does not have dimensions [2,3]</exception>
    public static void Zpv(double[,] pv) {
        SofaValidate.VerifyDouble2x3(pv);
        iauZpv(pv);
    }


    // DECLSPEC void iauZr(double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauZr(double[,] r);

    /// <summary>
    /// Initialize an r-matrix to the null matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <exception cref="ArgumentException"> if argument does not have dimensions [3,3]</exception>
    public static void Zr(double[,] r) {
        SofaValidate.VerifyDouble3x3(r);
        iauZr(r);
    }


    /* VectorMatrix/MatrixOps */


    // DECLSPEC void iauRxr(double a[3][3], double b[3][3], double atb[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRxr(double[,] a, double[,] b, double[,] atb);

    /// <summary>
    /// Multiply two r-matrices.
    /// </summary>
    /// <param name="a">first r-matrix</param>
    /// <param name="b">second r-matrix</param>
    /// <param name="atb">a * b</param>
    /// <exception cref="ArgumentException"> unless all arguments have dimensions [3,3]</exception>
    /// <remarks>It is permissible to use either input to receive the output.</remarks>
    public static void Rxr(double[,] a, double[,] b, double[,] atb) {
        SofaValidate.VerifyDouble3x3(a);
        SofaValidate.VerifyDouble3x3(b);
        SofaValidate.VerifyDouble3x3(atb);
        iauRxr(a, b, atb);
    }

    
    // DECLSPEC void iauTr(double r[3][3], double rt[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauTr(double[,] r, double[,] rt);

    /// <summary>
    /// Transpose an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="rt">transpose</param>
    /// <exception cref="ArgumentException"> if the arguments are not dimensioned [3,3]</exception>
    /// <remarks>It is permissible for r and rt to be the same array.</remarks>
    public static void Tr(double[,] r, double[,] rt) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble3x3(rt);
        iauTr(r, rt);
    }


    /* VectorMatrix/MatrixVectorProducts */

    
    // DECLSPEC void iauRxp(double r[3][3], double p[3], double rp[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRxp(double[,] r, double[] p, double[] rp);

    /// <summary>
    /// Multiply a p-vector by an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="p">p-vector</param>
    /// <param name="rp">product of r-matrix and p-vector</param>
    /// <exception cref="ArgumentException"> if arguments do not have correct dimensions: [3] and
    /// [3,3].</exception>
    /// <remarks>It is permissible for <paramref name="p"/> and <paramref name="rp"/> to be the same
    /// array.</remarks>
    public static void Rxp(double[,] r, double[] p, double[] rp) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(rp);
        iauRxp(r, p, rp);
    }


    // DECLSPEC void iauRxpv(double r[3][3], double pv[2][3], double rpv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRxpv(double[,] r, double[,] pv, double[,] rpv);

    /// <summary>
    /// Multiply a pv-vector by an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="pv">pv-vector</param>
    /// <param name="rpv">product of r-matrix and pv-vector</param>
    /// <exception cref="ArgumentException"> if arguments do not have correct dimensions: [3,3] and
    /// [2,3].</exception>
    /// <remarks>It is permissible for <paramref name="pv"/> and <paramref name="rpv"/> to be the
    /// same array.</remarks>
    public static void Rxpv(double[,] r, double[,] pv, double[,] rpv) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(rpv);
        iauRxpv(r, pv, rpv);
    }


    // DECLSPEC void iauTrxp(double r[3][3], double p[3], double trp[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauTrxp(double[,] r, double[] p, double[] trp);

    /// <summary>
    /// Multiply a p-vector by the transpose of an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="p">p-vector</param>
    /// <param name="trp">product of transposed r-matrix and p-vector</param>
    /// <exception cref="ArgumentException"> if arguments do not have correct dimensions: [3,3] and
    /// [3].</exception>
    /// <remarks><paramref name="p"/> and <paramref name="trp"/> can be the same array.</remarks>
    public static void Trxp(double[,] r, double[] p, double[] trp) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(trp);
        iauTrxp(r, p, trp);
    }

    // DECLSPEC void iauTrxpv(double r[3][3], double pv[2][3], double trpv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauTrxpv(double[,] r, double[,] pv, double[,] trpv);

    /// <summary>
    /// Multiply a pv-vector by the transpose of an r-matrix.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="pv">pv-vector</param>
    /// <param name="trpv">product of transposed r-matrix and pv-vector</param>
    /// <exception cref="ArgumentException"> if arguments do not have correct dimensions: [3,3] and
    /// [2,3].</exception>
    /// <remarks><paramref name="pv"/> and <paramref name="trpv"/> can be the same array.</remarks>
    public static void Trxpv(double[,] r, double[,] pv, double[,] trpv) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(trpv);
        iauTrxpv(r, pv, trpv);
    }


    /* VectorMatrix/RotationVectors */


    // DECLSPEC void iauRm2v(double r[3][3], double w[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRm2v(double[,] r, double[] w);

    /// <summary>
    /// Express a rotation matrix as a rotation vector.
    /// </summary>
    /// <param name="r">r-matrix</param>
    /// <param name="w">r-vector</param>
    /// <exception cref="ArgumentException"> if arguments do not have dimensions [3,3] and [3].
    /// </exception>
    public static void Rm2v(double[,] r, double[] w) {
        SofaValidate.VerifyDouble3x3(r);
        SofaValidate.VerifyDouble1x3(w);
        iauRm2v(r, w);
    }


    // DECLSPEC void iauRv2m(double w[3], double r[3][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauRv2m(double[] w, double[,] r);

    /// <summary>
    /// Form the rotation matrix corresponding to a given rotation vector.
    /// </summary>
    /// <param name="w">r-vector</param>
    /// <param name="r">r-matrix</param>
    /// <exception cref="ArgumentException"> if arguments do not have dimensions [3] and [3,3].
    /// </exception>
    public static void Rv2m(double[] w, double[,] r) {
        SofaValidate.VerifyDouble1x3(w);
        SofaValidate.VerifyDouble3x3(r);
        iauRv2m(w,r);
    }


    /* VectorMatrix/SeparationAndAngle */


    // DECLSPEC double iauPap(double a[3], double b[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauPap(double[] a, double[] b);

    /// <summary>
    /// Position angle from two vectors.
    /// </summary>
    /// <param name="a">direction of reference point</param>
    /// <param name="b">direction of point whose PA is required</param>
    /// <returns>position angle (radians)</returns>
    /// <exception cref="ArgumentException"> unless both arguments have dimension [3].</exception>
    /// <remarks>Zero is returned if either vector is null.</remarks>
    public static double Pap(double[] a, double[] b) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        return iauPap(a, b);
    }

    
    // DECLSPEC double iauPas(double al, double ap, double bl, double bp);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauPas(double al, double ap, double bl, double bp);

    /// <summary>
    /// Position angle from spherical coordinates.
    /// </summary>
    /// <param name="al">longitude of first point (radians)</param>
    /// <param name="ap">latitude of first point (radians)</param>
    /// <param name="bl">longitude of second point (radians)</param>
    /// <param name="bp">latitude of second point (radians)</param>
    /// <returns>position angle of second point with respect to first (radians, -pi to +pi)
    /// </returns>
    /// <remarks>The angle is zero north and increases east. Zero is returned if the
    /// points coincide.</remarks>
    public static double Pas(double al, double ap, double bl, double bp) {
        return iauPas(al, ap, bl, bp);
    }


    // DECLSPEC double iauSepp(double a[3], double b[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauSepp(double[] a, double[] b);

    /// <summary>
    /// Angular separation between two p-vectors.
    /// </summary>
    /// <param name="a">first vector</param>
    /// <param name="b">second vector</param>
    /// <returns>angular separation (radians, always positive)</returns>
    /// <exception cref="ArgumentException"> if either vector does not have dimension [3].
    /// </exception>
    /// <remarks>If either vector is null, a zero result is returned.</remarks>
    public static double Sepp(double[] a, double[] b) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        return iauSepp(a, b);
    }


    // DECLSPEC double iauSeps(double al, double ap, double bl, double bp);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauSeps(double al, double ap, double bl, double bp);

    /// <summary>
    /// Angular separation between two sets of spherical coordinates.
    /// </summary>
    /// <param name="al">longitude, point 1 (radians)</param>
    /// <param name="ap">latitude, point 1 (radians)</param>
    /// <param name="bl">longitude, point 2 (radians)</param>
    /// <param name="bp">latitude, point 2 (radians)</param>
    /// <returns>separation angle (radians)</returns>
    public static double Seps(double al, double ap, double bl, double bp) {
        return iauSeps(al, ap, bl, bp);
    }

    
    /* VectorMatrix/SphericalCartesian */


    // DECLSPEC void iauC2s(double p[3], double *theta, double *phi);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauC2s(double[] p, out double theta, out double phi);

    /// <summary>
    /// P-vector to spherical coordinates.
    /// </summary>
    /// <param name="p">p-vector</param>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <exception cref="ArgumentException"> if vector does not have dimension [3].</exception>
    public static void C2s(double[] p, out double theta, out double phi) {
        SofaValidate.VerifyDouble1x3(p);
        iauC2s(p, out theta, out phi);
    }


    // DECLSPEC void iauP2s(double p[3], double *theta, double *phi, double *r);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauP2s(double[] p, out double theta, out double phi, out double r);

    /// <summary>
    /// Convert vector to spherical polar coordinates.
    /// </summary>
    /// <param name="p">p-vector</param>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <param name="r">radial distance</param>
    /// <exception cref="ArgumentException"> if <paramref name="p"/> is not dimensioned [3]
    /// </exception>
    public static void P2s(double[] p, out double theta, out double phi, out double r) {
        SofaValidate.VerifyDouble1x3(p);
        iauP2s(p, out theta, out phi, out r);
    }

    // DECLSPEC void iauPv2s(double pv[2][3],
    //             double *theta, double *phi, double *r,
    //             double *td, double *pd, double *rd);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPv2s(double[,] pv, out double theta, out double phi, out double r,
        out double td, out double pd, out double rd);

    /// <summary>
    /// Convert position/velocity from Cartesian to spherical coordinates.
    /// </summary>
    /// <param name="pv">pv-vector</param>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <param name="r">radial distance</param>
    /// <param name="td">rate of change of theta</param>
    /// <param name="pd">rate of change of phi</param>
    /// <param name="rd">rate of change of r</param>
    /// <exception cref="ArgumentException"> if PV vector dimensions are not [2,3]</exception>
    public static void Pv2s(double[,] pv, out double theta, out double phi, out double r,
        out double td, out double pd, out double rd)
    {
        SofaValidate.VerifyDouble2x3(pv);
        iauPv2s(pv, out theta, out phi, out r, out td, out pd, out rd);
    }


    // DECLSPEC void iauS2c(double theta, double phi, double c[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauS2c(double theta, double phi, double[] c);

    /// <summary>
    /// Convert spherical coordinates to Cartesian.
    /// </summary>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <param name="c">direction cosines</param>
    /// <exception cref="ArgumentException"> if <paramref name="c"/> is not dimensioned [3]
    /// </exception>
    public static void S2c(double theta, double phi, double[] c) {
        SofaValidate.VerifyDouble1x3(c);
        iauS2c(theta, phi, c);
    }


    // DECLSPEC void iauS2p(double theta, double phi, double r, double p[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauS2p(double theta, double phi, double r, double[] p);

    /// <summary>
    /// Convert spherical polar coordinates to p-vector.
    /// </summary>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <param name="r">radial distance</param>
    /// <param name="p">Cartesian coordinates</param>
    /// <exception cref="ArgumentException"> if <paramref name="p"/> does not have dimension [3]
    /// </exception>
    public static void S2p(double theta, double phi, double r, double[] p) {
        SofaValidate.VerifyDouble1x3(p);
        iauS2p(theta, phi, r, p);
    }


    // DECLSPEC void iauS2pv(double theta, double phi, double r,
    //             double td, double pd, double rd,
    //             double pv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauS2pv(double theta, double phi, double r, double td, double pd, double rd,
        double[,] pv);

    /// <summary>
    /// Convert position/velocity from spherical to Cartesian coordinates.
    /// </summary>
    /// <param name="theta">longitude angle (radians)</param>
    /// <param name="phi">latitude angle (radians)</param>
    /// <param name="r">radial distance</param>
    /// <param name="td">rate of change of theta</param>
    /// <param name="pd">rate of change of phi</param>
    /// <param name="rd">rate of change of r</param>
    /// <param name="pv">pv-vector</param>
    /// <exception cref="ArgumentException"> if <paramref name="pv"/> is not dimensioned [2,3]
    /// </exception>
    public static void S2pv(double theta, double phi, double r, double td, double pd, double rd,
        double[,] pv)
    {
        SofaValidate.VerifyDouble2x3(pv);
        iauS2pv(theta, phi, r, td, pd, rd, pv);
    }


    /* VectorMatrix/VectorOps */
    
    
    // DECLSPEC double iauPdp(double a[3], double b[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauPdp(double[] a, double[] b);

    /// <summary>
    /// Vector inner (scalar or dot) product.
    /// </summary>
    /// <param name="a">first vector</param>
    /// <param name="b">second vecto</param>
    /// <returns>a . b</returns>
    /// <exception cref="ArgumentException"> if arguments are not dimensioned [3]</exception>
    public static double Pdp(double[] a, double[] b) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        return iauPdp(a, b);
    }


    // DECLSPEC double iauPm(double p[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static double iauPm(double[] p);

    /// <summary>
    /// Modulus of a vector.
    /// </summary>
    /// <param name="p">vector</param>
    /// <returns>modulus</returns>
    /// <exception cref="ArgumentException"> if vector is not dimensioned [3]</exception>
    public static double Pm(double[] p) {
        SofaValidate.VerifyDouble1x3(p);
        return iauPm(p);
    }


    // DECLSPEC void iauPmp(double a[3], double b[3], double amb[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPmp(double[] a, double[] b, double[] amb);

    /// <summary>
    /// Vector subtraction.
    /// </summary>
    /// <param name="a">first vector</param>
    /// <param name="b">second vector</param>
    /// <param name="amb">a - b</param>
    /// <remarks>It is permissible to re-use either input to receive output.</remarks>
    /// <exception cref="ArgumentException"> if arguments are not dimensioned [3]</exception>
    public static void Pmp(double[] a, double[] b, double[] amb) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        SofaValidate.VerifyDouble1x3(amb);
        iauPmp(a, b, amb);
    }


    // DECLSPEC void iauPn(double p[3], double *r, double u[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPn(double[] p, out double r, double[] u);

    /// <summary>
    /// Convert a vector into modulus and unit vector.
    /// </summary>
    /// <param name="p">vector</param>
    /// <param name="r">modulus</param>
    /// <param name="u">unit vector</param>
    /// <exception cref="ArgumentException"> unless both vector arguments are dimensioned [3]
    /// </exception>
    /// <remarks>If <paramref name="p"/> is null, the result is null. The same array may be used
    /// for input and output.</remarks>
    public static void Pn(double[] p, out double r, double[] u) {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(u);
        iauPn(p, out r, u);
    }


    // DECLSPEC void iauPpp(double a[3], double b[3], double apb[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPpp(double[] a, double[] b, double[] apb);

    /// <summary>
    /// P-vector addition.
    /// </summary>
    /// <param name="a">first p-vector</param>
    /// <param name="b">second p-vector</param>
    /// <param name="apb">p-vector sum</param>
    /// <remarks>It is permissible to receive the output in one of the input arrays.</remarks>
    /// <exception cref="ArgumentException"> if arguments are not dimensioned [3]</exception>
    public static void Ppp(double[] a, double[] b, double[] apb) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        SofaValidate.VerifyDouble1x3(apb);
        iauPpp(a, b, apb);
    }


    // DECLSPEC void iauPpsp(double a[3], double s, double b[3], double apsb[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPpsp(double[] a, double s, double[] b, double[] apsb);

    /// <summary>
    /// P-vector plus scaled p-vector.
    /// </summary>
    /// <param name="a">first p-vector</param>
    /// <param name="s">scalar (multiplier for second vector)</param>
    /// <param name="b">second p-vector</param>
    /// <param name="apsb">a + s*b</param>
    /// <exception cref="ArgumentException"> if any array is not length 3.</exception>
    /// <remarks>It is permissible to receive the output in one of the input arrays.</remarks>
    public static void Ppsp(double[] a, double s, double[] b, double[] apsb) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        SofaValidate.VerifyDouble1x3(apsb);
        iauPpsp(a, s, b, apsb);
    }


    // DECLSPEC void iauPvdpv(double a[2][3], double b[2][3], double adb[2]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvdpv(double[,] a, double[,] b, double[] adb);

    /// <summary>
    /// Inner (=scalar=dot) product of two pv-vectors.
    /// </summary>
    /// <param name="a">first pv-vector</param>
    /// <param name="b">second pv-vector</param>
    /// <param name="adb">dot-product of the two p-vectors and its derivative</param>
    /// <exception cref="ArgumentException"> if the inputs are not dimensioned [2,3] or output not
    /// [2]</exception>
    public static void Pvdpv(double[,] a, double[,] b, double[] adb) {
        SofaValidate.VerifyDouble2x3(a);
        SofaValidate.VerifyDouble2x3(b);
        if (adb.Length != 2)
            throw new ArgumentException("Array must have dimension [2].");
        iauPvdpv(a, b, adb);
    }


    // DECLSPEC void iauPvm(double pv[2][3], double *r, double *s);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvm(double[,] pv, out double r, out double s);

    /// <summary>
    /// Modulus of pv-vector.
    /// </summary>
    /// <param name="pv">pv-vector</param>
    /// <param name="r">modulus of position component</param>
    /// <param name="s">modulus of velocity component</param>
    /// <exception cref="ArgumentException"> if array is not dimensioned [2,3]</exception>
    public static void Pvm(double[,] pv, out double r, out double s) {
        SofaValidate.VerifyDouble2x3(pv);
        iauPvm(pv, out r, out s);
    }


    // DECLSPEC void iauPvmpv(double a[2][3], double b[2][3], double amb[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvmpv(double[,] a, double[,] b, double[,] amb);

    /// <summary>
    /// Subtract one pv-vector from another.
    /// </summary>
    /// <param name="a">first pv-vector</param>
    /// <param name="b">second pv-vector</param>
    /// <param name="amb">a - b</param>
    /// <exception cref="ArgumentException"> if any array is not dimensioned [2,3]</exception>
    /// <remarks>It is permissible to use an input array for the output.</remarks>
    public static void Pvmpv(double[,] a, double[,] b, double[,] amb) {
        SofaValidate.VerifyDouble2x3(a);
        SofaValidate.VerifyDouble2x3(b);
        SofaValidate.VerifyDouble2x3(amb);
        iauPvmpv(a, b, amb);
    }


    // DECLSPEC void iauPvppv(double a[2][3], double b[2][3], double apb[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvppv(double[,] a, double[,] b, double[,] apb);

    /// <summary>
    /// Add one pv-vector to another.
    /// </summary>
    /// <param name="a">first pv-vector</param>
    /// <param name="b">second pv-vector</param>
    /// <param name="apb">a + b</param>
    /// <exception cref="ArgumentException"> if any array is not dimensioned [2,3]</exception>
    public static void Pvppv(double[,] a, double[,] b, double[,] apb) {
        SofaValidate.VerifyDouble2x3(a);
        SofaValidate.VerifyDouble2x3(b);
        SofaValidate.VerifyDouble2x3(apb);
        iauPvppv(a, b, apb);
    }


    // DECLSPEC void iauPvu(double dt, double pv[2][3], double upv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvu(double dt, double[,] pv, double[,] upv);

    /// <summary>
    /// Update a pv-vector.
    /// </summary>
    /// <param name="dt">time interval</param>
    /// <param name="pv">pv-vector</param>
    /// <param name="upv">p updated, v unchanged</param>
    /// <exception cref="ArgumentException"> if the arrays do not have dimensions [2,3]</exception>
    /// <remarks>It is permissible to receive the output in the input array.</remarks>
    public static void Pvu(double dt, double[,] pv, double[,] upv) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(upv);
        iauPvu(dt, pv, upv);
    }


    // DECLSPEC void iauPvup(double dt, double pv[2][3], double p[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvup(double dt, double[,] pv, double[] p);

    /// <summary>
    /// Update a pv-vector, discarding the velocity component.
    /// </summary>
    /// <param name="dt">time interval</param>
    /// <param name="pv">pv-vector</param>
    /// <param name="p">p-vector</param>
    /// <exception cref="ArgumentException"> if the arguments are not dimensioned [2,3] and [1,3]
    /// </exception>
    public static void Pvup(double dt, double[,] pv, double[] p) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble1x3(p);
        iauPvup(dt, pv, p);
    }


    // DECLSPEC void iauPvxpv(double a[2][3], double b[2][3], double axb[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPvxpv(double[,] a, double[,] b, double[,] axb);

    /// <summary>
    /// Outer (=vector=cross) product of two pv-vectors.
    /// </summary>
    /// <param name="a">first pv-vector</param>
    /// <param name="b">second pv-vector</param>
    /// <param name="axb">a x b</param>
    /// <exception cref="ArgumentException"> if any argument is not dimensioned [2,3]</exception>
    /// <remarks>It is permissible to receive the output in one of the input arrays.</remarks>
    public static void Pvxpv(double[,] a, double[,] b, double[,] axb) {
        SofaValidate.VerifyDouble2x3(a);
        SofaValidate.VerifyDouble2x3(b);
        SofaValidate.VerifyDouble2x3(axb);
        iauPvxpv(a, b, axb);
    }

    
    // DECLSPEC void iauPxp(double a[3], double b[3], double axb[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauPxp(double[] a, double[] b, double[] axb);

    /// <summary>
    /// p-vector outer (=vector=cross) product.
    /// </summary>
    /// <param name="a">first p-vector</param>
    /// <param name="b">second p-vector</param>
    /// <param name="axb">cross product</param>
    /// <exception cref="ArgumentException"> if any argument is not dimensioned [3]</exception>
    /// <remarks>It is permissible to receive the output in one of the input arrays.</remarks>
    public static void Pxp(double[] a, double[] b, double[] axb) {
        SofaValidate.VerifyDouble1x3(a);
        SofaValidate.VerifyDouble1x3(b);
        SofaValidate.VerifyDouble1x3(axb);
        iauPxp(a, b, axb);
    }


    // DECLSPEC void iauS2xpv(double s1, double s2, double pv[2][3], double spv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauS2xpv(double s1, double s2, double[,] pv, double[,] spv);

    /// <summary>
    /// Multiply a pv-vector by two scalars.
    /// </summary>
    /// <param name="s1">scalar to multiply position component by</param>
    /// <param name="s2">scalar to multiply velocity component by</param>
    /// <param name="pv">pv-vector</param>
    /// <param name="spv">scaled pv-vector out</param>
    /// <exception cref="ArgumentException"> if either array is not dimensioned [2,3]</exception>
    /// <remarks>It is permissible to receive the output in the input array.</remarks>
    public static void S2xpv(double s1, double s2, double[,] pv, double[,] spv) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(spv);
        iauS2xpv(s1, s2, pv, spv);
    }

    
    // DECLSPEC void iauSxp(double s, double p[3], double sp[3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauSxp(double s, double[] p, double[] sp);

    /// <summary>
    /// Multiply a p-vector by a scalar.
    /// </summary>
    /// <param name="s">scalar</param>
    /// <param name="p">p-vector</param>
    /// <param name="sp">p-vector times scalar</param>
    /// <exception cref="ArgumentException"> if either array is not dimensioned [3]</exception>
    /// <remarks>It is permissible to receive the output in the input array.</remarks>
    public static void Sxp(double s, double[] p, double[] sp) {
        SofaValidate.VerifyDouble1x3(p);
        SofaValidate.VerifyDouble1x3(sp);
        iauSxp(s, p, sp);
    }


    // DECLSPEC void iauSxpv(double s, double pv[2][3], double spv[2][3]);
    [DllImport("SofaWin32", CallingConvention = CallingConvention.Cdecl)]
    extern static void iauSxpv(double s, double[,] pv, double[,] spv);

    /// <summary>
    /// Multiply a pv-vector by a scalar.
    /// </summary>
    /// <param name="s">scalar</param>
    /// <param name="pv">p-vector</param>
    /// <param name="spv">scalar times pv-vector</param>
    /// <exception cref="ArgumentException"> if either array is not dimensioned [2,3]</exception>
    /// <remarks>It is permissible to receive the output in the input array.</remarks>
    public static void Sxpv(double s, double[,] pv, double[,] spv) {
        SofaValidate.VerifyDouble2x3(pv);
        SofaValidate.VerifyDouble2x3(spv);
        iauSxpv(s, pv, spv);
    }


    // Self test this part of the class.
#if DEBUG

    /// <summary>
    /// Self test vector methods.
    /// </summary>
    public static void TestVector() {
        double[,] r1 = new double[3, 3];
        double[,] r2 = new double[3, 3];
        double[,] pv1 = new double[2, 3];
        double[,] pv2 = new double[2, 3];
        double[] p1 = new double[3];
        double[] p2 = new double[3];
        double[] pvdpv = new double[2];
        double phi, theta, psi, r, td, pd, rd;
        int j;

        /* VectorMatrix/BuildRotations */

        // DECLSPEC void iauRx(double phi, double r[3][3]);
        phi = 0.3456789;
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Rx(phi, r1);
        SofaValidate.EqualsApprox(r1, new double[,]{
            {2.0, 3.0, 2.0},
            {3.839043388235612460, 3.237033249594111899, 4.516714379005982719},
            {1.806030415924501684, 3.085711545336372503, 3.687721683977873065}}, 1e-12);

        // DECLSPEC void iauRy(double theta, double r[3][3]);
        theta = 0.3456789;
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Ry(theta, r1);
        SofaValidate.EqualsApprox(r1, new double[,]{
            {0.8651847818978159930, 1.467194920539316554, 0.1875137911274457342},
            {3, 2, 3},
            {3.500207892850427330, 4.779889022262298150, 5.381899160903798712}}, 1e-12);

        // DECLSPEC void iauRz(double psi, double r[3][3]);
        psi = 0.3456789;
        r1 = new double[,]{{2.0, 3.0, 2.0},{3.0, 2.0, 3.0},{3.0, 4.0, 5.0}};
        Rz(psi, r1);
        SofaValidate.EqualsApprox(r1, new double[,]{
            {2.898197754208926769, 3.500207892850427330, 2.898197754208926769},
            {2.144865911309686813, 0.865184781897815993, 2.144865911309686813},
            {3.0, 4.0, 5.0}}, 1e-12);

        /* VectorMatrix/CopyExtendExtract */

        // DECLSPEC void iauCp(double p[3], double c[3]);
        p1 = new double[] { 0.3, 1.2, -2.5 };
        Cp(p1, p2);
        SofaValidate.EqualsApprox(p2, p1, 0.0);

        // DECLSPEC void iauCpv(double pv[2][3], double c[2][3]);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { -0.5, 3.1, 0.9 } };
        Cpv(pv1, pv2);
        SofaValidate.EqualsApprox(pv2, pv1, 0.0);

        // DECLSPEC void iauCr(double r[3][3], double c[3][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Cr(r1, r2);
        SofaValidate.EqualsApprox(r2, r1, 0.0);

        // DECLSPEC void iauP2pv(double p[3], double pv[2][3]);
        p1 = new double[] { 0.25, 1.2, 3.0 };
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { -0.5, 3.1, 0.9 } };
        P2pv(p1, pv1);
        SofaValidate.EqualsApprox(pv1, new double[,]{
            {0.25, 1.2, 3.0},{0.0, 0.0,  0.0}}, 0.0);

        // DECLSPEC void iauPv2p(double pv[2][3], double p[3]);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { -0.5, 3.1, 0.9 } };
        Pv2p(pv1, p1);
        SofaValidate.EqualsApprox(p1, new double[] { 0.3, 1.2, -2.5 }, 0.0);

        /* VectorMatrix/Initialization */

        // DECLSPEC void iauIr(double r[3][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Ir(r1);
        SofaValidate.EqualsApprox(r1, new double[,] {
            {1.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0.0, 0.0, 1.0}}, 0.0);

        // DECLSPEC void iauZp(double p[3]);
        p1 = new double[] { 0.3, 1.2, -2.5 };
        Zp(p1);
        SofaValidate.EqualsApprox(p1, new double[3], 0.0);

        // DECLSPEC void iauZpv(double pv[2][3]);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { -0.5, 3.1, 0.9 } };
        Zpv(pv1);
        SofaValidate.EqualsApprox(pv1, new double[,] { { 0.0, 0.0, 0.0 }, { 0.0, 0.0, 0.0 }}, 0);

        // DECLSPEC void iauZr(double r[3][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Zr(r1);
        SofaValidate.EqualsApprox(r1, new double[3, 3], 0.0);

        /* VectorMatrix/MatrixOps */

        // DECLSPEC void iauRxr(double a[3][3], double b[3][3], double atb[3][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        r2 = new double[,] { { 1.0, 2.0, 2.0 }, { 4.0, 1.0, 1.0 }, { 3.0, 0.0, 1.0 } };
        Rxr(r1, r2, r2);
        SofaValidate.EqualsApprox(r2, new double[,]{
            {20.0, 7.0, 9.0},{20.0, 8.0, 11.0},{34.0, 10.0, 15.0}}, 1e-12);

        // DECLSPEC void iauTr(double r[3][3], double rt[3][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        Tr(r1, r2);
        SofaValidate.EqualsApprox(r2, new double[,]{
            {2.0, 3.0, 3.0}, {3.0, 2.0, 4.0}, {2.0, 3.0, 5.0}}, 0.0);

        /* VectorMatrix/MatrixVectorProducts */

        // DECLSPEC void iauRxp(double r[3][3], double p[3], double rp[3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        p1 = new double[] { 0.2, 1.5, 0.1 };
        Rxp(r1, p1, p2);
        SofaValidate.EqualsApprox(p2, new double[] { 5.1, 3.9, 7.1 }, 1e-12);

        // DECLSPEC void iauRxpv(double r[3][3], double pv[2][3], double rpv[2][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        pv1 = new double[,] { { 0.2, 1.5, 0.1 }, { 1.5, 0.2, 0.1 } };
        Rxpv(r1, pv1, pv1);
        SofaValidate.EqualsApprox(pv1, new double[,]{{5.1, 3.9, 7.1},{3.8, 5.2, 5.8}}, 1e-12);

        // DECLSPEC void iauTrxp(double r[3][3], double p[3], double trp[3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        p1 = new double[] { 0.2, 1.5, 0.1 };
        Trxp(r1, p1, p2);
        SofaValidate.EqualsApprox(p2, new double[] { 5.2, 4.0, 5.4 }, 1e-12);

        // DECLSPEC void iauTrxpv(double r[3][3], double pv[2][3], double trpv[2][3]);
        r1 = new double[,] { { 2.0, 3.0, 2.0 }, { 3.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } };
        pv1 = new double[,] { { 0.2, 1.5, 0.1 }, { 1.5, 0.2, 0.1 } };
        Trxpv(r1, pv1, pv2);
        SofaValidate.EqualsApprox(pv2, new double[,]{{5.2, 4.0, 5.4}, {3.9, 5.3, 4.1}}, 1e-12);

        /* VectorMatrix/RotationVectors */

        // DECLSPEC void iauRm2v(double r[3][3], double w[3]);
        r1 = new double[,] { { 0.00, -0.80, -0.60 }, { 0.80, -0.36, 0.48 }, { 0.60, 0.48, -0.64 } };
        Rm2v(r1, p1);
        SofaValidate.EqualsApprox(p1,
            new double[] { 0.0, 1.413716694115406957, -1.884955592153875943 }, 1e-12);

        // DECLSPEC void iauRv2m(double w[3], double r[3][3]);
        Rv2m(new double[] { 0.0, 1.41371669, -1.88495559 }, r1);
        SofaValidate.EqualsApprox(r1, new double[,]{
            {-0.7071067782221119905, -0.5656854276809129651, -0.4242640700104211225},
            {0.5656854276809129651, -0.0925483394532274246, -0.8194112531408833269},
            {0.4242640700104211225, -0.8194112531408833269, 0.3854415612311154341}}, 1e-14);

        /* VectorMatrix/SeparationAndAngle */

        // DECLSPEC double iauPap(double a[3], double b[3]);
        p1 = new double[] { 1.0, 0.1, 0.2 };
        p2 = new double[] { -3.0, 1e-3, 0.2 };
        SofaValidate.EqualsApprox(Pap(p1, p2), 0.3671514267841113674, 1e-12);

        // DECLSPEC double iauPas(double al, double ap, double bl, double bp);
        SofaValidate.EqualsApprox(Pas(1.0, 0.1, 0.2, -1.0), -2.724544922932270424, 1e-12);

        // DECLSPEC double iauSepp(double a[3], double b[3]);
        p1 = new double[] { 1.0, 0.1, 0.2 };
        p2 = new double[] { -3.0, 1e-3, 0.2 };
        SofaValidate.EqualsApprox(Sepp(p1, p2), 2.860391919024660768, 1e-12);

        // DECLSPEC double iauSeps(double al, double ap, double bl, double bp);
        SofaValidate.EqualsApprox(Seps(1.0, 0.1, 0.2, -3.0), 2.346722016996998842, 1e-14);

        /* VectorMatrix/SphericalCartesian */

        // DECLSPEC void iauC2s(double p[3], double *theta, double *phi);
        p1 = new double[] { 100.0, -50.0, 25.0 };
        C2s(p1, out theta, out phi);
        SofaValidate.EqualsApprox(theta, -0.4636476090008061162, 1e-14);
        SofaValidate.EqualsApprox(phi, 0.2199879773954594463, 1e-14);

        // DECLSPEC void iauP2s(double p[3], double *theta, double *phi, double *r);
        p1 = new double[] { 100.0, -50.0, 25.0 };
        P2s(p1, out theta, out phi, out r);
        SofaValidate.EqualsApprox(theta, -0.4636476090008061162, 1e-12);
        SofaValidate.EqualsApprox(phi, 0.2199879773954594463, 1e-12);
        SofaValidate.EqualsApprox(r, 114.5643923738960002, 1e-9);

        // DECLSPEC void iauPv2s(double pv[2][3],
        //             double *theta, double *phi, double *r,
        //             double *td, double *pd, double *rd);
        pv1 = new double[,]{
            {-0.4514964673880165,  0.03093394277342585,  0.05594668105108779},
            {1.292270850663260e-5, 2.652814182060692e-6, 2.568431853930293e-6}};
        Pv2s(pv1, out theta, out phi, out r, out td, out pd, out rd);
        SofaValidate.EqualsApprox(theta, 3.073185307179586515, 1e-12);
        SofaValidate.EqualsApprox(phi, 0.1229999999999999992, 1e-12);
        SofaValidate.EqualsApprox(r, 0.4559999999999999757, 1e-12);
        SofaValidate.EqualsApprox(td, -0.7800000000000000364e-5, 1e-16);
        SofaValidate.EqualsApprox(pd, 0.9010000000000001639e-5, 1e-16);
        SofaValidate.EqualsApprox(rd, -0.1229999999999999832e-4, 1e-16);

        // DECLSPEC void iauS2c(double theta, double phi, double c[3]);
        S2c(3.0123, -0.999, p1);
        SofaValidate.EqualsApprox(p1, new double[]{
            -0.5366267667260523906, 0.0697711109765145365, -0.8409302618566214041}, 1e-12);

        // DECLSPEC void iauS2p(double theta, double phi, double r, double p[3]);
        S2p(-3.21, 0.123, 0.456, p1);
        SofaValidate.EqualsApprox(p1, new double[]{
            -0.4514964673880165228, 0.0309339427734258688, 0.0559466810510877933}, 1e-12);

        // DECLSPEC void iauS2pv(double theta, double phi, double r,
        //             double td, double pd, double rd,
        //             double pv[2][3]);
        S2pv(-3.21, 0.123, 0.456, -7.8e-6, 9.01e-6, -1.23e-5, pv1);
        SofaValidate.EqualsApprox(pv1, new double[,]{
            {-0.4514964673880165228, 0.0309339427734258688, 0.0559466810510877933},
            {0.1292270850663260170e-4, 0.2652814182060691422e-5, 0.2568431853930292259e-5}},
            1e-12, 1e-16);

        /* VectorMatrix/VectorOps */

        // DECLSPEC double iauPdp(double a[3], double b[3]);
        p1 = new double[] { 2.0, 2.0, 3.0 };
        p2 = new double[] { 1.0, 3.0, 4.0 };
        SofaValidate.EqualsApprox(Pdp(p1, p2), 20, 1e-12);

        // DECLSPEC double iauPm(double p[3]);
        p1 = new double[] { 0.3, 1.2, -2.5 };
        SofaValidate.EqualsApprox(Pm(p1), 2.789265136196270604, 1e-12);

        // DECLSPEC void iauPmp(double a[3], double b[3], double amb[3]);
        p1 = new double[] { 2.0, 2.0, 3.0 };
        p2 = new double[] { 1.0, 3.0, 4.0 };
        Pmp(p1, p2, p1);
        SofaValidate.EqualsApprox(p1, new double[] { 1.0, -1.0, -1.0 }, 1e-12);

        // DECLSPEC void iauPn(double p[3], double *r, double u[3]);
        p1 = new double[] { 0.3, 1.2, -2.5 };
        Pn(p1, out r, p2);
        SofaValidate.EqualsApprox(r, 2.789265136196270604, 1e-12);
        SofaValidate.EqualsApprox(p2, new double[]{
            0.1075552109073112058, 0.4302208436292448232, -0.8962934242275933816}, 1e-12);

        // DECLSPEC void iauPpp(double a[3], double b[3], double apb[3]);
        p1 = new double[] { 2.0, 2.0, 3.0 };
        p2 = new double[] { 1.0, 3.0, 4.0 };
        Ppp(p1, p2, p1);
        SofaValidate.EqualsApprox(p1, new double[] { 3.0, 5.0, 7.0 }, 1e-12);

        // DECLSPEC void iauPpsp(double a[3], double s, double b[3], double apsb[3]);
        p1 = new double[] { 2.0, 2.0, 3.0 };
        p2 = new double[] { 1.0, 3.0, 4.0 };
        Ppsp(p1, 5.0, p2, p1);
        SofaValidate.EqualsApprox(p1, new double[] { 7.0, 17.0, 23.0 }, 1e-12);

        // DECLSPEC void iauPvdpv(double a[2][3], double b[2][3], double adb[2]);
        pv1 = new double[,] { { 2.0, 2.0, 3.0 }, { 6.0, 0.0, 4.0 } };
        pv2 = new double[,] { { 1.0, 3.0, 4.0 }, { 0.0, 2.0, 8.0 } };
        Pvdpv(pv1, pv2, pvdpv);
        SofaValidate.EqualsApprox(pvdpv[0], 20.0, 1e-12);
        SofaValidate.EqualsApprox(pvdpv[1], 50.0, 1e-12);

        // DECLSPEC void iauPvm(double pv[2][3], double *r, double *s);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { 0.45, -0.25, 1.1 } };
        Pvm(pv1, out r, out rd);
        SofaValidate.EqualsApprox(r, 2.789265136196270604, 1e-12);
        SofaValidate.EqualsApprox(rd, 1.214495780149111922, 1e-12);

        // DECLSPEC void iauPvmpv(double a[2][3], double b[2][3], double amb[2][3]);
        pv1 = new double[,] { { 2.0, 2.0, 3.0 }, { 5.0, 6.0, 3.0 } };
        pv2 = new double[,] { { 1.0, 3.0, 4.0 }, { 3.0, 2.0, 1.0 } };
        Pvmpv(pv1, pv2, pv1);
        SofaValidate.EqualsApprox(pv1, new double[,] {
            { 1.0, -1.0, -1.0 }, { 2.0, 4.0, 2.0 } }, 1e-12);

        // DECLSPEC void iauPvppv(double a[2][3], double b[2][3], double apb[2][3]);
        pv1 = new double[,] { { 2.0, 2.0, 3.0 }, { 5.0, 6.0, 3.0 } };
        pv2 = new double[,] { { 1.0, 3.0, 4.0 }, { 3.0, 2.0, 1.0 } };
        Pvppv(pv1, pv2, pv2);
        SofaValidate.EqualsApprox(pv2, new double[,]{{3.0, 5.0, 7.0},{8.0, 8.0, 4.0}}, 1e-12);

        // DECLSPEC void iauPvu(double dt, double pv[2][3], double upv[2][3]);
        pv1 = new double[,]{
            {126668.5912743160734, 2136.792716839935565, -245251.2339876830229},
            {-0.4051854035740713039e-2, -0.6253919754866175788e-2,  0.1189353719774107615e-1}};
        Pvu(2920.0, pv1, pv2);
        SofaValidate.EqualsApprox(pv2, new double[,]{
            {126656.7598605317105, 2118.531271155726332, -245216.5048590656190},
            {-0.4051854035740713039e-2, -0.6253919754866175788e-2, 0.1189353719774107615e-1}},
            1e-12);

        // DECLSPEC void iauPvup(double dt, double pv[2][3], double p[3]);
        pv1 = new double[,]{
            {126668.5912743160734, 2136.792716839935565, -245251.2339876830229},
            {-0.4051854035740713039e-2, -0.6253919754866175788e-2,  0.1189353719774107615e-1}};
        Pvup(2920.0, pv1, p1);
        SofaValidate.EqualsApprox(p1, new double[]{
            126656.7598605317105, 2118.531271155726332, -245216.5048590656190}, 1e-12);

        // DECLSPEC void iauPvxpv(double a[2][3], double b[2][3], double axb[2][3]);
        pv1 = new double[,] { { 2.0, 2.0, 3.0 }, { 6.0, 0.0, 4.0 } };
        pv2 = new double[,] { { 1.0, 3.0, 4.0 }, { 0.0, 2.0, 8.0 } };
        Pvxpv(pv1, pv2, pv1);
        SofaValidate.EqualsApprox(pv1, new double[,]{
            {-1.0, -5.0, 4.0}, {-2.0, -36.0, 22.0}}, 1e-12);

        // DECLSPEC void iauPxp(double a[3], double b[3], double axb[3]);
        p1 = new double[] { 2.0, 2.0, 3.0 };
        p2 = new double[] { 1.0, 3.0, 4.0 };
        Pxp(p1, p2, p1);
        SofaValidate.EqualsApprox(p1, new double[] { -1.0, -5.0, 4.0 }, 1e-12);

        // DECLSPEC void iauS2xpv(double s1, double s2, double pv[2][3], double spv[2][3]);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { 0.5, 2.3, -0.4 } };
        S2xpv(2.0, 3.0, pv1, pv2);
        SofaValidate.EqualsApprox(pv2, new double[,] { { 0.6, 2.4, -5.0 }, { 1.5, 6.9, -1.2 } },
            1e-12);

        // DECLSPEC void iauSxp(double s, double p[3], double sp[3]);
        p1 = new double[] { 0.3, 1.2, -2.5 };
        Sxp(2.0, p1, p2);
        SofaValidate.EqualsApprox(p2, new double[] { 0.6, 2.4, -5.0 }, 1e-12);

        // DECLSPEC void iauSxpv(double s, double pv[2][3], double spv[2][3]);
        pv1 = new double[,] { { 0.3, 1.2, -2.5 }, { 0.5, 3.2, -0.7 } };
        Sxpv(2.0, pv1, pv2);
        SofaValidate.EqualsApprox(pv2, new double[,] { { 0.6, 2.4, -5.0 }, { 1.0, 6.4, -1.4 } },
            1e-12);
    }
#endif  // end debug self test code
}   // end partial class Sofa
}   // end namespace HirosePS
